@@ -1,11 +1,12 @@
 # Vend::Cart - Interchange shopping cart management routines
 #
-# $Id: Cart.pm,v 2.1.2.1 2002/01/24 05:07:01 jon Exp $
+# $Id: Cart.pm,v 2.10 2003/11/17 14:24:11 mheins Exp $
 #
-# Copyright (C) 1996-2002 Red Hat, Inc. <interchange@redhat.com>
+# Copyright (C) 2002-2003 Interchange Development Group
+# Copyright (C) 1996-2002 Red Hat, Inc.
 #
 # This program was originally based on Vend 0.2 and 0.3
-# Copyright 1995 by Andrew M. Wilcox <awilcox@world.std.com>
+# Copyright 1995 by Andrew M. Wilcox <amw@wilcoxsolutions.com>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -24,27 +25,64 @@
 
 package Vend::Cart;
 
-$VERSION = substr(q$Revision: 2.1.2.1 $, 10);
+$VERSION = substr(q$Revision: 2.10 $, 10);
 
 use strict;
 
 sub TIESCALAR {
 	my $class = shift;
 	my $instance = shift || undef;
+	$Vend::CurrentCart = 'main';
+	$::Levies = $Vend::Session->{levies}{main} ||= [];
 	return bless \$instance => $class;
 }
 
 sub FETCH {
-	return scalar ($::Carts->{$Vend::CurrentCart || 'main'} ||= []);
+	my $cartname = $Vend::CurrentCart;
+	$::Levies = $Vend::Session->{levies}{$cartname} ||= [];
+	return scalar ($::Carts->{$cartname} ||= []);
 }
 
 sub STORE {
 	my ($self, $cart) = @_;
 	my $name;
-	if(ref $cart) {
+	if( ref($cart) eq 'ARRAY' ) {
 		for(keys %$::Carts) {
+#::logDebug("checking name $_ via ref comparison");
 			$name = $_ if $::Carts->{$_} eq $cart;
 		}
+
+		if (! $name) {
+			$name = $cart->[0]{mv_cartname} if $cart->[0]{mv_cartname};
+		}
+
+		if (! $name) {
+			for my $pname (keys %$::Carts) {
+#::logDebug("checking name $pname via line comparison");
+				my $pros = $::Carts->{$pname};
+				next if ref($pros) ne 'ARRAY';
+				next if @$pros != @$cart;
+				CHECKLINES: {
+					for( my $i = 0; $i < @$pros; $i++ ) {
+						my $p = $pros->[$i];
+						my $c = $cart->[$i];
+						my @k1 = keys %$p;
+						my @k2 = keys %$c;
+						last CHECKLINES if @k1 != @k2;
+						foreach my $k (@k1) {
+							last CHECKLINES
+								unless exists $c->{$k};
+							last CHECKLINES
+								unless $c->{$k} eq $p->{$k};
+						}
+					}
+#::logDebug("found name $pname via line comparison");
+					$name = $pname;
+				}
+				last if $name;
+			}
+		}
+
 		if (! $name) {
 			$name = 'UNKNOWN';
 			$::Carts->{UNKNOWN} = $cart;
@@ -54,6 +92,7 @@ sub STORE {
 	else {
 		$Vend::CurrentCart = $cart;
 	}
+	$::Levies = $Vend::Session->{levies}{$Vend::CurrentCart} ||= [];
 	return $::Carts->{$Vend::CurrentCart};
 }
 
@@ -135,40 +174,82 @@ sub DESTROY { }
 # If the user has put in "0" for any quantity, delete that item
 # from the order list.
 sub toss_cart {
-	my($s) = @_;
+	my($s, $cartname) = @_;
 	my $i;
 	my $sub;
 	my (@master);
 	my (@cascade);
 	DELETE: for (;;) {
 		foreach $i (0 .. $#$s) {
+			my $item = $s->[$i];
 			if ($sub = $Vend::Cfg->{ItemAction}{$s->[$i]{code}}) {
-				$sub->($s->[$i]);
+				$sub->($item);
 			}
-			if ($s->[$i]->{quantity} <= 0) {
-				next if defined $s->[$i]->{mv_control} and
-								$s->[$i]->{mv_control} =~ /\bnotoss\b/;
-				if ($s->[$i]->{mv_mi} && ! $s->[$i]->{mv_si}) {
-					push (@master, $s->[$i]->{mv_mi});
+			if ($item->{quantity} <= 0) {
+				next if defined $item->{mv_control} and
+								$item->{mv_control} =~ /\bnotoss\b/;
+				if ($item->{mv_mi} && ! $item->{mv_si}) {
+					push (@master, $item->{mv_mi});
 				}
-				elsif ( $s->[$i]->{mv_ci} ) {
-					push (@master, $s->[$i]->{mv_ci});
+				elsif ( $item->{mv_ci} ) {
+					push (@master, $item->{mv_ci});
 				}
 				splice(@$s, $i, 1);
 				next DELETE;
 			}
+
+			if($Vend::Cfg->{MinQuantityField}) {
+				if(! defined $item->{mv_min_quantity}) {
+					my ($tab, $col) = split /:+/, $Vend::Cfg->{MinQuantityField};
+					if(! length $col) {
+						$col = $tab;
+						$tab = $item->{mv_ib} || $Vend::Cfg->{ProductFiles}[0];
+					}
+					$item->{mv_min_quantity} = ::tag_data($tab, $col, $item->{code})
+											 || '';
+				}
+
+				if(
+					length $item->{mv_min_quantity}
+					and 
+					$item->{quantity} < $item->{mv_min_quantity}
+					)
+				{
+					$item->{quantity} = $item->{mv_min_quantity};
+					$item->{mv_min_under} = 1;
+				}
+			}
+
+			if($Vend::Cfg->{MaxQuantityField}) {
+				my ($tab, $col) = split /:+/, $Vend::Cfg->{MaxQuantityField};
+				if(! length $col) {
+					$col = $tab;
+					$tab = $item->{mv_ib} || $Vend::Cfg->{ProductFiles}[0];
+				}
+				$item->{mv_max_quantity} = ::tag_data($tab, $col, $item->{code});
+
+				if(
+					length $item->{mv_max_quantity}
+					and 
+					$item->{quantity} > $item->{mv_max_quantity}
+					)
+				{
+					$item->{quantity} = $item->{mv_max_quantity};
+					$item->{mv_max_over} = 1;
+				}
+			}
+
 			next unless $Vend::Cfg->{Limit}{cart_quantity_per_line};
 			
-			$s->[$i]->{quantity} = $Vend::Cfg->{Limit}{cart_quantity_per_line}
+			$item->{quantity} = $Vend::Cfg->{Limit}{cart_quantity_per_line}
 				if
-					$s->[$i]->{quantity}
+					$item->{quantity}
 						>
 					$Vend::Cfg->{Limit}{cart_quantity_per_line};
 		}
 		last DELETE;
 	}
 
-	return 1 unless @master;
 	my $mi;
 	my %save;
 	my @items;
@@ -195,7 +276,8 @@ sub toss_cart {
 			@{$s} = @items[sort {$a <=> $b} keys %save];
 		}
 	}
-	1;
+	Vend::Interpolate::levies(1, $cartname);
+	return 1;
 }
 
 =head2 Test footer for item toss
