@@ -1,8 +1,9 @@
 # Vend::Error - Handle Interchange error pages and messages
 # 
-# $Id: Error.pm,v 2.1.2.2 2002/01/24 05:07:01 jon Exp $
+# $Id: Error.pm,v 2.8 2004/01/30 17:35:03 mheins Exp $
 #
-# Copyright (C) 1996-2002 Red Hat, Inc. <interchange@redhat.com>
+# Copyright (C) 2002-2003 Interchange Development Group
+# Copyright (C) 1996-2002 Red Hat, Inc.
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -37,7 +38,7 @@ use strict;
 
 use vars qw/$VERSION/;
 
-$VERSION = substr(q$Revision: 2.1.2.2 $, 10);
+$VERSION = substr(q$Revision: 2.8 $, 10);
 
 sub get_locale_message {
 	my ($code, $message, @arg) = @_;
@@ -81,7 +82,8 @@ sub interaction_error {
     $page = readin(find_special_page('interact'));
     if (defined $page) {
 		$page =~ s#\[message\]#$msg#ig;
-		::response(::interpolate_html($page, 1));
+		::interpolate_html($page, 1);
+		::response();
     }
 	else {
 		logError( "Missing special page: interact" , '');
@@ -109,10 +111,11 @@ EOF
 
 sub full_dump {
 	my $portion = shift;
+	my $opt = shift || {};
 	my $out = '';
 	if($portion) {
 		$out .= "###### SESSION ($portion) #####\n";
-		$out .= ::uneval($Vend::Session->{$portion});
+		$out .= uneval($Vend::Session->{$portion});
 		$out .= "\n###### END SESSION    #####\n";
 		$out =~ s/\0/\\0/g;
 		return $out;
@@ -120,22 +123,36 @@ sub full_dump {
 
 	$out = minidump();
 	local($Data::Dumper::Indent) = 2;
-	unless(caller() eq 'Vend::SOAP') {
+	unless ($opt->{no_env}) {
 		$out .= "###### ENVIRONMENT     #####\n";
-		$out .= ::uneval(::http()->{env});
+		if(my $h = ::http()) {
+			$out .= uneval($h->{env});
+		}
+		else {
+			$out .= uneval(\%ENV);
+		}
 		$out .= "\n###### END ENVIRONMENT #####\n";
 	}
-	$out .= "###### CGI VALUES      #####\n";
-	$out .= ::uneval(\%CGI::values);
-	$out .= "\n###### END CGI VALUES  #####\n";
-	$out .= "###### SESSION         #####\n";
-	$out .= ::uneval($Vend::Session);
-	$out .= "\n###### END SESSION    #####\n";
+	unless($opt->{no_cgi}) {
+		my %cgi = %CGI::values;
+		unless($opt->{show_all}) {
+			for(@Global::HideCGI) {
+				delete $cgi{$_};
+			}
+		}
+		$out .= "###### CGI VALUES      #####\n";
+		$out .= uneval(\%cgi);
+		$out .= "\n###### END CGI VALUES  #####\n";
+	}
+	unless($opt->{no_session}) {
+		$out .= "###### SESSION         #####\n";
+		$out .= uneval($Vend::Session);
+		$out .= "\n###### END SESSION    #####\n";
+	}
 	$out =~ s/\0/\\0/g;
 	return $out;
 }
 
-
 sub do_lockout {
 	my ($cmd);
 	my $msg = shift || '';
