@@ -1,11 +1,12 @@
 # Vend::Page - Handle Interchange page routing
 # 
-# $Id: Page.pm,v 2.0.2.3 2002/01/24 05:07:01 jon Exp $
+# $Id: Page.pm,v 2.18 2004/04/02 17:19:20 mheins Exp $
 #
-# Copyright (C) 1996-2002 Red Hat, Inc. <interchange@redhat.com>
+# Copyright (C) 2002-2003 Interchange Development Group
+# Copyright (C) 1996-2002 Red Hat, Inc.
 #
 # This program was originally based on Vend 0.2 and 0.3
-# Copyright 1995 by Andrew M. Wilcox <awilcox@world.std.com>
+# Copyright 1995 by Andrew M. Wilcox <amw@wilcoxsolutions.com>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -45,7 +46,7 @@ use strict;
 
 use vars qw/$VERSION/;
 
-$VERSION = substr(q$Revision: 2.0.2.3 $, 10);
+$VERSION = substr(q$Revision: 2.18 $, 10);
 
 my $wantref = 1;
 
@@ -53,25 +54,31 @@ sub display_special_page {
 	my($name, $subject) = @_;
 	my($page);
 
-	$name =~ m/[\[<]+/g
+	undef $Vend::write_redirect;
+
+	$name =~ m/[\[<]|[\@_]_[A-Z]\w+_[\@_]|\@\@[A-Z]\w+\@\@/
 		and do {
 			::logGlobal(
 					"Security violation -- scripting character in page name '%s'.",
 					$name,
 				);
-			$name = 'violation';
+			$name = find_special_page('violation');
+			1 while $subject =~ s/[\@_]_/_/g;
 		};
 
-	$subject = $subject || 'unspecified error';
+	$subject ||= 'unspecified error';
+
+	my $noname = $name;
+	$noname =~ s:^\.\./::;
 	
-#::logDebug("looking for special_page=$name");
-	$page = readfile($name, $Global::NoAbsolute, 1) || readin($name);
+	$page = readfile($noname, $Global::NoAbsolute, 1) || readin($name);
 
 	die ::get_locale_message(412, "Missing special page: %s\n", $name)
 		unless defined $page;
-#::logDebug("displaying special_page=$name");
 	$page =~ s#\[subject\]#$subject#ig;
-	return ::response(::interpolate_html($page, 1));
+	$Vend::PageInit = 0;
+	interpolate_html($page, 1);
+	::response();
 }
 
 # Displays the catalog page NAME.  If the file is not found, displays
@@ -79,49 +86,71 @@ sub display_special_page {
 # 
 
 sub display_page {
-	my($name) = @_;
+	my($name, $opt) = @_;
 	my($page);
 
-	$name =~ m/[\[<]+/g
+	$name ||= $CGI::values{mv_nextpage};
+
+	$name =~ m/[\[<]|[\@_]_[A-Z]\w+_[\@_]|\@\@[A-Z]\w+\@\@/
 		and do {
 			::logGlobal(
 					"Security violation -- scripting character in page name '%s'.",
 					$name,
 				);
-			$name = 'violation';
+			$name = find_special_page('violation');
 			return display_special_page($name);
 		};
 
-	$name = $CGI::values{mv_nextpage} unless $name;
-#::logDebug("display_page: $name");
 	if($Vend::Cfg->{ExtraSecure} and
 		$Vend::Cfg->{AlwaysSecure}->{$name}
 		and !$CGI::secure) {
 		$name = find_special_page('violation');
 	}
 
-	$page = readin($name);
+	$page = $Vend::VirtualPage || readin($name);
 # TRACK
-	if (defined $page) {
+	if (defined $page && $Vend::Track) {
 		$Vend::Track->view_page($name);
 	}
 # END TRACK	
 		
-	my $opt;
+	my $inth_opt;
 	# Try for on-the-fly if not there
 	if(! defined $page) {
 		$page = Vend::Interpolate::fly_page($name)
-			and $opt->{onfly} = 1;
+			and $inth_opt->{onfly} = 1;
+	}
+
+	# Try one last time for page with index
+	if(! defined $page and $Vend::Cfg->{DirectoryIndex}) {
+		my $try = $name;
+		$try =~ s!/*$!/$Vend::Cfg->{DirectoryIndex}!;
+		$page = readin($try);
 	}
 
 	if (defined $page) {
-		::response(::interpolate_html($page, 1, $opt));
+		$Vend::PageInit = 0;
+		if ($opt->{return}) {
+			return ::interpolate_html($page, 1, $inth_opt);
+		} else {
+			::interpolate_html($page, 1, $inth_opt);
+		}
+		::response();
 		return 1;
 	}
 	else {
-		$name =~ s/\&/&amp;/g;
-		$name =~ s/\[/&#91;/g;
-		$name =~ s/\</&lt;/g;
+		my $handled;
+		my $newpage;
+		if(my $subname = $Vend::Cfg->{SpecialSub}{missing}) {
+			my $sub = $Vend::Cfg->{Sub}{$subname} || $Global::GlobalSub->{$subname};
+			($handled, $newpage) = $sub->($name)
+				if $sub;
+		}
+		if($handled) {
+			return display_page($newpage) if $newpage;
+			return 0;
+		}
+		HTML::Entities::encode($name, $ESCAPE_CHARS::std);
 		display_special_page(find_special_page('missing'), $name);
 		return 0;
 	}
@@ -138,7 +167,7 @@ sub do_page {
 sub do_search {
 	my($c) = \%CGI::values;
 	::update_user();
-#::logDebug($more);
+
 	if ($c->{mv_more_matches}) {
 		$Vend::Session->{last_search} = "scan/MM=$c->{mv_more_matches}";
 		$c->{mv_more_matches} =~ m/([a-zA-Z0-9])+/;
@@ -188,4 +217,83 @@ sub do_scan {
 	return 1;
 }
 
+sub output_test {
+	my ($tag) = @_;
+	my $ary;
+	return '' unless $ary = $Vend::OutPtr{lc $tag};
+	for(@$ary) {
+		next unless $Vend::Output[$_];
+		next unless length(${$Vend::Output[$_]});
+		return 1;
+	}
+	return '';
+}
+
+sub output_cat {
+	my ($tag) = @_;
+	my $ary;
+	return '' unless $ary = $Vend::OutPtr{lc $tag};
+	my $out = '';
+	for(@$ary) {
+		next unless $Vend::Output[$_];
+		$out .= ${$Vend::Output[$_]};
+		undef $Vend::Output[$_];
+	}
+	return $out;
+}
+
+sub output_ary {
+	my ($tag) = @_;
+	my $ary;
+	return '' unless $ary = $Vend::OutPtr{lc $tag};
+	my @out;
+	for(@$ary) {
+		next unless $Vend::Output[$_];
+		push @out, ${$Vend::Output[$_]};
+		undef $Vend::Output[$_];
+	}
+	return \@out;
+}
+
+sub output_rest {
+	my ($tag) = @_;
+	my $out = '';
+	for(@$Vend::Output) {
+		next unless $_;
+		$out .= ${$Vend::Output[$_]};
+		undef $Vend::Output[$_];
+	}
+	return $out;
+}
+
+sub templatize {
+	my ($template) = @_;
+	$template ||= $Vend::Cfg->{PageTemplate} || '{:REST}';
+#::logDebug("Templatizing, template length=" . length($template));
+	my $body = $template;
+
+	$body =~ s!\{\{\@([A-Z][A-Z_]*[A-Z])\}\}(.*?)\{\{/\@\1\}\}!
+					my $tag = lc $1;
+					my $ary;
+					return '' unless $ary = $Vend::OutPtr{$tag};
+					my $tpl = $2;
+					my $out = '';
+					for(@$ary) {
+						my $ref = $Vend::Output[$_]
+							or next;
+						my $chunk = $tpl;
+						$chunk =~ s/\{$tag\}/$$ref/;
+						undef $Vend::Output[$_];
+						$out .= $chunk;
+					}
+					$out;
+				!sge;
+	1 while $body =~ s!\{\{([A-Z][A-Z_]*[A-Z])\?\}\}(.*?)\{\{/\1\?\}\}! output_test(lc $1) ? $2 : ''!egs;
+	1 while $body =~ s!\{\{([A-Z][A-Z_]*[A-Z])\:\}\}(.*?)\{\{/\1\:\}\}! output_test(lc $1) ? '' : $2!egs;
+	$body =~ s!\{\{([A-Z][A-Z_]*[A-Z])\}\}!output_cat($1)!eg;
+	$body =~ s!\{\{:DEFAULT\}\}!output_cat('')!e;
+	$body =~ s!\{\{:REST\}\}!output_rest('')!e;
+	@Vend::Output = (\$body);
+}
+
 1;
