@@ -1,18 +1,21 @@
 # Vend::Payment::AuthorizeNet - Interchange AuthorizeNet support
 #
-# $Id: AuthorizeNet.pm,v 2.1.2.3 2002/01/24 05:07:03 jon Exp $
+# Connection routine for AuthorizeNet version 3 using the 'ADC Direct Response'
+# method.
 #
-# Copyright (C) 1999-2002 Red Hat, Inc. <interchange@redhat.com>
+# $Id: AuthorizeNet.pm,v 2.12 2003/09/11 20:12:56 kwalsh Exp $
 #
-# by mark@summersault.com with code reused and inspired by
-#	Mike Heins <mheins@redhat.com>
-#	webmaster@nameastar.net
-#   Jeff Nappi <brage@cyberhighway.net>
-#   Paul Delys <paul@gi.alaska.edu>
-#  Edited by Ray Desjardins <ray@dfwmicrotech.com>
-
-# Patches for AUTH_CAPTURE and VOID support contributed by
-# nferrari@ccsc.com (Nelson H Ferrari)
+# Copyright (C) 2003 Interchange Development Group, http://www.icdevgroup.org/
+# Copyright (C) 1999-2002 Red Hat, Inc.
+#
+# Authors:
+# mark@summersault.com
+# Mike Heins <mike@perusion.com>
+# Jeff Nappi <brage@cyberhighway.net>
+# Paul Delys <paul@gi.alaska.edu>
+# webmaster@nameastar.net
+# Ray Desjardins <ray@dfwmicrotech.com>
+# Nelson H Ferrari <nferrari@ccsc.com>
 
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -29,16 +32,11 @@
 # Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 # MA  02111-1307  USA.
 
-# Connection routine for AuthorizeNet version 3 using the 'ADC Direct Response'
-# method.
-
-# Reworked extensively to support new Interchange payment stuff by Mike Heins
-
 package Vend::Payment::AuthorizeNet;
 
-=head1 Interchange AuthorizeNet Support
+=head1 NAME
 
-Vend::Payment::AuthorizeNet $Revision: 2.1.2.3 $
+Vend::Payment::AuthorizeNet - Interchange AuthorizeNet Support
 
 =head1 SYNOPSIS
 
@@ -209,8 +207,8 @@ That should show what happened.
 
 =item *
 
-If all else fails, Red Hat and other consultants are available to help
-with integration for a fee.
+If all else fails, consultants are available to help with integration for a fee.
+See http://www.icdevgroup.org/ for mailing lists and other information.
 
 =back
 
@@ -221,8 +219,8 @@ to Vend::Payment and places things there.
 
 =head1 AUTHORS
 
-Mark Stosberg <mark@summersault.com>, based on original code by Mike Heins
-<mheins@redhat.com>.
+Mark Stosberg <mark@summersault.com>.
+Based on original code by Mike Heins <mike@perusion.com>.
 
 =head1 CREDITS
 
@@ -271,10 +269,14 @@ BEGIN {
 }
 
 package Vend::Payment;
+use strict;
+
 sub authorizenet {
 	my ($user, $amount) = @_;
 
 	my $opt;
+	my $secret;
+	
 	if(ref $user) {
 		$opt = $user;
 		$user = $opt->{id} || undef;
@@ -316,6 +318,18 @@ sub authorizenet {
 	my $referer   =  $opt->{referer}
 					|| charge_param('referer');
 
+	my @override = qw/
+						order_id
+						auth_code
+						mv_credit_card_exp_month
+						mv_credit_card_exp_year
+						mv_credit_card_number
+					/;
+	for(@override) {
+		next unless defined $opt->{$_};
+		$actual->{$_} = $opt->{$_};
+	}
+
 	## Authorizenet does things a bit different, ensure we are OK
 	$actual->{mv_credit_card_exp_month} =~ s/\D//g;
     $actual->{mv_credit_card_exp_month} =~ s/^0+//;
@@ -339,17 +353,17 @@ sub authorizenet {
 		CREDIT					=>	'CREDIT',
 		PRIOR_AUTH_CAPTURE		=>	'PRIOR_AUTH_CAPTURE',
 		VOID					=>	'VOID',
-		auth		 			=>	'ONLY',
+		auth		 			=>	'AUTH_ONLY',
 		authorize		 		=>	'AUTH_ONLY',
 		mauthcapture 			=>	'AUTH_CAPTURE',
 		mauthonly				=>	'AUTH_ONLY',
 		return					=>	'CREDIT',
-		reverse           		=>	'PRIOR_AUTH_CAPTURE',
+		settle_prior        	=>	'PRIOR_AUTH_CAPTURE',
 		sale		 			=>	'AUTH_CAPTURE',
 		settle      			=>  'CAPTURE_ONLY',
 		void					=>	'VOID',
 	);
-	
+
 	if (defined $type_map{$transtype}) {
         $transtype = $type_map{$transtype};
     }
@@ -361,34 +375,44 @@ sub authorizenet {
         $amount = Vend::Util::round_to_frac_digits($amount,$precision);
     }
 
-	$order_id = gen_order_id($opt);
+	my $order_id = gen_order_id($opt);
 
+#::logDebug("auth_code=$actual->{auth_code} order_id=$opt->{order_id}");
     my %query = (
-                    x_Test_Request	=> $opt->{test} || charge_param('test'),
-                    x_Card_Num		=> $actual->{mv_credit_card_number},
-                    x_First_Name    => $actual->{b_fname},
-                    x_Last_Name     => $actual->{b_lname},
-                    x_Address       => $actual->{address},
-                    x_City          => $actual->{b_city},
-                    x_State         => $actual->{b_state},
-                    x_Zip			=> $actual->{b_zip},
-                    x_Country		=> $actual->{b_country},
-					x_Type			=> $transtype,
-                    x_Amount    	=> $amount,
-                    x_Exp_Date  	=> $exp,
-                    x_Method    	=> 'CC',
-					x_Trans_ID		=> $actual->{order_id},
-					x_Auth_Code		=> $actual->{auth_code},
-                    x_Invoice_Num   => $actual->{mv_order_number},
-#                    x_Company      => $actual->{company},
-                    x_Email         => $actual->{email},
-                    x_Phone        => $actual->{phone_day},
-                    x_Password  	=> $secret,
-                    x_Login     	=> $user,
-                    x_Version   	=> '3.0',
-                    x_ADC_URL   	=> 'FALSE',
-                    x_ADC_Delim_Data	=> 'TRUE',
-
+		x_Test_Request			=> $opt->{test} || charge_param('test'),
+		x_First_Name			=> $actual->{b_fname},
+		x_Last_Name				=> $actual->{b_lname},
+		x_Company				=> $actual->{b_company},
+		x_Address				=> $actual->{b_address},
+		x_City					=> $actual->{b_city},
+		x_State					=> $actual->{b_state},
+		x_Zip					=> $actual->{b_zip},
+		x_Country				=> $actual->{b_country},
+		x_Ship_To_First_Name	=> $actual->{fname},
+		x_Ship_To_Last_Name		=> $actual->{lname},
+		x_Ship_To_Company		=> $actual->{company},
+		x_Ship_To_Address		=> $actual->{address},
+		x_Ship_To_City			=> $actual->{city},
+		x_Ship_To_State			=> $actual->{state},
+		x_Ship_To_Zip			=> $actual->{zip},
+		x_Ship_To_Country		=> $actual->{country},
+		x_Email					=> $actual->{email},
+		x_Phone					=> $actual->{phone_day},
+		x_Type					=> $transtype,
+		x_Amount				=> $amount,
+		x_Method				=> 'CC',
+		x_Card_Num				=> $actual->{mv_credit_card_number},
+		x_Exp_Date				=> $exp,
+		x_Card_Code				=> $actual->{cvv2},
+		x_Trans_ID				=> $actual->{order_id},
+		x_Auth_Code				=> $actual->{auth_code},
+		x_Invoice_Num			=> $actual->{mv_order_number},
+		x_Password				=> $secret,
+		x_Login					=> $user,
+		x_Version				=> '3.1',
+		x_ADC_URL				=> 'FALSE',
+		x_ADC_Delim_Data		=> 'TRUE',
+		x_ADC_Delim_Character	=> "\037",
     );
 
     my @query;
@@ -423,6 +447,7 @@ sub authorizenet {
             pop.avs_code          x_avs_code
             pop.avs_zip           x_zip
             pop.avs_addr          x_address
+            pop.cvv2_resp_code    x_cvv2_resp_code
     /
     );
 
@@ -439,9 +464,41 @@ sub authorizenet {
 			x_auth_code
 			x_avs_code
 			x_trans_id
+			x_invoice_num
+			x_description
+			x_amount
+			x_method
+			x_type
+			x_cust_id
+			x_first_name
+			x_last_name
+			x_company
+			x_address
+			x_city
+			x_state
+			x_zip
+			x_country
+			x_phone
+			x_fax
+			x_email
+			x_ship_to_first_name
+			x_ship_to_last_name
+			x_ship_to_company
+			x_ship_to_address
+			x_ship_to_city
+			x_ship_to_state
+			x_ship_to_zip
+			x_ship_to_country
+			x_tax
+			x_duty
+			x_freight
+			x_tax_exempt
+			x_po_num
+			x_MD5_hash
+			x_cvv2_resp_code			
 		/
 		}
-		 = split (/,/,$page);
+		 = split (/\037/,$page);
     	
 #::logDebug(qq{authorizenet response_reason_text=$result{x_response_reason_text} response_code: $result{x_response_code}});    	
 
@@ -470,6 +527,7 @@ sub authorizenet {
     		$result{MErrMsg} = errmsg($msg, $result{x_response_reason_text});
     	}
     }
+#::logDebug(qq{authorizenet result=} . uneval(\%result));    	
 
     return (%result);
 }
