@@ -1,12 +1,11 @@
 # Vend::Payment::Skipjack - Interchange Skipjack support
 #
-# $Id: Skipjack.pm,v 2.0.2.2 2002/02/09 03:44:24 mheins Exp $
+# $Id: Skipjack.pm,v 2.6 2003/08/04 22:01:09 racke Exp $
 #
-# Copyright (C) 1999-2002 Red Hat, Inc. <interchange@redhat.com>
+# Copyright (C) 2002-2003 Interchange Development Group
+# Copyright (C) 1999-2002 Red Hat, Inc.
 #
-# Written by Cameron Prince <cprince@redhat.com> and
-# Mark Johnson <markj@redhat.com>,
-# based on code by Mike Heins <mheins@redhat.com>
+# Written by Cameron Prince and Mark Johnson, based on code by Mike Heins
 
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -25,9 +24,9 @@
 
 package Vend::Payment::Skipjack;
 
-=head1 Interchange Skipjack Support
+=head1 NAME
 
-Vend::Payment::Skipjack $Revision: 2.0.2.2 $
+Vend::Payment::Skipjack - Interchange Skipjack Support
 
 =head1 SYNOPSIS
 
@@ -235,8 +234,8 @@ That should show what happened.
 
 =item *
 
-If all else fails, Red Hat and other consultants are available to help
-with integration for a fee.
+If all else fails, consultants are available to help with integration for a fee.
+See http://www.icdevgroup.org/
 
 =back
 
@@ -255,7 +254,7 @@ Modified for GetCareer.com by Slipstream.com by Troy Davis <troy@slipstream.com>
 LWP/Crypt::SSLeay interface code by Matthew Schick,
 <mschick@brightredproductions.com>.
 
-Interchange implementation by Mike Heins, <mheins@redhat.com>.
+Interchange implementation by Mike Heins.
 
 =cut
 
