@@ -1,8 +1,9 @@
 # Vend::Scan - Prepare searches for Interchange
 #
-# $Id: Scan.pm,v 2.0.2.4 2002/03/01 20:10:53 racke Exp $
+# $Id: Scan.pm,v 2.25 2003/07/30 04:00:45 mheins Exp $
 #
-# Copyright (C) 1996-2002 Red Hat, Inc. <interchange@redhat.com>
+# Copyright (C) 2002-2003 Interchange Development Group
+# Copyright (C) 1996-2002 Red Hat, Inc.
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -29,12 +30,17 @@ require Exporter;
 			perform_search
 			);
 
-$VERSION = substr(q$Revision: 2.0.2.4 $, 10);
+$VERSION = substr(q$Revision: 2.25 $, 10);
 
 use strict;
 use Vend::Util;
+use Vend::File;
+use Vend::SQL_Parser;
 use Vend::Interpolate;
 use Vend::Data qw(product_code_exists_ref column_index);
+use Vend::TextSearch;
+use Vend::DbSearch;
+use Vend::RefSearch;
 
 my @Order = ( qw(
 	mv_dict_look
@@ -73,6 +79,7 @@ my @Order = ( qw(
 	mv_more_id
 	mv_min_string
 	mv_max_matches
+	mv_no_hide
 	mv_orsearch
 	mv_range_min
 	mv_range_max
@@ -83,6 +90,7 @@ my @Order = ( qw(
 	mv_return_file_name
 	mv_return_reference
 	mv_substring_match
+	mv_small_data
 	mv_start_match
 	mv_return_spec
 	mv_spelling_errors
@@ -99,9 +107,15 @@ my @Order = ( qw(
 	mv_unique
 	mv_more_matches
 	mv_value
+	mv_next_search
+	mv_search_reference
 	prefix
 ));
 
+## Place marker, not used in search specs but is reserved
+##  rt  mv_real_table
+##  hf  mv_header_fields
+##
 my %Scan = ( qw(
 	ac  mv_all_chars
 	bd  mv_base_directory
@@ -121,6 +135,7 @@ my %Scan = ( qw(
 	er  mv_spelling_errors
 	ff  mv_field_file
 	fi  mv_search_file
+	ft  mv_field_title
 	fm  mv_first_match
 	fn  mv_field_names
 	hs  mv_head_skip
@@ -141,13 +156,16 @@ my %Scan = ( qw(
 	ms  mv_min_string
 	ne  mv_negate
 	ng  mv_negate
+	nh  mv_no_hide
 	np  mv_nextpage
+	ns  mv_next_search
 	nu  mv_numeric
 	op  mv_column_op
 	os  mv_orsearch
 	pf  prefix
 	ra  mv_return_all
 	rd  mv_return_delim
+	re	mv_search_reference
 	rf  mv_return_fields
 	rg  mv_range_alpha
 	rl  mv_range_look
@@ -156,7 +174,7 @@ my %Scan = ( qw(
 	rr  mv_return_reference
 	rs  mv_return_spec
 	rx  mv_range_max
-	SE  mv_raw_searchspec
+	sd  mv_small_data
 	se  mv_searchspec
 	sf  mv_search_field
 	sg  mv_search_group
@@ -189,14 +207,15 @@ my %Parse = (
 	mv_substring_match      =>  \&_yes_array,
 	mv_column_op            =>  \&_array,
 	mv_coordinate           =>  \&_yes,
+	mv_no_hide              =>  \&_yes,
 
 	mv_field_names          =>	\&_array,
 	mv_spelling_errors      => 	sub { my $n = int($_[1]); $n < 8 ? $n : 1; },
 	mv_dict_limit           =>  \&_dict_limit,
 	mv_exact_match          =>  \&_yes,
 	mv_head_skip            =>  \&_number,
-	mv_matchlimit           =>  sub { $_[1] =~ /(\d+)/ ? $1 : 50 },
-	mv_max_matches          =>  sub { $_[1] =~ /(\d+)/ ? $1 : 2000 },
+	mv_matchlimit           =>  \&_matchlimit,
+	mv_max_matches          =>  sub { $_[1] =~ /(\d+)/ ? $1 : -1 },
 	mv_min_string           =>  sub { $_[1] =~ /(\d+)/ ? $1 : 1 },
 	mv_profile              =>  \&parse_profile,
 	mv_range_alpha          =>  \&_array,
@@ -252,7 +271,7 @@ sub find_search_params {
 		$c = \%CGI::values;
 	}
 	else {
-		$param =~ s/__NULL__/\0/g;
+		$param =~ s/-_NULL_-/\0/g;
 		@args = split m:/:, $param;
 	}
 
@@ -381,7 +400,6 @@ sub finish_search {
 	$::Values->{mv_first_match}    = $q->{mv_first_match}
 			if defined $q->{mv_first_match};
 	$::Values->{mv_searchspec} 	   = $q->{mv_searchspec};
-	$::Values->{mv_raw_searchspec} = $q->{mv_raw_searchspec} || undef;
 	$::Values->{mv_raw_dict_look}  = $q->{mv_raw_dict_look}  || undef;
 	$::Values->{mv_dict_look}      = $q->{mv_dict_look} || undef;
 }
@@ -494,6 +512,9 @@ sub perform_search {
 		elsif (! $options{mv_searchtype} or $options{mv_searchtype} eq 'text') {
 			$q = new Vend::TextSearch %options;
 		}
+		elsif ( $options{mv_searchtype} eq 'ref'){
+			$q = new Vend::RefSearch %options;
+		}
 # GLIMPSE
 		elsif ( $options{mv_searchtype} eq 'glimpse'){
 			$q = new Vend::Glimpse %options;
@@ -533,11 +554,7 @@ sub perform_search {
 
 }
 
-BEGIN {
-	eval { require SQL::Statement; };
-}
-
-my %scalar = (qw/ st 1 ra 1 co 1 os 1/);
+my %scalar = (qw/ st 1 ra 1 co 1 os 1 sr 1 ml 1 ms 1/);
 
 sub push_spec {
 	my ($parm, $val, $ary, $hash) = @_;
@@ -567,29 +584,24 @@ sub sql_statement {
 	}
 
 	if ($table) {
-		push_spec('fi', $table, $ary, $hash)
+		push_spec('fi', $table, $ary, $hash), push_spec('rt', $table, $ary, $hash)
 # GLIMPSE
 			unless "\L$table" eq 'glimpse';
 # END GLIMPSE
 	}
 
-	die "SQL is not enabled for Interchange. Get the SQL::Statement module.\n"
-		unless defined &SQL::Statement::new;
-
-	my $parser = SQL::Parser->new('Ansi');
-
 	# Strip possible leading stuff
 	$text =~ s/^\s*sq\s*=//;
 	my $stmt;
 	eval {
-		$stmt = SQL::Statement->new($text, $parser);
+		$stmt = Vend::SQL_Parser->new($text, $ref);
 	};
 	if($@ and $text =~ s/^\s*sq\s*=(.*)//m) {
 #::logDebug("failed first query, error=$@");
 		my $query = $1;
 		push @$ary, $text if $ary;
 		eval {
-			$stmt = SQL::Statement->new($query, $parser);
+			$stmt = Vend::SQL_Parser->new($text, $ref);
 		};
 	}
 	if($@) {
@@ -601,27 +613,29 @@ sub sql_statement {
 	my $nuhash;
 	my $codename;
 
+#::logDebug("SQL statement=" . ::uneval($stmt));
+
 	my $update = $stmt->command();
+#::logDebug("SQL command=$update");
 	undef $update if $update eq 'SELECT';
-#	CODECHECK: {
-#		last CODECHECK if ! $update;
-#		my $i = 0;
-#		for($stmt->columns()) {
-#			($stmt->{MV_VALUE_RELOCATE} = $i, last)
-#				if $_ eq $codename || $_ eq '0';
-#			$i++;
-#		}
-#	}
 
 	for($stmt->tables()) {
 		my $t = $_->name();
+		if($ref->{table_only}) {
+			return $t;
+		}
+#::logDebug("found table=$t");
 
 		my $codename;
 		my $db = Vend::Data::database_exists_ref($t);
 		if($db) {
 			$codename = $db->config('KEY') || 'code';
-			$nuhash = $db->config('NUMERIC') || undef;
-			push_spec( 'fi', $Vend::Cfg->{Database}{$t}{'file'}, $ary, $hash);
+			# Only for first table, what else can we do?
+			$nuhash ||= $db->config('NUMERIC') || undef;
+			push_spec( 'fi', $db->config('file'), $ary, $hash);
+			push_spec( 'rt', $t, $ary, $hash);
+			$stmt->verbatim_fields(1)
+				if $db->config('VERBATIM_FIELDS');
 		}
 # GLIMPSE
 		elsif ("\L$t" eq 'glimpse') {
@@ -632,106 +646,72 @@ sub sql_statement {
 # END GLIMPSE
 		else {
 			push_spec('fi', $t, $ary, $hash);
+			push_spec('rt', $t, $ary, $hash);
 		}
 #::logDebug("t=$t obj=$_ db=$db nuhash=" . ::uneval($nuhash));
 	}
 
-	$text =~ /\bselect\s+distinct\s+/i and push_spec( 'un', 'yes', $ary, $hash);
+	if(my $l = $stmt->limit()) {
+#::logDebug("found limit=" . $l->limit());
+		push_spec('ml', $l->limit(), $ary, $hash);
+		if(my $fm = $l->offset()) {
+#::logDebug("found offset=$fm");
+			push_spec('fm', $fm, $ary, $hash);
+		}
+	}
 
+	my $distincted;
 	for($stmt->columns()) {
 		my $name = $_->name();
-		#($stmt->{MV_VALUE_RELOCATE} = 0, last) if $name eq '*';
+#::logDebug("found column=$name");
+		push_spec('un', 1, $ary, $hash) if $_->distinct() and ! $distincted++;
 		push_spec('rf', $name, $ary, $hash);
+		push_spec('hf', $_->as(), $ary, $hash);
 		last if $name eq '*';
 #::logDebug("column name=" . $_->name() . " table=" . $_->table());
 	}
-#	if(! $update) {
-#		# do nothing
-#	}
-#	elsif ($stmt->{mv_value_relocate}) {
-#		splice(@{$hash->{rf}}, $stmt->{mv_value_relocate}, 1);
-#	}
-#	elsif ($update eq 'insert') {
-#		$stmt->{mv_value_relocate} = 0 if ! $stmt->columns();
-#	}
-#
+
+	for my $v ($stmt->params()) {
+		my $val = $v->value();
+		my $type = $v->type();
+#::logDebug(qq{found value="$val" type=$type});
+		push_spec('vv', $val, $ary, $hash);
+		push_spec('vt', $type, $ary, $hash);
+	}
+
 	my @order;
 
 	@order = $stmt->order();
 	for(@order) {
 		my $c = $_->column();
+#::logDebug("found order column=$c");
 		push_spec('tf', $c, $ary, $hash);
 		my $d = $_->desc() ? 'fr' : 'f';
+		$d =~ s/f/n/ if exists $nuhash->{$c};
+#::logDebug("found order sense=$d");
 		push_spec('to', $d, $ary, $hash);
 	}
 
-	my $where;
+#::logDebug("ary spec to this point=" . ::uneval($ary));
+#::logDebug("hash spec to this point=" . ::uneval($hash));
 	my @where;
-	my $numeric;
 	@where = $stmt->where();
-# Account for undocumented behavior in SQL::Statement
-# Fix by Kestutis Lasys
-	if(CORE::ref $where[0]) {
-	  my $or;
-	  push_spec('co', 'yes', $ary, $hash);
-	  do {
-	  	my $where = shift @where;
-		my $op = $where->op();
-		my $col = $where->arg1();
-		my $spec = $where->arg2();
-#::logDebug("where=$where op=$op arg1=$col arg2=$spec");
-		OP: {
-			if($op eq 'OR') {
-				push_spec( 'os', 'yes', $ary, $hash)     unless $or++;
-				push(@where, $where->arg1() , $where->arg2());
-			}
-			elsif($op eq 'AND') {
-				push(@where, $where->arg1() , $where->arg2());
-			}
-			else {
-
-				my ($col, $spec);
-
-				# Search spec is a variable if a ref
-				$spec = $where->arg2();
-#::logDebug("where col=$col spec=$spec");
-				$spec = $ref->{$spec->name()}		if ref $spec;
-
-				last OP unless defined $spec;
-
-				# Column name is a variable if a string
-				$col = $where->arg1();
-				$col = ref $col ? $col->name() : $::Values->{$col};
-
-				last OP unless $col;
-
-				$numeric = (defined $nuhash)
-							? (exists $nuhash->{$col})
-							: (
-								$spec =~ /^-?\d+\.?\d*$/
-								and
-								$spec !~ /^0\d+$/			 );
-
-#::logDebug("where col=$col spec=$spec");
-
-#::logDebug("numeric for $col=$numeric");
-				push_spec('nu', ($numeric || 0),      $ary, $hash); 
-				push_spec('se', $spec,                $ary, $hash);
-				push_spec('op', $op,                  $ary, $hash);
-				push_spec('sf', $col,                 $ary, $hash);
-				push_spec('ne', ($where->neg() || 0), $ary, $hash);
-
-				
-			}
+#::logDebug("where returned=" . ::uneval(\@where));
+	if(@where) {
+		## In a SQL query, we never want to drop out on empty string
+		push_spec('ms', 0, $ary, $hash);
+		for(@where) {
+			push_spec( @$_, $ary, $hash );
 		}
-	  } while @where;
-
 	}
 	else {
 		push_spec('ra', 'yes', $ary, $hash);
 	}
 	
-#::logDebug("sql_statement output=" . Vend::Util::uneval($hash)) if $hash;
+	if($hash->{sg} and ! $hash->{sr}) {
+		delete $hash->{sg};
+	}
+#::logDebug("sql_statement output=" . Vend::Util::uneval_it($hash)) if $hash;
 	return ($hash, $stmt) if $hash;
 
 	my $string = join "\n", @$ary;
@@ -855,6 +835,7 @@ sub _verbatim_array {
 	my @fields;
 #::logDebug("receiving verbatim_array: " . ::uneval (\@_));
 	@fields = ref $_[1] ? @{$_[1]} : split /\0/, $_[1], -1;
+	@fields = ('') if ! @fields;
 	unshift(@fields, @{$_[2]}) if $_[2];
 	return \@fields;
 }
@@ -879,20 +860,18 @@ sub _scalar {
 	defined $_[1] ? $_[1] : '';
 }
 
-my $Pat = ($^O =~ /win32/i) ? '([A-Za-z]:)?[\\/]' : '/';
-
 sub _file_security {
 	my ($junk, $param, $passed) = @_;
 	$passed = [] unless $passed;
 	my(@files) = grep /\S/, split /\s*[,\0]\s*/, $param, -1;
 	for(@files) {
-		my $ok = (m:^$Pat:o || /\.\./) ? 0 : 1;
+		my $ok = allowed_file($_);
 		if(!$ok) {
 			$ok = 1 if $_ eq $::Variable->{MV_SEARCH_FILE};
 			$ok = 1 if $::Scratch->{$_};
 		}
-		if($_ !~ /\./) {
-			$_ = $Vend::Cfg->{Database}{$_}{'file'}
+		if(/^\w+$/ and ! $::Variable->{MV_DEFAULT_SEARCH_DB}) {
+			$_ = $Vend::Cfg->{Database}{$_}{file}
 				if defined $Vend::Cfg->{Database}{$_};
 		}
 		if ($ok and $Vend::Cfg->{NoSearch} and /$Vend::Cfg->{NoSearch}/) {
@@ -945,5 +924,12 @@ sub _dict_limit {
 	return $_[1];
 }
 
+sub _matchlimit {
+	shift;
+	my $val = lc(shift);
+	return -1 if $val eq 'none' or $val eq 'all';
+	return int($val) || $::Variable->{MV_DEFAULT_MATCHLIMIT} || 50;
+}
+
 1;
 __END__
