@@ -1,8 +1,9 @@
 # Vend::Search - Base class for search engines
 #
-# $Id: Search.pm,v 2.0.2.2 2002/02/06 03:48:36 mheins Exp $
+# $Id: Search.pm,v 2.22 2004/04/08 15:05:09 mheins Exp $
 #
-# Copyright (C) 1996-2002 Red Hat, Inc. <interchange@redhat.com>
+# Copyright (C) 2002-2003 Interchange Development Group
+# Copyright (C) 1996-2002 Red Hat, Inc.
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -21,7 +22,7 @@
 
 package Vend::Search;
 
-$VERSION = substr(q$Revision: 2.0.2.2 $, 10);
+$VERSION = substr(q$Revision: 2.22 $, 10);
 
 use strict;
 use vars qw($VERSION);
@@ -91,6 +92,25 @@ my %maytag = (
 
 my (@hashable) = (qw/mv_return_fields mv_range_look mv_search_field mv_sort_field/);
 
+sub search_reference {
+	my ($s, $ref) = @_;
+	my $c = { mv_searchtype => 'ref', label => ref($s), mv_search_file => '__none__' };
+
+	my $ns = $s->{mv_next_search};
+	$ns = $::Scratch->{$ns} unless $ns =~ /=/;
+	my $params = Vend::Interpolate::escape_scan($ns);
+#::logDebug("search_params: $params");
+	Vend::Scan::find_search_params($c, $params);
+
+	$c->{mv_return_filtered} = 1;
+	$c->{mv_return_fields} = '*';
+	$c->{mv_field_names} = $s->{mv_field_names};
+	$c->{mv_search_reference} = $ref;
+#::logDebug("Ref ready to search: " . ::uneval($c));
+	my $o = Vend::Scan::perform_search($c);
+	return @{$o->{mv_results} || []};
+}
+
 sub hash_fields {
 	my ($s, $fn, @laundry) = @_;
 	my %fh;
@@ -130,10 +150,10 @@ sub hash_fields {
 
 sub escape {
     my($s, @text) = @_;
-#::logDebug( "text=@text");
+#::logDebug( "escaped text=" . ::uneval(\@text));
 	return @text if ! $s->{mv_all_chars}[0];
 	@text = map {quotemeta $_} @text;
-#::logDebug( "text=@text");
+#::logDebug( "escaped text=" . ::uneval(\@text));
     return @text;
 }
 
@@ -194,15 +214,15 @@ sub dump_coord {
 			$msg,
             $s->{mv_coordinate},
 			scalar @$specs,
-			::uneval($specs),
+			Vend::Util::uneval($specs),
 			scalar @{$s->{mv_search_field}},
-			::uneval($s->{mv_search_field}),
+			Vend::Util::uneval($s->{mv_search_field}),
 			scalar @{$s->{mv_column_op}},
-			::uneval($s->{mv_column_op}),
+			Vend::Util::uneval($s->{mv_column_op}),
 			scalar @{$s->{mv_numeric}},
-			::uneval($s->{mv_numeric}),
+			Vend::Util::uneval($s->{mv_numeric}),
 			scalar @{$s->{mv_negate}},
-			::uneval($s->{mv_negate}),
+			Vend::Util::uneval($s->{mv_negate}),
 			;
 }
 
@@ -222,8 +242,8 @@ sub spec_check {
 	my $all_chars = $s->{mv_all_chars}[0];
 
 	while ($i < @specs) {
-#::logDebug("i=$i specs=$#specs");
-		if($#specs and length($specs[$i]) == 0) { # should add a switch
+#::logDebug("i=$i specs=$#specs mv_min_string=$s->{mv_min_string}");
+		if($#specs and length($specs[$i]) == 0 and $s->{mv_min_string} != 0) { # should add a switch
 			if($s->{mv_coordinate}) {
 		        splice(@{$s->{mv_search_group}}, $i, 1);
 		        splice(@{$s->{mv_search_field}}, $i, 1);
@@ -238,14 +258,6 @@ sub spec_check {
 		    splice(@specs, $i, 1);
 		}
 		else {
-			if(length($specs[$i]) < $s->{mv_min_string}) {
-				my $msg = <<EOF;
-Search strings must be at least $s->{mv_min_string} characters.
-You had '$specs[$i]' as one of your search strings.
-EOF
-				$s->{matches} = -1;
-				return undef;
-			}
 			COLOP: {
 				last COLOP unless $s->{mv_coordinate};
 #::logDebug("i=$i, begin_string=$s->{mv_begin_string}[$i]");
@@ -278,6 +290,7 @@ EOF
 	}
 
 #::logDebug("regex_specs=" . ::uneval($s->{regex_specs}));
+#::logDebug("eq_specs_sql=" . ::uneval($s->{eq_specs_sql}));
 
 	if ( ! $s->{mv_exact_match} and ! $s->{mv_coordinate}) {
 		my $string = join ' ', @specs;
@@ -337,7 +350,7 @@ sub more_matches {
 	my $id = $s->{mv_more_id} || $s->{mv_session_id};
 	$id .= ".$s->{mv_cache_key}";
 	
-	my $file = Vend::Util::get_filename($id,undef,undef,$Vend::Cfg->{StaticScratch});
+	my $file = Vend::Util::get_filename($id);
 #::logDebug("more_matches: $id from $file");
 
 	my $obj;
@@ -362,12 +375,12 @@ sub more_matches {
 
 sub more_alpha {
 	my ($s, $out) = @_;
-	my ($sfpos, $letter, $sortkey, $last, @alphaspecs, $i);
+	my ($sfpos, @letters, @compare, $changed, @pos, $sortkey, $last, @alphaspecs, $i, $l);
 	my $alphachars = $s->{mv_more_alpha_chars} || 3;
 
 	# determine position of sort field within results
 	for ($i = 0; $i < @{$s->{mv_return_fields}}; $i++) {
-		last if $s->{mv_return_fields}->[$i] == $s->{mv_sort_field}->[0];
+		last if $s->{mv_return_fields}->[$i] eq $s->{mv_sort_field}->[0];
 	}
 	$sfpos = $i;
 	
@@ -377,9 +390,20 @@ sub more_alpha {
 	$last = 0;
 	for ($i = 0; $i < @$out; $i++) {
 		$sortkey = $out->[$i]->[$sfpos];
-		$letter = substr($sortkey,0,1);
+		@letters = split(//, $sortkey, $alphachars + 1);
+		pop(@letters);
+		$changed = 0;
 
-		if ($letter ne substr($alphaspecs[$last]->[0],0,1)) {
+		for ($l = 0; $l < $alphachars; $l++) {
+			if ($letters[$l] ne $compare[$l]) {				
+				$changed = 1;
+			}
+			next unless $changed;
+			$pos[$l] = $i;
+			$compare[$l] = $letters[$l];
+		}
+
+		if ($pos[0] == $i) {
 			# add record if first letter has changed
 			push (@alphaspecs, [$sortkey, 1, $i]);
 			# add last pointer to previous record
@@ -389,11 +413,11 @@ sub more_alpha {
 			# add record if match limit is exceeded and significant
 			# letters are different
 			for (my $c = 2; $c <= $alphachars; $c++) {
-				if (substr($sortkey,0,$c)
+				if (substr($sortkey, 0, $c)
 					ne substr($alphaspecs[$last]->[0],0,$c)) {
-					push (@alphaspecs, [$sortkey, $c, $i]);
+					push (@alphaspecs, [$sortkey, $c, $pos[$c - 1]]);
 					# add last pointer to previous record
-					push (@{$alphaspecs[$last++]}, $i - 1);
+					push (@{$alphaspecs[$last++]}, $pos[$c - 1] - 1);
 					last;
 				}
 			}
@@ -410,42 +434,122 @@ sub more_alpha {
 # Returns a field weeding function based on the search specification.
 # Input is the raw line and the delimiter, output is the fields
 # specified in the return_field specification
+# If we get a third parameter, $makeref, we need to build a reference
 sub get_return {
-	my($s, $final) = @_;
+	my($s, $final, $makeref) = @_;
 	my ($return_sub);
 
-	# We will pick out the return fields later if sorting
-	if(! $final and $s->{mv_sort_field}) {
-		return ( sub {@_}, 1);
-	}
+	if($makeref) {
+		# Avoid the hash key lookup, it is a closure
+		my $delim = $s->{mv_index_delim};
 
-	if(! $s->{mv_return_fields}) {
-		my $delim = $s->{mv_index_delim} || "\t";
-#::logDebug("code return. delim='$delim'");
-		$return_sub = sub {
-				$_[0] =~ s/$delim.*//s;
-				my $ary = [ $_[0] ];
-#::logDebug("ary is:" . ::uneval($ary));
-				return $ary;
-				};
+		# We will pick out the return fields later if sorting
+		# This returns
+		if( $s->{mv_sort_field} || $s->{mv_next_search}) {
+			return ( 
+				sub {
+					[ split /$delim/o, shift(@_) ]
+				},
+				1,
+			);
+		}
+		elsif($s->{mv_return_fields}) {
+			my @fields = @{$s->{mv_return_fields}};
+			$return_sub = sub {
+							return [ (split /$delim/o, shift(@_))[@fields] ]
+						};
+		}
+		else {
+			$return_sub = sub {
+							$_[0] =~ s/$delim.*//s;
+							return [ $_[0] ];
+					};
+		}
 	}
 	else {
-		my $delim = $s->{mv_index_delim};
-#::logDebug("rf[0]='$s->{mv_return_fields}[0]'");
-		my @fields = @{$s->{mv_return_fields}};
-#::logDebug("delim='$delim' fields='@fields'");
-		$return_sub = sub {
-			chomp($_[0]);
-			my $ary = [
-				(split /\Q$delim/o, $_[0])[@fields]
-				];
-#::logDebug("line is:$_[0]\nary is:" . ::uneval($ary));
-				return $ary;
-		};
+		# We will pick out the return fields later if sorting
+		# This returns
+		if(! $final and $s->{mv_sort_field} || $s->{mv_next_search}) {
+			return ( sub { [ @{ shift(@_) } ] }, 1);
+		}
+
+		if(! $s->{mv_return_fields}) {
+			$return_sub = sub {
+								return [ $_[0]->[0] ];
+						};
+		}
+		else {
+			my @fields = @{$s->{mv_return_fields}};
+			$return_sub = sub {
+				my $ref = [ @{$_[0]}[@fields] ];
+				return $ref;
+			};
+		}
 	}
 	return $return_sub;
 }
 
+my $TextQuery;
+
+BEGIN {
+	eval {
+		require Text::Query;
+		import Text::Query;
+		$TextQuery = 1;
+	};
+}
+
+sub create_text_query {
+	my ($s, $i, $string, $op) = @_;
+
+	if(! $TextQuery) {
+		die ::errmsg("No Text::Query module installed, cannot use op=%s", $op);
+	}
+
+	$s ||= {};
+	$op ||= 'sq';
+	my $q;
+#::logDebug("query creation called, op=$op");
+
+	my $cs	= ref ($s->{mv_case})
+			? $s->{mv_case}[$i]
+			: $s->{mv_case};
+	my $ac	= ref ($s->{mv_all_chars})
+			? $s->{mv_all_chars}[$i]
+			: $s->{mv_all_chars};
+	my $su	= ref ($s->{mv_substring_match})
+			? $s->{mv_substring_match}[$i]
+			: $s->{mv_substring_match};
+
+#::logDebug("query creation called, op=$op cs=$cs ac=$ac");
+	if($op eq 'aq') {
+		$q = new Text::Query($string,
+								-parse => 'Text::Query::ParseAdvanced',
+                               -solve => 'Text::Query::SolveAdvancedString',
+                               -build => 'Text::Query::BuildAdvancedString',
+				);
+	}
+	else {
+		$q = new Text::Query($string,
+								-parse => 'Text::Query::ParseSimple',
+                               -solve => 'Text::Query::SolveSimpleString',
+                               -build => 'Text::Query::BuildSimpleString',
+				);
+	}
+	$q->prepare($string,
+					-litspace => $s->{mv_exact_match},
+					-case => $cs,
+					-regexp => ! $ac,
+					-whole => ! $su,
+				);
+#::logDebug("query object called, is: " . ::uneval($q));
+	return sub {
+		my $str = shift;
+#::logDebug("query routine called string=$str");
+		$q->matchscalar($str);
+    };
+}
+
 my %numopmap  = (
 				'!=' => [' != '],
 				'!~' => [' !~ m{', '}'],
@@ -491,6 +595,8 @@ my %stropmap  = (
 				'em' => [' =~ m{^', '$}i'],
 				'rm' => [' =~ m{', '}i'],
 				'rn' => [' !~ m{', '}i'],
+				'aq' => [\&create_text_query, 'aq'],
+				'tq' => [\&create_text_query, 'tq'],
 				'like' => [' =~ m{LIKE', '}i'],
 				'LIKE' => [' =~ m{LIKE', '}i'],
 );
@@ -503,10 +609,17 @@ sub map_ops {
 	my $op;
 	for($i = 0; $i < $count; $i++) {
 		next unless $c->[$i];
-		$c->[$i] =~ tr/ 	//;
+		$c->[$i] =~ tr/ \t//;
+		my $o = $c->[$i];
 		$c->[$i] = $s->{mv_numeric}[$i]
-				? $numopmap{$c->[$i]}
-				: $stropmap{$c->[$i]};
+				? $numopmap{$o}
+				: $stropmap{$o};
+		if(! $c->[$i]) {
+			my $r;
+			$c->[$i] = [$r, $o], next
+				if  $r = $Global::CodeDef->{SearchOp}
+				and $r = $r->{Routine}{$o};
+		}
 	}
 	@{$s->{mv_column_op}};
 }
@@ -535,12 +648,20 @@ sub create_field_hash {
 # Returns a screening function based on the search specification.
 # The $f is a reference to previously created search function which does
 # a pattern match on the line.
+# Makeref says we have to build a reference from the supplied text line
 sub get_limit {
-	my($s, $f) = @_;
+	my($s, $f, $makeref) = @_;
 	my $limit_sub;
 	my $range_code = '';
 	my $rd = $s->{mv_record_delim} || '\n';
-	my $code       = "sub {\nmy \$line = shift; chomp \$line; \$line =~ tr/$rd//d;\n";
+	my $code;
+	if($makeref) {
+		$code       = "sub {\nmy \$line = [split /$s->{mv_index_delim}/, shift(\@_)];\n";
+	}
+	else {
+		$code       = "sub {\nmy \$line = shift;\n";
+	}
+	$code .= "my \@fields = \@\$line;\n";
 	my $join_key;
 	$join_key = defined $s->{mv_return_fields} ? $s->{mv_return_fields}[0] : 0;
 	$join_key = 0 if $join_key eq '*';
@@ -574,15 +695,14 @@ sub get_limit {
 		$s->{mv_field_hash} = create_field_hash($s) 
 			unless $s->{mv_field_hash};
 		$code .= <<EOF;
-my \$key = (split m{$s->{mv_index_delim}}, \$line)[$join_key];
+my \$key = \$line->[$join_key];
 EOF
 		for(@join_fields) {
 			my ($table, $col) = split /:+/, $_, 2;
 			if($table) {
 				$wild_card = 0;
 				$code .= <<EOF;
-\$line .= qq{$s->{mv_index_delim}} .
-		  Vend::Data::database_field('$table', \$key, '$col');
+push \@fields, Vend::Data::database_field('$table', \$key, '$col');
 EOF
 			}
 			elsif ($col =~ tr/:/,/) {
@@ -594,16 +714,15 @@ EOF
 				$wild_card = 1;
 				$col =~ s/[^\d,.]//g;
 			$code .= <<EOF;
-my \$addl = join " ", (split m{\Q$s->{mv_index_delim}\E}, \$line)[$col];
-\$line .= qq{$s->{mv_index_delim}} . \$addl;
+my \$addl = join " ", \@fields[$col];
+push \@fields, \$addl;
 EOF
 			}
 			else {
 				$wild_card = 1;
 				$code .= <<EOF;
-my \$addl = \$line;
-\$addl =~ tr/$s->{mv_index_delim}/ /;
-\$line .= qq{$s->{mv_index_delim}} . \$addl;
+my \$addl = join " ", \@fields;
+push \@fields, \$addl;
 EOF
 			}
 		}
@@ -613,7 +732,7 @@ EOF
 
 	if ( ref $s->{mv_range_look} )  {
 		$range_code = <<EOF;
-return $joiner \$s->range_check(qq{$s->{mv_index_delim}},\$line);
+return $joiner \$s->range_check(\$line);
 EOF
 	}
 	if ( $s->{mv_coordinate} ) {
@@ -624,7 +743,6 @@ EOF
 		 }
 		 my $callchar = $fields =~ /,/ ? '@' : '$';
 		 $code .= <<EOF;
-	my \@fields = split /\\Q$s->{mv_index_delim}/, \$line;
 	\@fields = ${callchar}fields[$fields];
 EOF
 		$code .= <<EOF if $Global::DebugFile and $CGI::values{debug};
@@ -647,13 +765,35 @@ EOF
 #::logDebug("Begin=" . join ",", @begin);
 #::logDebug("Group=" . join ",", @group);
 #::logDebug("Ors=" . join ",", @{$s->{mv_orsearch}});
+#::logDebug("Field count=$field_count");
 		my @code;
 		my $candidate = '';
 		my ($i, $start, $term, $like);
 		for($i = 0; $i < $field_count; $i++) {
 			undef $candidate, undef $f 
 				if $begin[$i] or $s->{mv_orsearch}[$i];
-			if($ops[$i]) {
+			my $subfrag;
+			if(! $ops[$i]) {
+				$start = '=~ m{';
+				$start .=  '^' if $begin[$i];
+				if($bounds[$i]) {
+					$term = '}';
+				}
+				else {
+					$term = '\b}';
+					$start .= '\b' unless $begin[$i];
+				}
+				$term .= 'i' unless $cases[$i];
+				$candidate = 1 if defined $candidate and ! $begin[$i];
+			}
+			elsif(ref($ops[$i][0]) eq 'CODE') {
+					undef $f; undef $candidate;
+					my $o = shift(@{$ops[$i]});
+					$s->{search_routines} ||= [];
+					$s->{search_routines}[$i] = $o->($s, $i, $specs[$i], @{$ops[$i]});
+					$subfrag = qq{ \$s->{search_routines}[$i]->(\$fields[$i])};
+			}
+			else {
 				$ops[$i][0] =~ s/m\{$/m{^/ if $begin[$i];
 				! $bounds[$i] 
 					and $ops[$i][0] =~ s/=~\s+m\{$/=~ m{\\b/
@@ -667,19 +807,7 @@ EOF
 					and $candidate = 1;
 #::logDebug("Candidate now=$candidate");
 			}
-			else {
-				$start = '=~ m{';
-				$start .=  '^' if $begin[$i];
-				if($bounds[$i]) {
-					$term = '}';
-				}
-				else {
-					$term = '\b}';
-					$start .= '\b' unless $begin[$i];
-				}
-				$term .= 'i' unless $cases[$i];
-				$candidate = 1 if defined $candidate and ! $begin[$i];
-			}
+			
 			if ($start =~ s/LIKE$//) {
 				$specs[$i] =~ s/^(%)?([^%]*)(%)?$/$2/;
 				# Substitute if only one present
@@ -691,20 +819,27 @@ EOF
 								: '^' . $specs[$i];
 					$like = 1;
 				}
-			 }
-			 if ($i >= $k + $field_count) {
-				 undef $candidate if ! $wild_card;
-#::logDebug("triggered wild_card: $wild_card");
-				 $wild_card = 0;
-			 }
-			 if(defined $candidate and ! $like) {
-				undef $f if $candidate;
-			 	$f = "sub { return 1 if $negates[$i]\$_ $start$specs[$i]$term ; return 0}"
-					if ! defined $f and $start =~ m'=~';
-				undef $candidate if $candidate;
-			 }
-			 my $grp = $group[$i] || 0;
-			 my $frag = qq{$negates[$i]\$fields[$i] $start$specs[$i]$term};
+			}
+			if ($i >= $k + $field_count) {
+			    undef $candidate if ! $wild_card;
+#::logDebug(triggered wild_card: $wild_card");
+			    $wild_card = 0;
+			}
+			if(defined $candidate and ! $like) {
+			   undef $f if $candidate;
+				$f = "sub { return 1 if $negates[$i]\$_ $start$specs[$i]$term ; return 0}"
+			   	if ! defined $f and $start =~ m'=~';
+			   undef $candidate if $candidate;
+			}
+			my $grp = $group[$i] || 0;
+			my $frag;
+			if($subfrag) {
+			   $frag = $subfrag;
+			}
+			else {
+			    $frag = qq{$negates[$i]\$fields[$i] $start$specs[$i]$term};
+			}
+#::logDebug("Code fragment is q!$frag!");
 			 unless ($code[$grp]) {
 				 $code[$grp] = [ $frag ];
 			 }
@@ -715,13 +850,15 @@ EOF
 		}
 #::logDebug("coderef=" . ::uneval_it(\@code));
 
+		undef $f if $s->{mv_search_relate} =~ s/\bor\b/or/ig;
+		undef $f unless $s->{regex_specs} or $s->{eq_specs};
 		DOLIMIT: {
 #::logDebug(::uneval_it({%$s}));
 #::logDebug("do_limit.");
-			undef $f, last DOLIMIT unless $s->{regex_specs};
 			last DOLIMIT if $f;
 #::logDebug("do_limit past f.");
 			last DOLIMIT if $s->{mv_small_data};
+			last DOLIMIT if $s->{eq_specs_sql};
 			last DOLIMIT if (grep $_, @{$s->{mv_orsearch}});
 			last DOLIMIT if defined $s->{mv_search_relate}
 							&& $s->{mv_search_relate} =~ s/\bor\b/or/i;
@@ -777,7 +914,7 @@ EOF
 	}
 	elsif ( @{$s->{mv_search_field}} )  {
 		if(! $s->{mv_begin_string}[0]) {
-#::logDebug("Not begin, sub=f");
+#::logDebug("Not begin, sub=$f");
 			$sub = $f;
 		}
 		elsif (! $s->{mv_orsearch}[0] ) {
@@ -800,10 +937,8 @@ EOF
 		}
 		 $code .= $range_code;
 		 $code .= <<EOF;
-	my \@fields = (split /\\Q$s->{mv_index_delim}/, \$line)[$fields];
-	my \$field = join q{$s->{mv_index_delim}}, \@fields;
-	\$_ = \$field;
-	return(\$_ = \$line) if &\$sub();
+	local(\$_) = join q{$s->{mv_index_delim}}, \@\$line[$fields];
+	return(1) if &\$sub();
 	return undef;
 }
 EOF
@@ -831,8 +966,8 @@ EOF
 # Check to see if the fields specified in the range_look array
 # meet the criteria
 sub range_check {
-	my($s,$index_delim,$line) = @_;
-	my @fields = (split /\Q$index_delim/, $line)[@{$s->{mv_range_look}}];
+	my($s,$line) = @_;
+	my @fields = @$line[@{$s->{mv_range_look}}];
 	my $i = 0;
 	for(@fields) {
 		no strict 'refs';
@@ -954,7 +1089,7 @@ sub save_context {
 	for (@save) {
 		$return->{$_} = $s->{$_};
 	}
-	::uneval_fast($return);
+	Vend::Util::uneval_fast($return);
 }
 
 sub dump_options {
@@ -964,15 +1099,16 @@ sub dump_options {
 		$Data::Dumper::Indent = 3;
 		$Data::Dumper::Terse = 1;
 	}
-	return ::uneval($s);
+	return Vend::Util::uneval($s);
 }
 
 sub search_error {
-	my ($s, $msg) = @_;
+	my ($s, $msg, @args) = @_;
 	$s->{mv_search_error} = [] if ! $s->{mv_search_error};
+	$msg = ::errmsg($msg, @args);
 	push @{$s->{mv_search_error}}, $msg;
 	$s->{matches} = -1;
-	::logError ("search error: $msg");
+	::logError ("search error: %s", $msg);
 	return undef;
 }
 
@@ -983,7 +1119,7 @@ sub save_more {
 	delete $s->{dbref} if defined $s->{dbref};
 	my $id = $s->{mv_more_id} || $Vend::SessionID;
 	$id .= ".$s->{mv_cache_key}";
-	if ($s->{matches} > $s->{mv_matchlimit}) {
+	if ($s->{matches} > $s->{mv_matchlimit} and $s->{mv_matchlimit} > 0) {
 		$s->{overflow} = 1;
 		$s->{mv_next_pointer} = $s->{mv_matchlimit};
 	}
@@ -998,10 +1134,11 @@ sub save_more {
 		more_alpha($s,$out);
 	}
 	
-	$file = Vend::Util::get_filename($id,undef,undef,$Vend::Cfg->{StaticScratch}); 
+	$file = Vend::Util::get_filename($id); 
 #::logDebug("save_more: $id to $file.");
 	my $new = { %$s };
 	$new->{mv_results} = $out;
+	delete $new->{search_routines};
 #::logDebug("save_more:object:" . ::uneval($new));
 	eval {
 		Vend::Util::uneval_file($new, $file);
@@ -1030,7 +1167,7 @@ use vars qw/ %Sort_field /;
 sub sort_search_return {
     my ($s, $target) = @_;
 
-	@Flds	= @{$s->{mv_sort_field}};
+	@Flds	= @{$s->{mv_sort_field} || []};
 	for(@Flds) {
 		next if /^\d+$/;
 		$_ = $s->{field_hash}{$_}
@@ -1054,18 +1191,13 @@ my %Sorter = (
 	rn	=> sub { $_[1] <=> $_[0]			},
 );
 
-	my $last = 'none';
 	my $i;
 	my $max = 0;
 	for($i = 0; $i < @Flds; $i++) {
 		$max = $Flds[$i] if $Flds[$i] > $max;
-		if (! $Opts[$i]) {
-			$Opts[$i] = $last;
-			next;
-		}
+		$Opts[$i] = 'none', next unless $Opts[$i];
 		$Opts[$i] = lc $Opts[$i];
 		$Opts[$i] = 'none' unless defined $Sort_field{$Opts[$i]};
-		$last = $Opts[$i];
 	}
 #::logDebug("sort_search_return: flds='@Flds' opts='@Opts'");
 
@@ -1074,8 +1206,8 @@ my %Sorter = (
 	my $delim = quotemeta $s->{mv_index_delim};
 	my $code = <<EOF;
 sub {
-	my \@a = (split /$delim/, \$a, $max)[$f_string];
-	my \@b = (split /$delim/, \$b, $max)[$f_string];
+	my \@a = \@{\$a}[$f_string];
+	my \@b = \@{\$b}[$f_string];
 	my \$r;
 EOF
 #::logDebug("No define of Sort_field") if ! defined $Sort_field{'none'};
@@ -1113,7 +1245,7 @@ eval {
 
 	# Prime sort routine
 	use locale;
-	sort { $routine } ('30','31') or 1;
+	local($^W);
 
 	@$target = sort { &$routine } @$target;
 #::logDebug("target is $target: " . Vend::Util::uneval_it($target));
