@@ -1,11 +1,12 @@
 # Vend::Server - Listen for Interchange CGI requests as a background server
 #
-# $Id: Server.pm,v 2.0.2.1 2002/01/24 05:07:01 jon Exp $
+# $Id: Server.pm,v 2.50 2004/03/07 06:59:50 mheins Exp $
 #
-# Copyright (C) 1996-2002 Red Hat, Inc. <interchange@redhat.com>
+# Copyright (C) 2002-2003 Interchange Development Group
+# Copyright (C) 1996-2002 Red Hat, Inc.
 #
 # This program was originally based on Vend 0.2 and 0.3
-# Copyright 1995 by Andrew M. Wilcox <awilcox@world.std.com>
+# Copyright 1995 by Andrew M. Wilcox <amw@wilcoxsolutions.com>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -25,7 +26,7 @@
 package Vend::Server;
 
 use vars qw($VERSION);
-$VERSION = substr(q$Revision: 2.0.2.1 $, 10);
+$VERSION = substr(q$Revision: 2.50 $, 10);
 
 use POSIX qw(setsid strftime);
 use Vend::Util;
@@ -36,18 +37,10 @@ use Socket;
 use Symbol;
 use strict;
 
+my $ppidsub = sub { return getppid };
+
 sub new {
     my ($class, $fh, $env, $entity) = @_;
-	if(@Global::argv > 1) {
-		(
-			$CGI::script_name,
-			$CGI::values{mv_session_id}, 
-			$CGI::query_string
-		) = @Global::argv;
-		map_cgi();
-		$Global::FastMode = 1;
-		return bless { fh => $fh }, $class;
-	}
     populate($env);
     my $http = {
 					fh => $fh,
@@ -65,29 +58,37 @@ sub new {
     bless $http, $class;
 }
 
-my @Map =
-    (
-     'authorization' => 'AUTHORIZATION',
-     'content_length' => 'CONTENT_LENGTH',
-     'content_type' => 'CONTENT_TYPE',
-     'content_encoding' => 'HTTP_CONTENT_ENCODING',
-     'cookie' => 'HTTP_COOKIE',
-     'http_host' => 'HTTP_HOST',
-     'path_info' => 'PATH_INFO',
-     'pragma' => 'HTTP_PRAGMA',
-     'query_string' => 'QUERY_STRING',
-     'referer' => 'HTTP_REFERER',
-     'remote_addr' => 'REMOTE_ADDR',
-     'remote_host' => 'REMOTE_HOST',
-     'remote_user' => 'REMOTE_USER',
-     'request_method', => 'REQUEST_METHOD',
-     'script_name' => 'SCRIPT_NAME',
-     'secure' => 'HTTPS',
-     'server_name' => 'SERVER_NAME',
-     'server_host' => 'HTTP_HOST',
-     'server_port' => 'SERVER_PORT',
-     'useragent' => 'HTTP_USER_AGENT',
-);
+my @Map = qw/
+    authorization      AUTHORIZATION
+    content_length     CONTENT_LENGTH
+    content_type       CONTENT_TYPE
+    content_encoding   HTTP_CONTENT_ENCODING
+    cookie             HTTP_COOKIE
+    http_host          HTTP_HOST
+    path_info          PATH_INFO
+    pragma             HTTP_PRAGMA
+    query_string       QUERY_STRING
+    referer            HTTP_REFERER
+    remote_addr        REMOTE_ADDR
+    remote_host        REMOTE_HOST
+    remote_user        REMOTE_USER
+    request_method     REQUEST_METHOD
+    request_uri        REQUEST_URI
+    script_name        SCRIPT_NAME
+    secure             HTTPS
+    server_name        SERVER_NAME
+    server_host        HTTP_HOST
+    server_port        SERVER_PORT
+    useragent          HTTP_USER_AGENT
+/;
+
+my @RedirMap = qw/
+    path_info          REDIRECT_URL
+    query_string       REDIRECT_QUERY_STRING
+    error_notes        REDIRECT_ERROR_NOTES
+    redirect_status    REDIRECT_STATUS
+    request_method     REDIRECT_REQUEST_METHOD
+/;
 
 ### This is to account for some bad Socket.pm implementations
 ### which don't set SOMAXCONN, I think SCO is the big one
@@ -119,6 +120,30 @@ sub populate {
         ${"CGI::$field"} = $cgivar->{$cgi} if defined $cgivar->{$cgi};
 #::logDebug("CGI::$field=" . ${"CGI::$field"});
     }
+
+	# try to get originating host's IP address if request was
+	# forwarded through a trusted proxy
+	my $ip;
+	if ($Global::TrustProxy
+		and ($CGI::remote_addr =~ $Global::TrustProxy
+			or $CGI::remote_host =~ $Global::TrustProxy)
+		and $ip = $cgivar->{HTTP_X_FORWARDED_FOR}) {
+		# trim off intermediate proxies in comma-separated list
+		$ip =~ s/,.*//;
+		$ip =~ s/^\s+//; $ip =~ s/\s+$//;
+		if ($ip =~ /^\d\d?\d?\.\d\d?\d?\.\d\d?\d?\.\d\d?\d?$/) {
+			$CGI::remote_addr = $ip;
+			undef $CGI::remote_host;
+		}
+		else {
+			::logGlobal(
+				{ level => 'info' },
+				"Unknown HTTP_X_FORWARDED_FOR header set from trusted proxy %s: '%s'",
+				$CGI::remote_addr,
+				$cgivar->{HTTP_X_FORWARDED_FOR},
+			);
+		}
+	}
 }
 
 sub log_http_data {
@@ -149,8 +174,8 @@ sub log_http_data {
 }
 
 sub map_misc_cgi {
-	$CGI::user = $CGI::remote_user;
 	$CGI::host = $CGI::remote_host || $CGI::remote_addr;
+	$CGI::user = $CGI::remote_user;
 
 	$CGI::script_path = $CGI::script_name;
 	$CGI::script_name = $CGI::server_host . $CGI::script_path
@@ -174,7 +199,7 @@ sub map_cgi {
 			my $content_type = $msg =~ /<html/i ? 'text/html' : 'text/plain';
 			my $len = length($msg);
 			$Vend::StatusLine = <<EOF;
-Status: 404
+Status: 404 Not found
 Content-Type: $content_type
 Content-Length: $len
 EOF
@@ -182,21 +207,77 @@ EOF
 			die($msg);
 		};
 
+	if($Global::DNSBL) {
+		my @quads = split /\./, $CGI::remote_addr;
+		my $intro = join ".", reverse(@quads), '';
+		my $blocked;
+		for(@{$Global::DNSBL}) {
+			my $addr = gethostbyname($intro . $_)
+			 or next;
+			$blocked = 1;
+		}
+		if($blocked) {
+			my $msg = ::get_locale_message( 403, "Listed on avoid list.",);
+			my $content_type = $msg =~ /<html/i ? 'text/html' : 'text/plain';
+			my $len = length($msg);
+			$Vend::StatusLine = <<EOF;
+Status: 403 Forbidden
+Content-Type: $content_type
+Content-Length: $len
+EOF
+			respond('', \$msg);
+			die($msg);
+		}
+	}
+
 	($::IV, $::VN, $::SV) = $g->{VarName}
 			? ($g->{IV}, $g->{VN}, $g->{IgnoreMultiple})
 			: ($Global::IV, $Global::VN, $Global::IgnoreMultiple);
 
+	# Vend::ModPerl has already handled GET/POST parsing
+	return if $Global::mod_perl;
+
 #::logDebug("CGI::query_string=" . $CGI::query_string);
 #::logDebug("entity=" . ${$h->{entity}});
+
 	if ("\U$CGI::request_method" eq 'POST') {
 		parse_post(\$CGI::query_string)
 			if $Global::TolerateGet;
 		parse_post($h->{entity});
 	}
+	elsif ("\U$CGI::request_method" eq 'PUT') {
+#::logDebug("Put operation.");
+		parse_post(\$CGI::query_string);
+		$CGI::put_ref = $h->{entity};
+#::logDebug("Put contents: $$CGI::put_ref");
+		$$CGI::put_ref =~ s/^\s*--+\s+begin\s+content\s+--+\r?\n//i;
+		$$CGI::put_ref =~ s/^\r?\n--+\s+end\s+content\s+--+\s*$//i;
+	}
 	else {
 		 parse_post(\$CGI::query_string);
 	}
 
+	return if $CGI::values{mv_tmp_session};
+
+#::logDebug("Check robot UA=$Global::RobotUA IP=$Global::RobotIP");
+	if ($Global::RobotUA and $CGI::useragent =~ $Global::RobotUA) {
+#::logDebug("It is a robot by UA!");
+		$CGI::values{mv_tmp_session} = 1;
+	}
+	elsif ($Global::RobotIP and $CGI::remote_addr =~ $Global::RobotIP) {
+#::logDebug("It is a robot by IP!");
+		$CGI::values{mv_tmp_session} = 1;
+	}
+	elsif ($Global::HostnameLookups && $Global::RobotHost) {
+		if (!$CGI::remote_host && $CGI::remote_addr) {
+			$CGI::remote_host = gethostbyaddr(Socket::inet_aton($CGI::remote_addr),Socket::AF_INET);
+			$CGI::host = $CGI::remote_host || $CGI::remote_addr;
+		}
+		if ($CGI::remote_host && $CGI::remote_host =~ $Global::RobotHost) {
+#::logDebug("It is a robot by host!");
+			$CGI::values{mv_tmp_session} = 1;
+		}
+	}
 }
 
 # This is called by parse_multipart
@@ -217,7 +298,6 @@ sub store_cgi_kv {
 sub parse_post {
 	my $sref = shift;
 	my(@pairs, $pair, $key, $value);
-	undef %CGI::values;
 	return unless length $$sref;
 	if ($CGI::content_type =~ /^multipart/i) {
 		return parse_multipart($sref) if $CGI::useragent !~ /MSIE\s+5/i;
@@ -226,7 +306,7 @@ sub parse_post {
 		# even though it's sent as non-multipart data
 		# Contributed by Bill Randle
 		my ($boundary) = $CGI::content_type =~ /boundary=\"?([^\";]+)\"?/;
-		$boundary = "--$boundary";
+		$boundary = '--' . quotemeta $boundary;
 		return parse_multipart($sref) if $$sref =~ /^\s*$boundary\s+/;
 	}
 	@pairs = split($Global::UrlSplittor, $$sref);
@@ -244,7 +324,7 @@ sub parse_post {
 	my $redo;
   CGIVAL: {
   	# This loop semi-duplicated in store_cgi_kv
-	foreach $pair (@pairs) {
+	foreach $pair (grep length, @pairs) {
 		($key, $value) = ($pair =~ m/([^=]+)=(.*)/)
 			or do {
 				if ($Global::TolerateGet) {
@@ -296,6 +376,7 @@ sub parse_post {
 sub parse_multipart {
 	my $sref = shift;
 	my ($boundary) = $CGI::content_type =~ /boundary=\"?([^\";]+)\"?/;
+	$boundary = quotemeta $boundary;
 #::logDebug("got to multipart");
 	# Stolen from CGI.pm, thanks Lincoln
 	$boundary = "--$boundary"
@@ -350,6 +431,7 @@ sub parse_multipart {
 sub create_cookie {
 	my($domain,$path) = @_;
 	my  $out;
+	return '' if $Vend::tmp_session;
 	my @jar;
 	push @jar, [
 				($::Instance->{CookieName} || 'MV_SESSION_ID'),
@@ -357,7 +439,6 @@ sub create_cookie {
 				$Vend::Expire || undef,
 			]
 		unless $Vend::CookieID;
-	push @jar, ['MV_STATIC', 1] if $Vend::Cfg->{Static};
 	push @jar, @{$::Instance->{Cookies}}
 		if defined $::Instance->{Cookies};
 	$out = '';
@@ -395,7 +476,7 @@ sub canon_status {
 	$_ = shift;
 	s:^\s+::;
 	s:\s+$::;
-	s:\s*\n:\r\n:mg;
+	s:\s*\n\s*:\r\n:g;
 	return "$_\r\n";
 }
 
@@ -404,19 +485,22 @@ sub respond {
     my ($s, $body) = @_;
 #show_times("begin response send") if $Global::ShowTimes;
 	my $status;
+	return if $Vend::Sent;
 	if($Vend::StatusLine) {
 		$status = $Vend::StatusLine =~ /(?:^|\n)Status:\s+(.*)/i
 				? "$1"
 				: "200 OK";
 	}
 
+	if($CGI::redirect_status and ! $Vend::StatusLine) {
+		$status = "200 OK";
+		$Vend::StatusLine = "Status: 200 OK\nContent-Type: text/html";
+	}
+
 	$$body =~ s/^\s+//
-		if ! $Vend::ResponseMade and $Vend::Cfg->{Pragma}{strip_white};
+		if ! $Vend::ResponseMade and $::Pragma->{strip_white};
 
 	if(! $s and $Vend::StatusLine) {
-		$Vend::StatusLine = "HTTP/1.0 $status\r\n$Vend::StatusLine"
-			if defined $Vend::InternalHTTP
-				and $Vend::StatusLine !~ m{^HTTP/};
 		$Vend::StatusLine .= ($Vend::StatusLine =~ /^Content-Type:/im)
 							? '' : "\r\nContent-Type: text/html\r\n";
 # TRACK
@@ -443,18 +527,35 @@ sub respond {
 #	select($oldfh);
 # END SUNOSDIGITAL
 
+	my $rfh = $s->{rfh};
+	if($Vend::write_redirect and ! $rfh) {
+		$rfh = gensym();
+		my $fn = $Vend::Cfg->{RedirectCache} . $CGI::path_info;
+		my $save = umask(022);
+		open $rfh, "> $fn"
+			or do {
+				::logError("Unable to write redirected page %s: %s", $fn, $!);
+				undef $Vend::write_redirect;
+				undef $rfh;
+			};
+		$s->{rfh} = $rfh;
+		umask $save;
+	}
+
 	if($Vend::ResponseMade || $CGI::values{mv_no_header} ) {
 		print $fh $$body;
+		print $rfh $$body if $rfh;
 #show_times("end response send") if $Global::ShowTimes;
 		return 1;
 	}
 
-	if (defined $Vend::InternalHTTP or defined $ENV{MOD_PERL} or $CGI::script_name =~ m:/nph-[^/]+$:) {
+	if (defined $ENV{MOD_PERL} or $CGI::script_name =~ m:/nph-[^/]+$:) {
 # TRACK
 		my $save = select $fh;
 		$| = 1;
 		select $save;
-        $Vend::StatusLine .= "X-Track: " . $Vend::Track->header() . "\r\n";
+        $Vend::StatusLine .= "X-Track: " . $Vend::Track->header() . "\r\n"
+			if $Vend::Track;
 # END TRACK                            
         $Vend::StatusLine .= "Pragma: no-cache\r\n"
 			if delete $::Scratch->{mv_no_cache};
@@ -469,7 +570,9 @@ sub respond {
 		else { print $fh "HTTP/1.0 $status\r\n"; }
 	}
 
-	if ( (	! $Vend::CookieID && ! $::Instance->{CookiesSet}
+	if ( ! $Vend::tmp_session
+		and (
+			! $Vend::CookieID && ! $::Instance->{CookiesSet}
 			or defined $Vend::Expire
 			or defined $::Instance->{Cookies}
 		  )
@@ -479,19 +582,19 @@ sub respond {
 
 		my @domains;
 		@domains = ('');
+		my @paths;
+		@paths = ('/');
+
 		if ($Vend::Cfg->{CookieDomain}) {
 			@domains = split /\s+/, $Vend::Cfg->{CookieDomain};
 		}
-
-		my @paths;
-		@paths = ('/');
-		if($Global::Mall) {
+		elsif($Global::Mall) {
 			my $ref = $Global::Catalog{$Vend::Cat};
 			@paths = ($ref->{script});
 			push (@paths, @{$ref->{alias}}) if defined $ref->{alias};
 			if ($Global::FullUrl) {
 				# remove domain from script
-				for (@paths) { s:^[^/]+/:/: ; }
+				for (@paths) { s:^[^/]+/:/: or $_ = '/'; }
 			}
 		}
 
@@ -519,6 +622,7 @@ sub respond {
 
     print $fh "\r\n";
     print $fh $$body;
+	print $rfh $$body if $rfh;
 #show_times("end response send") if $Global::ShowTimes;
     $Vend::ResponseMade = 1;
 }
@@ -556,9 +660,7 @@ sub _string {
 
 my $HTTP_enabled;
 my $Remote_addr;
-my $Remote_host;
 my %CGImap;
-my %CGIspecial;
 
 BEGIN {
 	eval {
@@ -597,10 +699,9 @@ BEGIN {
 
 					/
 		);
-		%CGIspecial = ();
 	};
-										 
-}                                    
+
+}
 
 sub http_log_msg {
 	my($status, $env, $request) = @_;
@@ -669,7 +770,9 @@ sub http_soap {
 
 	(undef, $Remote_addr) =
 				sockaddr_in(getpeername($fh));
-	$$env{REMOTE_HOST} = gethostbyaddr($Remote_addr, AF_INET);
+	if ($Global::HostnameLookups) {
+		$$env{REMOTE_HOST} = gethostbyaddr($Remote_addr, AF_INET);
+	}
 	$Remote_addr = inet_ntoa($Remote_addr);
 
 	$$env{REMOTE_ADDR} = $Remote_addr;
@@ -711,127 +814,6 @@ sub http_soap {
 	return $ref;
 }
 
-sub http_server {
-	my($status_line, $in, $argv, $env, $entity) = @_;
-
-	die "Need URI::URL for this functionality.\n"
-		unless defined $HTTP_enabled;
-
-	$Vend::InternalHTTP = 1;
-	my ($header, $request, $block);
-	my $waiting = 0;
-	($$env{REQUEST_METHOD},$request) = split /\s+/, $status_line;
-	for(;;) {
-        $block = _find(\$in, "\n");
-#::logDebug("read: $block");
-		$block =~ s/\s+$//;
-		if($block eq '') {
-			last;
-		}
-		if ( $block =~ s/^([^:]+):\s*//) {
-			$header = lc $1;
-			if(defined $CGImap{$header}) {
-				$$env{$CGImap{$header}} = $block;
-			}
-			elsif(defined $CGIspecial{$header}) {
-				&{$CGIspecial{$header}}($env, $block);
-			}
-			# else { throw_away() }
-			next;
-		}
-		else {
-			die "HTTP protocol error on '$block':\n$in";
-		}
-		last;
-	}
-
-	if ($$env{CONTENT_LENGTH}) {
-		_read(\$in) while length($in) < $$env{CONTENT_LENGTH};
-	}
-	$in =~ s/\s+$//;
-	$$entity = $in;
-
-#::logDebug("exiting loop");
-	my $url = new URI::URL $request;
-	@{$argv} = $url->keywords();
-
-	(undef, $Remote_addr) =
-				sockaddr_in(getpeername(MESSAGE));
-	$$env{REMOTE_HOST} = gethostbyaddr($Remote_addr, AF_INET);
-	$Remote_addr = inet_ntoa($Remote_addr);
-
-	$$env{QUERY_STRING} = $url->equery();
-	$$env{REMOTE_ADDR} = $Remote_addr;
-
-	my (@path) = $url->path_components();
-	my $path = $url->path();
-	my $doc;
-	my $status = 200;
-
-	shift(@path);
-	my $catname = shift(@path);
-
-	if ($Global::TcpMap->{$Global::TcpPort} =~ /^\w+/) {
-		$catname = $Global::TcpMap->{$Global::TcpPort};
-	}
-	my $cat = "/$catname";
-
-	if($Global::Selector{$cat} and $Global::AllowGlobal->{$cat}) {
-		if ($$env{AUTHORIZATION}) {
-			$$env{REMOTE_USER} =
-					Vend::Util::check_authorization( delete $$env{AUTHORIZATION} );
-		}
-		if (! $$env{REMOTE_USER}) {
-			$Vend::StatusLine = <<EOF;
-HTTP/1.0 401 Unauthorized
-WWW-Authenticate: Basic realm="Interchange Admin"
-EOF
-			$doc = "Requires correct username and password.\n";
-			$path = '';
-		}
-	}
-
-	if($Global::Selector{$cat} || $Global::SelectorAlias{$cat}) {
-		$$env{SCRIPT_NAME} = $cat;
-		$$env{PATH_INFO} = join "/", '', @path;
-	}
-	elsif(-f "$Global::VendRoot/doc$path") {
-		$Vend::StatusLine = "HTTP/1.0 200 OK";
-		$doc = readfile("$Global::VendRoot/doc$path");
-	}
-	else {
-		$status = 404;
-		$Vend::StatusLine = "HTTP/1.0 404 Not found";
-		$doc = "$path not a Interchange catalog or help file.\n";
-	}
-
-	if($$env{REQUEST_METHOD} eq 'HEAD') {
-		$Vend::StatusLine = "HTTP/1.0 200 OK\nLast-modified: "
-			. Vend::Util::logtime;
-		$doc = '';
-	}
-
-	logData("$Global::VendRoot/etc/access_log",
-			http_log_msg(
-						$status,
-						$env,
-						($$env{REQUEST_METHOD} .  " " .  $request),
-						)
-		);
-
-	if (defined $doc) {
-		my $type = Vend::Util::mime_type($path);
-		$Vend::StatusLine = '' unless defined $Vend::StatusLine;
-		$Vend::StatusLine .= "\r\nContent-type: $type";
-		respond(
-					'',
-					\$doc,
-				);
-		return;
-	}
-	return 1;
-}
-
 sub read_cgi_data {
     my ($argv, $env, $entity) = @_;
     my ($in, $block, $n, $i, $e, $key, $value);
@@ -853,9 +835,6 @@ sub read_cgi_data {
 		elsif ($block =~ m/^entity$/) {
             $$entity = _string(\$in);
 		}
-        elsif ($block =~ m/^[GPH]/) {
-           	return http_server($block, $in, @_);
-        }
 		elsif (($n) = ($block =~ m/^arg (\d+)$/)) {
             $#$argv = $n - 1;
             foreach $i (0 .. $n - 1) {
@@ -866,14 +845,6 @@ sub read_cgi_data {
 			die "Unrecognized block: $block\n";
         }
     }
-	if($Vend::OnlyInternalHTTP) {
-		my $msg = ::errmsg(
-						"attempt to connect from unauthorized host '%s'",
-						$Vend::OnlyInternalHTTP,
-					);
-		::logGlobal({ level => 'alert' }, $msg);
-		die "$msg\n";
-	}
 	return 1;
 }
 
@@ -902,6 +873,7 @@ sub connection {
 	show_times("begin dispatch") if $Global::ShowTimes;
     ::dispatch($http) if $http;
 	show_times("end connection") if $Global::ShowTimes;
+	close $http->{rfh} if $http->{rfh};
 	undef $Vend::Cfg;
 }
 
@@ -931,6 +903,8 @@ use vars qw(
 			%Page_pids
 			$SOAP_servers
 			%SOAP_pids
+			$Job_servers
+			%Lifetime
 			$vector
 			$p_vector
 			$s_vector
@@ -941,6 +915,8 @@ BEGIN {
 }
 $Num_servers = 0;
 $SOAP_servers = 0;
+$Job_servers = 0;
+%Lifetime = ();
 
 # might also trap: QUIT
 
@@ -964,6 +940,7 @@ sub reset_vars {
 	package CGI;
 	reset 'A-Z';
 	reset 'a-z';
+	srand();
 #::logDebug("Reset vars");
 }
 
@@ -1028,7 +1005,7 @@ sub housekeeping {
 	rand();
 	$Last_housekeeping = $now;
 
-	my ($c, $num,$reconfig, $restart, @files);
+	my ($c, $num,$reconfig, $restart, $jobs, @files);
 	my @pids;
 
 		if($Global::PreFork) {
@@ -1078,6 +1055,7 @@ sub housekeeping {
 		($reconfig) = grep $_ eq 'reconfig', @files;
 		($restart) = grep $_ eq 'restart', @files
 			if $Signal_Restart || $Global::Windows;
+		($jobs) = grep $_ eq 'jobsqueue', @files;
 		if($Global::PIDcheck) {
 			$Num_servers = 0;
 			@pids = grep /^pid\.\d+$/, @files;
@@ -1150,8 +1128,12 @@ EOF
                 }
 
 				eval {
-					$c = ::config_named_catalog($cat->{CatalogName},
-                                    "from running server ($$)", $table, $cfile);
+					$c = Vend::Config::config_named_catalog(
+									$cat->{CatalogName},
+                                    "from running server ($$)",
+									$table,
+									$cfile
+								);
 				};
 
 				if (defined $c) {
@@ -1180,6 +1162,66 @@ EOF
 			$respawn = 1;
 			
 		}
+		if (defined $jobs) {
+			my (@scheduled_jobs, @queued_jobs);
+			open(Vend::Server::JOBS, "+<$Global::RunDir/jobsqueue")
+				or die "open $Global::RunDir/jobsqueue: $!\n";
+			lockfile(\*Vend::Server::JOBS, 1, 1)
+				or die "lock $Global::RunDir/jobsqueue: $!\n";
+			while(<Vend::Server::JOBS>) {
+				chomp;
+				my ($directive,$value) = split /\s+/, $_, 2;
+				my ($cat, $delay, @jobs) = grep /\S/, split /[\s,\0]+/, $value;
+				if ($delay && $delay < time()) {
+					# job expired
+#::logDebug ("Jobs @jobs expired ($delay vs $now)\n");
+				} elsif ($Job_servers++ >= $Global::Jobs->{MaxServers}) {
+						# no slot for job
+						$Job_servers--;
+#::logDebug ("Jobs @jobs queued, already %d jobs running/scheduled", $Job_servers);
+                        push(@queued_jobs, "$directive $value");
+                } else {
+#::logDebug ("Scheduled job @jobs for running");
+					push (@scheduled_jobs, [$cat, @jobs]);
+				}
+                if (@queued_jobs > 20) {
+					::logGlobal({ level => 'notice' }, "Excessive size of job queue, stopping");
+					last;
+				}
+			}
+
+			truncate(Vend::Server::JOBS, 0)
+				or die "truncate $Global::RunDir/jobsqueue: $!\n";
+            seek(Vend::Server::JOBS, 0, 0)
+                or die "seek $Global::RunDir/jobsqueue: $!\n";
+
+            if (@queued_jobs) {
+#::logDebug("Size of queue $$: %s", scalar(@queued_jobs));
+				print Vend::Server::JOBS join("\n", @queued_jobs, '');
+                unlockfile(\*Vend::Server::JOBS)
+					or die "unlock $Global::RunDir/jobsqueue: $!\n";
+				close(Vend::Server::JOBS)
+					or die "close $Global::RunDir/jobsqueue: $!\n";
+			} else {
+				unlockfile(\*Vend::Server::JOBS)
+					or die "unlock $Global::RunDir/jobsqueue: $!\n";
+				close(Vend::Server::JOBS)
+					or die "close $Global::RunDir/jobsqueue: $!\n";
+				unlink "$Global::RunDir/jobsqueue"
+					or die "unlink $Global::RunDir/jobsqueue: $!\n";
+			}
+
+			# now we run the scheduled jobs
+			for my $jobref (@scheduled_jobs) {
+				eval {
+					run_jobs (@$jobref);
+				};
+
+				if($@) {
+					::logGlobal({ level => 'notice' }, $@);
+				}
+			}
+		}
 
 		if($respawn) {
 			if($Global::PreFork) {
@@ -1225,8 +1267,20 @@ EOF
             my $fn = "$Global::RunDir/$_";
             ($Num_servers--, next) if ! -f $fn;
             my $runtime = $now - (stat(_))[9];
-            next if $runtime < $Global::PIDcheck;
             s/^pid\.//;
+            my ($lifetime, $isjob);
+            if (exists $Lifetime{$_}) {
+ 				$lifetime = $Lifetime{$_};
+				$isjob = 1;
+			} else {
+				$lifetime = $Global::PIDcheck;
+			}
+            next if $runtime < $lifetime;
+			if ($isjob) {
+	            delete $Lifetime{$_};
+    	        $Job_servers--;
+			}
+
             if(kill 9, $_) {
                 unlink $fn and $Num_servers--;
                 ::logGlobal({ level => 'error' }, "hammered PID %s running %s seconds", $_, $runtime);
@@ -1259,8 +1313,12 @@ sub server_start_message {
 	push (@types, 'INET') if $Global::Inet_Mode;
 	push (@types, 'UNIX') if $Global::Unix_Mode;
 	push (@types, 'SOAP') if $Global::SOAP;
+	push (@types, 'mod_perl') if $Global::mod_perl;
 	my $server_type = join(" and ", @types);
-	my @args = $reverse ? ($server_type, $$) : ($$, $server_type);
+	my $pid = ( $Global::PreFork || $Global::Variable->{MV_BAD_LOCK} )
+			  ? $$
+			  : read_pidfile();
+	my @args = $reverse ? ($server_type, $pid) : ($pid, $server_type);
 	return ::errmsg ($fmt , @args );
 }
 
@@ -1501,26 +1559,7 @@ sub start_soap {
 		}
 		elsif (! $pid) {
 			unless( $pid = fork ) {
-				close(STDOUT);
-				close(STDERR);
-				close(STDIN);
-				if ($Global::DebugFile) {
-					open(Vend::DEBUG, ">>$Global::DebugFile.soap");
-					select Vend::DEBUG;
-					$| =1;
-					print "Start DEBUG at " . localtime() . "\n";
-				}
-				elsif (!$Global::DEBUG) {
-					# May as well turn warnings off, not going anywhere
-					$^W = 0;
-					open (Vend::DEBUG, ">/dev/null") unless $Global::Windows;
-				}
-
-				open(STDOUT, ">&Vend::DEBUG");
-				select(STDOUT);
-				$| = 1;
-				open(STDERR, ">&Vend::DEBUG");
-				select(STDERR); $| = 1; select(STDOUT);
+				setup_debug_log();
 
 				$Global::Foreground = 1;
 
@@ -1571,10 +1610,13 @@ sub server_page {
 	my $rout;
 	my $pid;
 	my $spawn;
+	my $start_time = $Global::ChildLife ? time() : 0;
+	my $end_of_life;
 	my $handled = 0;
 	
 	$Global::Foreground ||= $no_fork;
 
+#::logDebug("Start time is $start_time");
     for (;;) {
 
 	  my $n;
@@ -1601,7 +1643,12 @@ my $pretty_vector = unpack('b*', $rin);
         }
 		elsif($n == 0) {
 			undef $spawn;
-			#indiv_housekeeping();
+			if($start_time) {
+				my $current_time = time();
+				next unless $current_time - $start_time > $Global::ChildLife;
+				$end_of_life = 1;
+				last;
+			}
 			next;
 		}
         else {
@@ -1625,7 +1672,6 @@ my $pretty_vector = unpack('b*', $rin);
 			}
 
 			CHECKHOST: {
-				undef $Global::OnlyInternalHTTP;
 				last CHECKHOST if $unix_socket{$p};
 				my $connector;
 				(undef, $ok) = sockaddr_in($ok);
@@ -1635,7 +1681,6 @@ my $pretty_vector = unpack('b*', $rin);
 				(undef, $dns_name) = gethostbyaddr($ok, AF_INET);
 				$dns_name = "UNRESOLVED_NAME" if ! $dns_name;
 				last CHECKHOST if $dns_name =~ /$Global::TcpHost/;
-				$Global::OnlyInternalHTTP = "$dns_name/$connector";
 			}
 			$spawn = 1;
 		}
@@ -1686,7 +1731,7 @@ my $pretty_vector = unpack('b*', $rin);
 					}
 
 					undef $::Instance;
-					select(undef,undef,undef,0.050) until getppid == 1;
+					select(undef,undef,undef,0.050) until &$ppidsub == 1;
 					&$Sig_dec and unlink_pid();
 					exit(0);
 				}
@@ -1716,13 +1761,16 @@ my $pretty_vector = unpack('b*', $rin);
 
 		}
 
-		return 1
-			if $no_fork
-			and $Global::MaxRequestsPerChild
-			and $handled >= $Global::MaxRequestsPerChild;
-
 		return if $Signal_Terminate;
 
+		next unless $no_fork;
+		
+		return 1   if $end_of_life;
+
+		return 1   if  $Global::MaxRequestsPerChild
+				   and $handled >= $Global::MaxRequestsPerChild;
+
+
     }
 }
 
@@ -1796,7 +1844,6 @@ my $pretty_vector = unpack('b*', $s_vector);
 				(undef, $dns_name) = gethostbyaddr($ok, AF_INET);
 				$dns_name = $connector if ! $dns_name;
 				last CHECKHOST if $dns_name =~ /$Global::TcpHost/;
-				$Global::OnlyInternalHTTP = "$dns_name/$connector";
 			}
 
 			$handled++;
@@ -1805,26 +1852,23 @@ my $pretty_vector = unpack('b*', $s_vector);
 			
 			reset_vars();
 
-			$Vend::OnlyInternalHTTP = $Global::OnlyInternalHTTP;
-
-			$Vend::Cfg = http_soap(\*MESSAGE, \%env, \$entity);
-			$Vend::Cat = $Vend::Cfg->{CatalogName};
+			if ($Vend::Cfg = http_soap(\*MESSAGE, \%env, \$entity)) {
+				$Vend::Cat = $Vend::Cfg->{CatalogName};
+			}
 
 			my $result;
 			my $error;
 			if(! $Vend::Cfg) {
 #::logDebug("we have no catalog");
 				$result = Vend::SOAP::Transport::Server
-					->new( error => 'Catalog not found' )
-					->dispatch_to('', 'Vend::SOAP::Error')
-					->handle();
+					->new()
+					->make_fault('Client.NotFound','Service not found');
 			}
 			elsif(! $Vend::Cfg->{SOAP}) {
 #::logDebug("we have no SOAP enable");
 				$result = Vend::SOAP::Transport::Server
-					->new( error => 'SOAP not enabled' )
-					->dispatch_to('', 'Vend::SOAP::Error')
-					->handle();
+					->new()
+					->make_fault('Client.NotAvailable','Service not available');
 			}
 			else {
 #::logDebug("we have our SOAP enable, entity is $entity");
@@ -1838,7 +1882,19 @@ my $pretty_vector = unpack('b*', $s_vector);
 					->handle;
 			}
 
+			unless ($Vend::StatusLine =~ m{^HTTP/}) {
+				my $status = $Vend::StatusLine =~ /(?:^|\n)Status:\s+(.*)/i
+					? "$1" : "200 OK";
+				$Vend::StatusLine = "HTTP/1.0 $status\r\n" . $Vend::StatusLine;
+			}
+			$Vend::StatusLine .= "\r\nContent-Type: text/xml\r\n"
+				unless $Vend::StatusLine =~ /^Content-Type:/im;
+
+			print MESSAGE canon_status($Vend::StatusLine);
+			print MESSAGE "\r\n";
 			print MESSAGE $result;
+			undef $Vend::StatusLine;
+			$Vend::ResponseMade = 1;
 			close MESSAGE;
 #::logDebug("SOAP port=$p n=$n unix=$unix_socket{$p} pid=$$ c=$c time=" . join '|', times);
 		}
@@ -1892,6 +1948,15 @@ sub process_ipc {
 		$SOAP_servers--;
 		start_soap(undef, 1);
 	}
+	elsif ($thing =~ /^running job (\d+)/) {
+#::logDebug("registered job pid $1");
+		$Lifetime{$1} = $Global::Jobs->{MaxLifetime} || 30;
+	}
+	elsif ($thing =~ /^finishing job (\d+)/) {
+#::logDebug("finished job pid $1");
+		$Job_servers--;
+		delete $Lifetime{$1};
+	}
 	elsif($thing =~ /^\d+$/) {
 		close $fh;
 		$Num_servers++;
@@ -1914,19 +1979,47 @@ sub send_ipc {
 	close SOCK;
 }
 
+sub setup_debug_log {
+	if ($Global::DebugFile) {
+		open(Vend::DEBUG, ">>$Global::DebugFile");
+		select Vend::DEBUG;
+		$| = 1;
+		print "Start DEBUG at " . localtime() . "\n";
+	}
+	elsif (!$Global::DEBUG) {
+		# May as well turn warnings off, not going anywhere
+		$^W = 0;
+		open (Vend::DEBUG, ">/dev/null") unless $Global::Windows;
+	}
+
+	close(STDIN);
+	close(STDOUT);
+	close(STDERR);
+
+	open(STDOUT, ">&Vend::DEBUG");
+	select(STDOUT);
+	$| = 1;
+
+	open(STDERR, ">&Vend::DEBUG");
+	select(STDERR);
+	$| = 1;
+}
+
 # The servers for both are now combined
 # Can have both INET and UNIX on same system
 sub server_both {
     my ($socket_filename) = @_;
     my ($n, $rin, $rout, $pid);
 
+	::logGlobal({ level => 'info' }, server_start_message());
+
 	$Vend::MasterProcess = $$;
 
 	$tick        = $Global::HouseKeeping || 60;
 
     setup_signals();
 
-#::logDebug("Starting server socket file='$socket_filename' hosts='$host'\n");
+#::logDebug("Starting server socket file='$socket_filename'\n");
 
 	my $spawn;
 
@@ -1944,7 +2037,6 @@ sub server_both {
 		$ipc_socket{$ipc} = $ipc;
 		$unix_socket{$ipc} = $ipc;
 		$ipc_vector = $vector;
-		
 	}
 
 	# Make UNIX-domain sockets if applicable. The sockets are mapped into the
@@ -1984,7 +2076,7 @@ sub server_both {
 			if($Global::SocketPerms & 033) {
 				::logGlobal( {
 					level => 'warn' },
-					"ALERT: %s socket permissions are insecure; are you sure you want permssions %o?",
+					"ALERT: %s socket permissions are insecure; are you sure you want permissions %o?",
 					$_,
 					$Global::SocketPerms,
 				);
@@ -2072,27 +2164,7 @@ sub server_both {
 		::logGlobal({ level => 'info' }, "Running in foreground, OS=$^O, debug=$Global::DEBUG\n");
 	}
 	else {
-		close(STDIN);
-		close(STDOUT);
-		close(STDERR);
-
-		if ($Global::DebugFile) {
-			open(Vend::DEBUG, ">>$Global::DebugFile");
-			select Vend::DEBUG;
-			$| =1;
-			print "Start DEBUG at " . localtime() . "\n";
-		}
-		elsif (!$Global::DEBUG) {
-			# May as well turn warnings off, not going anywhere
-			$^W = 0;
-			open (Vend::DEBUG, ">/dev/null") unless $Global::Windows;
-		}
-
-		open(STDOUT, ">&Vend::DEBUG");
-		select(STDOUT);
-		$| = 1;
-		open(STDERR, ">&Vend::DEBUG");
-		select(STDERR); $| = 1; select(STDOUT);
+		setup_debug_log();
 #::logDebug("s_vector=" . unpack('b*', $s_vector));
 		if($s_vector) {
 			start_soap(1);
@@ -2168,7 +2240,6 @@ my $pretty_vector = unpack('b*', $rin);
 			}
 
 			CHECKHOST: {
-				undef $Vend::OnlyInternalHTTP;
 				last CHECKHOST if $unix_socket{$p};
 				my $connector;
 				(undef, $ok) = sockaddr_in($ok);
@@ -2178,7 +2249,6 @@ my $pretty_vector = unpack('b*', $rin);
 				(undef, $dns_name) = gethostbyaddr($ok, AF_INET);
 				$dns_name = "UNRESOLVED_NAME" if ! $dns_name;
 				last CHECKHOST if $dns_name =~ /$Global::TcpHost/;
-				$Vend::OnlyInternalHTTP = "$dns_name/$connector";
 			}
 			$spawn = 1 unless $only_ipc;
 		}
@@ -2194,7 +2264,7 @@ my $pretty_vector = unpack('b*', $rin);
 	  eval {
 		SPAWN: {
 			last SPAWN unless defined $spawn;
-#::logDebug #("Spawning connection, " .  ($no_fork ? 'no fork, ' : 'forked, ') .  scalar localtime() . "\n");
+#::logDebug("Spawning connection, " .  ($no_fork ? 'no fork, ' : 'forked, ') .  scalar localtime() . "\n");
 			if(defined $no_fork) {
 				$::Instance = {};
 				connection();
@@ -2225,7 +2295,7 @@ my $pretty_vector = unpack('b*', $rin);
 					clean_up_after_fork();
 
 					undef $::Instance;
-					select(undef,undef,undef,0.050) until getppid == 1;
+					select(undef,undef,undef,0.050) until &$ppidsub == 1;
 					if ($Global::IPCsocket) {
 						&$Sig_dec and unlink_pid();
 					}
@@ -2313,6 +2383,73 @@ sub touch_pid {
 		or die "PID $$ conflict: can't lock\n";
 }
 
+sub jobs_job {
+	my ($cat, @jobs) = @_;
+	for my $job (@jobs) {
+		Vend::Dispatch::run_in_catalog($cat, $job);
+	}
+}
+
+sub run_jobs {
+	my ($cat, @jobs) = @_;
+
+#::logGlobal("Vend::Server::run_jobs: run jobs cat=cat jobs=@jobs");
+	my $pid;
+	if($Global::Foreground) {
+		$::Instance = {};
+		eval {
+			jobs_job($cat, @jobs);
+		};
+		if($@) {
+			my $msg = $@;
+			::logGlobal({ level => 'error' }, "Runtime error: %s" , $msg);
+			logError("Runtime jobs error: %s", $msg)
+				if defined $Vend::Cfg->{ErrorFile};
+		}
+		clean_up_after_fork();
+		undef $::Instance;
+	}
+	elsif(! defined ($pid = fork) ) {
+		my $msg = ::errmsg("Can't fork: %s", $!);
+		::logGlobal({ level => 'crit' },  $msg );
+		die ("$msg\n");
+	}
+	elsif (! $pid) {
+		#fork again
+		unless ($pid = fork) {
+
+			send_ipc("running job $$");
+			reset_per_fork();
+			$::Instance = {};
+			eval { 
+				touch_pid() if $Global::PIDcheck;
+				&$Sig_inc;
+				jobs_job($cat, @jobs);
+			};
+			if ($@) {
+				my $msg = $@;
+				::logGlobal({ level => 'error' }, "Runtime error: %s" , $msg);
+				logError("Runtime jobs error: %s", $msg)
+					if defined $Vend::Cfg->{ErrorFile};
+			}
+			clean_up_after_fork();
+			send_ipc("finishing job $$");
+
+			undef $::Instance;
+			select(undef,undef,undef,0.050) until &$ppidsub == 1;
+			if ($Global::PIDcheck) {
+				unlink_pid() and &$Sig_dec;
+			}
+			else {
+				&$Sig_dec;
+			}
+			exit(0);
+		}
+		exit(0);
+	}
+	wait unless $Global::Foreground;
+}
+
 sub unlink_pid {
 	close(TEMPPID);
 	unlink("$Global::RunDir/pid.$$");
@@ -2331,12 +2468,10 @@ sub grab_pid {
         no strict 'subs';
         truncate($fh, 0) or die "Couldn't truncate pid file: $!\n";
     }
-    print $fh $$, "\n";
+    print $fh ($Global::mod_perl ? &$ppidsub : $$), "\n";
     return 0;
 }
 
-
-
 sub open_pid {
 	my $fn = shift || $Global::PIDfile;
 	my $fh = gensym();
@@ -2349,34 +2484,89 @@ sub open_pid {
 	return $fh;
 }
 
+sub read_pidfile {
+	my $fn = shift || $Global::PIDfile;
+	my $fh = gensym();
+	open $fh, "<$fn" or return;
+	chomp (my $pid = <$fh>);
+	close $fh;
+	return $pid;
+}
+
 sub run_server {
     my $next;
-	
-    my $pidh = open_pid($Global::PIDfile);
+#::logDebug("trying to run server");
 
-	unless($Global::Inet_Mode || $Global::Unix_Mode || $Global::Windows) {
-		$Global::Inet_Mode = $Global::Unix_Mode = 1;
+	if($Global::Variable->{MV_GETPPID_BROKEN}) {
+#::logDebug("setting getppid broken");
+		my $num = $Global::Variable->{MV_GETPPID_BROKEN} > 1
+				? $Global::Variable->{MV_GETPPID_BROKEN}
+				: 64;
+		$ppidsub = sub {
+			return syscall($num);
+		};
+	}
+    my $pidh = open_pid($Global::PIDfile);
+#::logDebug("Opened pid file");
+
+	if($Global::AcceptRedirect) {
+		push @Map, @RedirMap
+			unless grep $_ eq 'REDIRECT_URL', @Map;
+	}
+
+	if ($Global::mod_perl) {
+		undef $Global::Unix_Mode;
+		undef $Global::Inet_Mode;
+		undef $Global::StartServers;
+		undef $Global::PreFork;
+		undef $Global::SOAP;
+		undef $Global::IPCsocket;
 	}
 	elsif ( $Global::Windows ) {
 		$Global::Inet_Mode = 1;
 	}
+	elsif (! $Global::Inet_Mode and ! $Global::Unix_Mode) {
+		$Global::Inet_Mode = $Global::Unix_Mode = 1;
+	}
 
-	::logGlobal({ level => 'info' }, server_start_message());
-
-	if($Global::PreFork || $Global::DEBUG || $Global::Windows) {
+	if($Global::mod_perl || $Global::PreFork || $Global::DEBUG || $Global::Windows) {
 		eval {
 			require Tie::ShadowHash;
 		};
 		if($@) {
 			my $reason;
-			if($Global::PreFork)	{ $reason = 'in PreFork mode' }
+			if($Global::mod_perl)	{ $reason = 'under mod_perl' }
+			elsif($Global::PreFork)	{ $reason = 'in PreFork mode' }
 			elsif($Global::DEBUG)	{ $reason = 'in DEBUG mode' }
 			elsif($Global::Windows)	{ $reason = 'under Windows' }
 			die ::errmsg("Running $reason requires Tie::ShadowHash module.") . "\n";
 		}
 	}
 
-    if ($Global::Windows || $Global::DEBUG) {
+	if ($Global::mod_perl) {
+		my $running = grab_pid($pidh);
+		if ($running) {
+			print errmsg(
+				"The Interchange server is already running (process id %s)\n",
+				$running,
+			);
+			undef $Global::mod_perl;
+			return;
+		}
+		# throw away pidfile -- Apache hasn't forked yet, so pid is wrong
+		unlockfile($pidh);
+		unlink $Global::PIDfile;
+		print server_start_message("Interchange server started (%s)\n", 1);
+		::logGlobal(
+			{ level => 'info' },
+			Vend::Server::server_start_message('START server (%s)', 1),
+		);
+		setup_debug_log();
+		# all done; now wait for Apache to call Vend::ModPerl::handler
+		return;
+	}
+
+	if ($Global::Windows || $Global::DEBUG) {
         my $running = grab_pid($pidh);
         if ($running) {
 			print errmsg(
@@ -2420,7 +2610,9 @@ sub run_server {
             }
             else {
                 # child 2
-                sleep 1 until getppid == 1;
+#::logDebug("getting ready to sleep ...");
+                sleep 1 until &$ppidsub == 1;
+#::logDebug("slept ...");
 
                 my $running = grab_pid($pidh);
                 if ($running) {
