@@ -1,11 +1,12 @@
 # Vend::Table::Common - Common access methods for Interchange databases
 #
-# $Id: Common.pm,v 2.0.2.4 2002/01/24 05:07:03 jon Exp $
+# $Id: Common.pm,v 2.34 2003/11/17 23:41:13 edl Exp $
 #
-# Copyright (C) 1996-2002 Red Hat, Inc. <interchange@redhat.com>
+# Copyright (C) 2002-2003 Interchange Development Group
+# Copyright (C) 1996-2002 Red Hat, Inc.
 #
 # This program was originally based on Vend 0.2 and 0.3
-# Copyright 1995 by Andrew M. Wilcox <awilcox@world.std.com>
+# Copyright 1995 by Andrew M. Wilcox <amw@wilcoxsolutions.com>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -22,13 +23,14 @@
 # Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 # MA  02111-1307  USA.
 
-$VERSION = substr(q$Revision: 2.0.2.4 $, 10);
+$VERSION = substr(q$Revision: 2.34 $, 10);
 use strict;
 
 package Vend::Table::Common;
 require Vend::DbSearch;
 require Vend::TextSearch;
-require File::CounterFile;
+require Vend::CounterFile;
+use Symbol;
 use Vend::Util;
 
 use Exporter;
@@ -99,8 +101,13 @@ sub create_columns {
 #::logDebug("set KEY_INDEX to $i: " . ::uneval($config));
     }
 
-    die("Cannot find key column $config->{KEY} in $config->{name} ($config->{file}): $!")
-        unless defined $config->{KEY_INDEX};
+    die errmsg(
+			"Cannot find key column %s in %s (%s): %s",
+			$config->{KEY},
+			$config->{name},
+			$config->{file},
+			$!,
+	    ) unless defined $config->{KEY_INDEX};
 
 	return $column_index;
 }
@@ -120,6 +127,37 @@ sub separate_definitions {
 	return;
 }
 
+sub clear_lock {
+	my $s = shift;
+	return unless $s->[$CONFIG]{IC_LOCKING};
+	if($s->[$CONFIG]{_lock_handle}) {
+		close $s->[$CONFIG]{_lock_handle};
+		delete $s->[$CONFIG]{_lock_handle};
+	}
+}
+
+sub lock_table {
+	my $s = shift;
+	return unless $s->[$CONFIG]{IC_LOCKING};
+	my $lockhandle;
+	if(not $lockhandle = $s->[$CONFIG]{_lock_handle}) {
+		my $lf = $s->[$CONFIG]{file} . '.lock';
+		$lockhandle = gensym;
+		$s->[$CONFIG]{_lock_file} = $lf;
+		$s->[$CONFIG]{_lock_handle} = $lockhandle;
+		open $lockhandle, ">> $lf"
+			or die errmsg("Cannot lock table %s: %s", $s->[$CONFIG]{name}, $!);
+	}
+#::logDebug("lock handle=$lockhandle");
+	Vend::Util::lockfile($lockhandle);
+}
+
+sub unlock_table {
+	my $s = shift;
+	return unless $s->[$CONFIG]{IC_LOCKING};
+	Vend::Util::unlockfile($s->[$CONFIG]{_lock_handle});
+}
+
 sub stuff {
     my ($val) = @_;
     $val =~ s,([\t\%]),$Hex_string[ord($1)],eg;
@@ -135,13 +173,20 @@ sub unstuff {
 sub autonumber {
 	my $s = shift;
 	my $start;
-	return $s->[$CONFIG]{SEQUENCE_VAL} if $s->[$CONFIG]{AUTO_SEQUENCE};
-	return '' if not $start = $s->[$CONFIG]->{AUTO_NUMBER};
+	my $cfg = $s->[$CONFIG];
+
+	return $s->autosequence() if $cfg->{AUTO_SEQUENCE};
+
+	return '' if not $start = $cfg->{AUTO_NUMBER};
 	local($/) = "\n";
 	my $c = $s->[$CONFIG];
 	if(! defined $c->{AutoNumberCounter}) {
-		$c->{AutoNumberCounter} = new File::CounterFile
-									"$c->{DIR}/$c->{name}.autonumber", $start;
+		my $dot = $c->{HIDE_AUTO_FILES} ? '.' : '';
+		$c->{AutoNumberCounter} = new Vend::CounterFile
+									"$c->{DIR}/$dot$c->{name}.autonumber",
+									$start,
+									$c->{AUTO_NUMBER_DATE},
+									;
 	}
 	my $num;
 	do {
@@ -190,8 +235,9 @@ sub close_table {
 	return 1 if ! defined $s->[$TIE_HASH];
 #::logDebug("closing table $s->[$FILENAME]");
 	undef $s->[$DBM];
+	$s->clear_lock();
     untie %{$s->[$TIE_HASH]}
-		or die "Could not close DBM table $s->[$FILENAME]: $!\n";
+		or $s->log_error("%s %s: %s", errmsg("untie"), $s->[$FILENAME], $!);
 	undef $s->[$TIE_HASH];
 #::logDebug("closed table $s->[$FILENAME], self=" . ::uneval($s));
 }
@@ -229,7 +275,11 @@ sub column_index {
     my ($s, $column) = @_;
 	$s = $s->import_db() if ! defined $s->[$TIE_HASH];
     my $i = $s->[$COLUMN_INDEX]{$column};
-    die "There is no column named '$column'" unless defined $i;
+    die $s->log_error(
+				"There is no column named '%s' in %s",
+				$column,
+				$s->[$FILENAME],
+			) unless defined $i;
     return $i;
 }
 
@@ -238,12 +288,16 @@ sub column_index {
 sub record_exists {
     my ($s, $key) = @_;
 	$s = $s->import_db() if ! defined $s->[$TIE_HASH];
-    # guess what?  The GDBM "exists" function got renamed to "EXISTS" 
-    # in 5.002.
     my $r = $s->[$DBM]->EXISTS("k$key");
     return $r;
 }
 
+sub name {
+	my ($s) = shift;
+	$s = $s->import_db() if ! defined $s->[$TIE_HASH];
+	return $s->[$CONFIG]{name};
+}
+
 sub row_hash {
     my ($s, $key) = @_;
 	$s = $s->import_db() if ! defined $s->[$TIE_HASH];
@@ -255,8 +309,14 @@ sub row_hash {
 
 sub unstuff_row {
     my ($s, $key) = @_;
+	$s->lock_table() if $s->[$CONFIG]{IC_LOCKING};
     my $line = $s->[$TIE_HASH]{"k$key"};
-    die "There is no row with index '$key'" unless defined $line;
+	$s->unlock_table() if $s->[$CONFIG]{IC_LOCKING};
+    die $s->log_error(
+					"There is no row with index '%s' in database %s",
+					$key,
+					$s->[$FILENAME],
+			) unless defined $line;
     return map(unstuff($_), split(/\t/, $line, 9999))
 		unless $s->[$CONFIG]{FILTER_FROM};
 	my @f = map(unstuff($_), split(/\t/, $line, 9999));
@@ -266,8 +326,11 @@ sub unstuff_row {
 
 sub thaw_row {
     my ($s, $key) = @_;
+	$s->lock_table() if $s->[$CONFIG]{IC_LOCKING};
     my $line = $s->[$TIE_HASH]{"k$key"};
-    die "There is no row with index '$key'" unless defined $line;
+	$s->unlock_table() if $s->[$CONFIG]{IC_LOCKING};
+    die $s->log_error( "There is no row with index '%s'", $key,)
+		unless defined $line;
     return (@{ Storable::thaw($line) })
 		unless $s->[$CONFIG]{FILTER_FROM};
 #::logDebug("filtering.");
@@ -329,7 +392,7 @@ sub set_slice {
 	$s = $s->import_db() if ! defined $s->[$TIE_HASH];
 
     if($s->[$CONFIG]{Read_only}) {
-		::logError(
+		$s->log_error(
 			"Attempt to set slice of %s in read-only table %s",
 			$key,
 			$s->[$CONFIG]{name},
@@ -355,10 +418,21 @@ sub set_slice {
 		unshift @$vary, $key;
 	}
 
-	my $sub = $s->row_settor(@$fary)
-		or die "failed to create row slice routine.";
+	my @current;
 
-	return $sub->(@$vary);
+	@current = $s->row($key)
+		if $s->record_exists($key);
+
+	@current[ map { $s->column_index($_) } @$fary ] = @$vary;
+
+	$key = $s->set_row(@current);
+	length($key) or
+		$s->log_error(
+			"Did set_slice with empty key on table %s",
+			$s->[$CONFIG]{name},
+		);
+
+	return $key;
 }
 
 sub field_settor {
@@ -402,11 +476,14 @@ sub clone_set {
 sub stuff_row {
     my ($s, @fields) = @_;
 	my $key = $fields[$s->[$KEY_INDEX]];
+#::logDebug("stuff key=$key");
 	$fields[$s->[$KEY_INDEX]] = $key = $s->autonumber()
-		if ! $key;
+		if ! length($key);
 	$s->filter(\@fields, $s->[$COLUMN_INDEX], $s->[$CONFIG]{FILTER_TO})
 		if $s->[$CONFIG]{FILTER_TO};
+	$s->lock_table();
     $s->[$TIE_HASH]{"k$key"} = join("\t", map(stuff($_), @fields));
+	$s->unlock_table();
 	return $key;
 }
 
@@ -418,7 +495,9 @@ sub freeze_row {
 		if ! length($key);
 	$s->filter(\@fields, $s->[$COLUMN_INDEX], $s->[$CONFIG]{FILTER_TO})
 		if $s->[$CONFIG]{FILTER_TO};
+	$s->lock_table();
 	$s->[$TIE_HASH]{"k$key"} = Storable::freeze(\@fields);
+	$s->unlock_table();
 	return $key;
 }
 
@@ -453,7 +532,10 @@ sub set_field {
     my ($s, $key, $column, $value) = @_;
 	$s = $s->import_db() if ! defined $s->[$TIE_HASH];
     if($s->[$CONFIG]{Read_only}) {
-		::logError("Attempt to set $s->[$CONFIG]{name}::${column}::$key in read-only table");
+		$s->log_error(
+			"Attempt to write %s in read-only table",
+			"$s->[$CONFIG]{name}::${column}::$key",
+		);
 		return undef;
 	}
     my @row;
@@ -474,7 +556,10 @@ sub inc_field {
 	$s = $s->import_db() if ! defined $s->[$TIE_HASH];
     my($value);
     if($s->[$CONFIG]{Read_only}) {
-		::logError("Attempt to set $s->[$CONFIG]{name}::${column}::$key in read-only table");
+		$s->log_error(
+			"Attempt to write %s in read-only table",
+			"$s->[$CONFIG]{name}::${column}::$key",
+		);
 		return undef;
 	}
     my @row = $s->row($key);
@@ -485,6 +570,10 @@ sub inc_field {
     return $value;
 }
 
+sub create_sql {
+    return undef;
+}
+
 sub touch {
     my ($s) = @_;
 	$s = $s->import_db() if ! defined $s->[$TIE_HASH];
@@ -545,19 +634,31 @@ sub each_record {
 my $sup;
 my $restrict;
 my $rfield;
+my $hfield;
 my $rsession;
 
 sub each_nokey {
-    my ($s) = @_;
+    my ($s, $qual) = @_;
 	$s = $s->import_db() if ! defined $s->[$TIE_HASH];
-    my ($key);
+    my ($key, $hf);
 
-	if (
-		! defined $restrict
-		and 
-		$restrict = ($Vend::Cfg->{TableRestrict}{$s->config('name')} || 0)
-		)
-	{
+	if (! defined $restrict) {
+		# Support hide_field
+		if($qual) {
+#::logDebug("Found qual=$qual");
+			$hfield = $qual;
+			if($hfield =~ s/^\s+WHERE\s+(\w+)\s*!=\s*1($|\s+)//) {
+				$hf = $1;
+#::logDebug("Found hf=$hf");
+				$s->test_column($hf) and $hfield = $s->column_index($hf);
+			}
+			else {
+				undef $hfield;
+			}
+
+#::logDebug("hf index=$hfield");
+		}
+		if($restrict = ($Vend::Cfg->{TableRestrict}{$s->config('name')} || 0)) {
 #::logDebug("restricted?");
 		$sup =  ! defined $Global::SuperUserFunction
 					||
@@ -571,6 +672,10 @@ sub each_nokey {
 				or $restrict = 0;
 			$rsession = $Vend::Session->{$rsession};
 		}
+	}
+
+		$restrict = 1 if $hfield and $s->[$CONFIG]{HIDE_FIELD} eq $hf;
+
 	}
 
     for (;;) {
@@ -583,11 +688,13 @@ sub each_nokey {
 		$key =~ s/^k// or next;
 		if($restrict) {
 			my (@row) = $s->row($key);
-##::logDebug("each_nokey: '$row[$rfield]' eq '$rsession' ??");
-			next if $row[$rfield] ne $rsession;
-			return @row;
+#::logDebug("each_nokey: rfield='$row[$rfield]' eq '$rsession' ??") if defined $rfield;
+#::logDebug("each_nokey: hfield='$row[$hfield]'") if defined $hfield;
+			next if defined $hfield and $row[$hfield];
+			next if defined $rfield and $row[$rfield] ne $rsession;
+			return \@row;
 		}
-		return $s->row($key);
+		return [ $s->row($key) ];
     }
 }
 
@@ -601,7 +708,11 @@ sub delete_record {
     my ($s, $key) = @_;
 	$s = $s->import_db() if ! defined $s->[$TIE_HASH];
     if($s->[$CONFIG]{Read_only}) {
-		::logError("Attempt to delete row '$key' in read-only table $s->[$CONFIG]{name}");
+		$s->log_error(
+			"Attempt to delete row '$key' in read-only table %s",
+			$key,
+			$s->[$CONFIG]{name},
+		);
 		return undef;
 	}
 
@@ -673,28 +784,39 @@ sub query {
 	}
 	else {
 		eval {
-			($spec, $stmt) = Vend::Scan::sql_statement($query, $ref);
+			($spec, $stmt) = Vend::Scan::sql_statement($query, $opt);
 		};
 		if($@) {
-			my $msg = ::errmsg("SQL query failed: %s\nquery was: %s", $@, $query);
+			my $msg = errmsg("SQL query failed: %s\nquery was: %s", $@, $query);
+			$s->log_error($msg);
 			Carp::croak($msg) if $Vend::Try;
-			::logError($msg);
 			return ($opt->{failure} || undef);
 		}
 		my @additions = grep length($_) == 2, keys %$opt;
-		if(@additions) {
-			@{$spec}{@additions} = @{$opt}{@additions};
+		for(@additions) {
+			next unless length $opt->{$_};
+			$spec->{$_} = $opt->{$_};
 		}
 	}
-	my @tabs = @{$spec->{fi}};
-	for (@tabs) {
-		s/\..*//;
+	my @tabs = @{$spec->{rt} || $spec->{fi}};
+
+	my $reroute;
+	my $tname = $s->[$CONFIG]{name};
+	if ($tabs[0] ne $tname) {
+		if("$tabs[0]_txt" eq $tname or "$tabs[0]_asc" eq $tname) {
+			$tabs[0] = $spec->{fi}[0] = $tname;
+		}
+		else {
+			$reroute = 1;
+		}
 	}
-	if (! defined $s || $tabs[0] ne $s->[$CONFIG]{name}) {
-		unless ($s = $Vend::Database{$tabs[0]}) {
-			::logError("Table %s not found in databases", $tabs[0]);
+
+	if($reroute) {
+		unless ($reroute = $Vend::Database{$tabs[0]}) {
+			$s->log_error("Table %s not found in databases", $tabs[0]);
 			return $opt->{failure} || undef;
 		}
+		$s = $reroute;
 #::logDebug("rerouting to $tabs[0]");
 		$opt->{STATEMENT} = $stmt;
 		$opt->{SPEC} = $spec;
@@ -703,31 +825,49 @@ sub query {
 
 eval {
 
+	my @vals;
 	if($stmt->command() ne 'SELECT') {
 		if(defined $s and $s->[$CONFIG]{Read_only}) {
-			die ("Attempt to write read-only database $s->[$CONFIG]{name}");
+			$s->log_error(
+					"Attempt to write read-only table %s",
+					$s->[$CONFIG]{name},
+			);
+			return undef;
 		}
 		$update = $stmt->command();
+		@vals = $stmt->row_values();
+#::logDebug("row_values returned=" . ::uneval(\@vals));
 	}
-	my @vals = $stmt->row_values();
-	
+
+
 	@na = @{$spec->{rf}}     if $spec->{rf};
 
+#::logDebug("spec->{ml}=$spec->{ml} opt->{ml}=$opt->{ml}");
+	$spec->{ml} = $opt->{ml} if $opt->{ml};
+	$spec->{ml} ||= '1000';
 	$spec->{fn} = [$s->columns];
 
 	my $sub;
 
 	if($update eq 'INSERT') {
-		@update_fields = $spec->{rf} ? @{$spec->{rf}} : @{$spec->{fn}};
+		if(! $spec->{rf} or  $spec->{rf}[0] eq '*') {
+			@update_fields = @{$spec->{fn}};
+		}
+		else {
+			@update_fields = @{$spec->{rf}};
+		}
+#::logDebug("update fields: " . uneval(\@update_fields));
 		@na = $codename;
 		$sub = $s->row_settor(@update_fields);
 	}
 	elsif($update eq 'UPDATE') {
 		@update_fields = @{$spec->{rf}};
-		@na = $codename;
+#::logDebug("update fields: " . uneval(\@update_fields));
+		my $key = $s->config('KEY');
+		@na = ($codename);
 		$sub = sub {
 					my $key = shift;
-					set_slice($s, $key, \@update_fields, \@_);
+					$s->set_slice($key, [@update_fields], \@_);
 				};
 	}
 	elsif($update eq 'DELETE') {
@@ -775,15 +915,14 @@ eval {
 		die "Bad row settor for columns @na"
 			if ! $sub;
 		if($update eq 'INSERT') {
-			&$sub(@vals);
+			$sub->(@vals);
 			$ref = [[ $vals[0] ]];
 		}
 		else {
-#::logDebug("Supposed to search..., spec=" . ::uneval($spec));
 			$ref = $search->array($spec);
-#::logDebug("Returning ref=" . ::uneval($ref) . "updating with vals=" . join ",", @vals);
-			for(@{$ref}) {
-				&$sub($_->[0], @vals);
+			for(@$ref) {
+#::logDebug("returned =" . uneval($_) . ", update values: " . uneval(\@vals));
+				$sub->($_->[0], @vals);
 			}
 		}
 	}
@@ -802,13 +941,24 @@ eval {
 #::logDebug("ref returned: " . substr(Vend::Util::uneval($ref), 0, 100));
 #::logDebug("opt is: " . Vend::Util::uneval($opt));
 	if($@) {
-		::logError("MVSQL query failed for $opt->{table}: $@\nquery was: $query");
+		$s->log_error(
+				"MVSQL query failed for %s: %s\nquery was: %s",
+				$opt->{table},
+				$@,
+				$query,
+			);
 		$return = $opt->{failure} || undef;
 	}
 
-	return scalar @{$ref || []}
-		if $opt->{row_count};
-	return Vend::Interpolate::region($opt, $text)
+	if ($opt->{row_count}) {
+		my $rc = $ref ? scalar @$ref : 0;
+		return $rc unless $opt->{list};
+		$ref = [ [ $rc ] ];
+		@na = [ 'row_count' ];
+		%nh = ( 'rc' => 0, 'count' => 0, 'row_count' => 0 );
+	}
+
+	return Vend::Interpolate::tag_sql_list($text, $ref, \%nh, $opt, \@na)
 		if $opt->{list};
 	return Vend::Interpolate::html_table($opt, $ref, \@na)
 		if $opt->{html};
@@ -870,13 +1020,13 @@ sub import_ascii_delimited {
 
 	if(! defined $realfile) {
 		open(IN, "+<$infile")
-			or die ::errmsg("Couldn't open '%s' read/write: %s", $infile, $!);
+			or die errmsg("%s %s: %s\n", errmsg("open read/write"), $infile, $!);
 		lockfile(\*IN, 1, 1)
-			or die ::errmsg("lock '%s': %s", $infile, $!);
+			or die errmsg("%s %s: %s\n", errmsg("lock"), $infile, $!);
 	}
 	else {
 		open(IN, "<$infile")
-			or die ::errmsg("Couldn't open '%s' for read: %s", $infile, $!);
+			or die errmsg("%s %s: %s\n", errmsg("open"), $infile, $!);
 	}
 
 	my $field_hash;
@@ -1044,8 +1194,8 @@ EndOfExcel
 			for($i = 0; $i < @i; $i++) {
 				my $fnum = $i[$i];
 				$fh = new IO::File "> $infile.$i[$i]";
-				die "Couldn't create $infile.$i[$i]: $!\n"
-					unless defined $fh;
+				die errmsg("%s %s: %s\n", errmsg("create"), "$infile.$i[$i]",
+				$!) unless defined $fh;
 				eval {
 					unlink "$infile.$n[$i]" if -l "$infile.$n[$i]";
 					symlink "$infile.$i[$i]", "$infile.$n[$i]";
@@ -1204,26 +1354,22 @@ EndOfRoutine
 );
 
     eval $format{$format};
-	die ::errmsg("$options->{name} import failed: %s", $@) if $@;
+	die errmsg("%s import failed: %s", $options->{name}, $@) if $@;
     if($realfile) {
 		close IN
-			or die ::errmsg("close preload file %s: %s", $infile, $!) . "\n";
+			or die errmsg("%s %s: %s\n", errmsg("close"), $infile, $!);
 		if(-f $realfile) {
 			open(IN, "+<$realfile")
-				or die ::errmsg(
-					"Couldn't open user file %s read/write: %s",
-					$realfile,
-					$!) . "\n";
-			lockfile(\*IN, 1, 1) or die "lock\n";
+				or die
+					errmsg("%s %s: %s\n", errmsg("open read/write"), $realfile, $!);
+			lockfile(\*IN, 1, 1)
+				or die errmsg("%s %s: %s\n", errmsg("lock"), $realfile, $!);
 			<IN>;
 			eval $format{$format};
-			die ::errmsg("%s import failed: %s", $options->{name}, $@) if $@;
+			die errmsg("%s %s: %s\n", errmsg("import"), $options->{name}, $!) if $@;
 		}
 		elsif (! open(IN, ">$realfile") ) {
-				warn ::errmsg(
-					"can't create %s import failed: %s",
-									$options->{file}, $@
-								);
+				die errmsg("%s %s: %s\n", errmsg("create"), $realfile, $!);
 		} 
 		else {
 			print IN join($options->{DELIMITER}, @field_names);
@@ -1266,11 +1412,17 @@ EndOfRoutine
 			File::Copy::copy(@{$_});
 		}
 	}
+
+	unless($options->{no_commit}) {
+		$out->commit() if $out->config('HAS_TRANSACTIONS');
+	}
 	delete $out->[$CONFIG]{Clean_start};
+	delete $out->[$CONFIG]{_Dirty};
 	unlockfile(\*IN) or die "unlock\n";
     close(IN);
+	my $dot = $out->[$CONFIG]{HIDE_AUTO_FILES} ? '.' : '';
 	if($numeric_guess) {
-		my $fn = Vend::Util::catfile($out->[$CONFIG]{DIR}, $out->[$CONFIG]{file});
+		my $fn = Vend::Util::catfile($out->[$CONFIG]{DIR}, "$dot$out->[$CONFIG]{file}");
 		Vend::Util::writefile(
 					">$fn.numeric",
 					join " ", map { $field_names[$_] } @possible,
@@ -1279,6 +1431,66 @@ EndOfRoutine
     return $out;
 }
 
+sub import_from_ic_db {
+    my ($infile, $options, $table_name) = @_;
+
+	my $tname = $options->{MIRROR}
+		or die errmsg(
+				"Memory mirror table not specified for table %s.",
+				$table_name,
+			);
+#::logDebug("Importing mirrored $table_name from $tname");
+
+	$Vend::Database{$tname} =
+		Vend::Data::import_database($Vend::Cfg->{Database}{$tname})
+			unless $Vend::Database{$tname};
+
+	my $idb = Vend::Data::database_exists_ref($tname)
+		or die errmsg(
+				"Memory mirror table %s does not exist (yet) to create mirror %s.\n",
+				$tname,
+				$table_name,
+			);
+
+	my @field_names = $idb->columns;
+
+	my $odb;
+
+	if($options->{ObjectType}) {
+		no strict 'refs';
+		$odb = &{"$options->{ObjectType}::create"}(
+									$options->{ObjectType},
+									$options,
+									\@field_names,
+									$table_name,
+								);
+	}
+	else {
+		$odb = $options->{Object};
+	}
+
+#::logDebug("idb=$idb odb=$odb");
+	eval {
+		my $f;
+		while($f = $idb->each_nokey($options->{MIRROR_QUAL})) {
+#::logDebug("importing key=$f->[0]");
+			$odb->set_row(@$f);
+		}
+	};
+
+	if($@) {
+		die errmsg(
+				"Problem with mirror import from source %s to target %s\n",
+				$tname,
+				$table_name,
+				);
+	}
+	
+	$odb->[$CONFIG]{Mirror_complete} = 1;
+	delete $odb->[$CONFIG]{Clean_start};
+    return $odb;
+}
+
 my $white = ' \t';
 
 sub read_quoted_fields {
@@ -1326,6 +1538,46 @@ sub parse {
     return @a;
 }
 
+sub reset {
+	undef $restrict;
+}
+
+sub errstr {
+	return shift(@_)->[$CONFIG]{last_error};
+}
+
+sub log_error {
+	my ($s, $tpl, @args) = @_;
+	if($tpl =~ /^(prepare|execute)$/) {
+		if(!@args) {
+			$tpl = "Statement $tpl failed: %s";
+		}
+		elsif (@args == 1) {
+			$tpl = "Statement $tpl failed: %s\nQuery was: %s";
+		}
+		else {
+			$tpl = "Statement $tpl failed: %s\nQuery was: %s";
+			$tpl .= "\nAdditional: %s" for (2 .. scalar(@args));
+		}
+		unshift @args, $DBI::errstr;
+	}
+	my $msg = errmsg($tpl, @args);
+	my $ekey = 'table ' . $s->[$CONFIG]{name};
+	my $cfg = $s->[$CONFIG];
+	unless(defined $cfg->{LOG_ERROR_CATALOG} and ! $cfg->{LOG_CATALOG}) {
+		logError($msg);
+	}
+	if($cfg->{LOG_ERROR_GLOBAL}) {
+		logGlobal($msg);
+	}
+	if($Vend::admin or ! defined($cfg->{LOG_ERROR_SESSION}) or $cfg->{LOG_ERROR_SESSION}) {
+		$Vend::Session->{errors} = {} unless ref($Vend::Session->{errors}) eq 'HASH';
+		$Vend::Session->{errors}{$ekey} = $msg;
+	}
+	die $msg if $cfg->{DIE_ERROR};
+	return $cfg->{last_error} = $msg;
+}
+
 1;
 
 __END__
