@@ -1,11 +1,12 @@
 # Vend::Table::GDBM - Access an Interchange table stored in a GDBM file
 #
-# $Id: GDBM.pm,v 2.0.2.1 2002/01/24 05:07:03 jon Exp $
+# $Id: GDBM.pm,v 2.12 2004/03/12 12:46:01 racke Exp $
 #
-# Copyright (C) 1996-2002 Red Hat, Inc. <interchange@redhat.com>
+# Copyright (C) 2002-2003 Interchange Development Group
+# Copyright (C) 1996-2002 Red Hat, Inc.
 #
 # This program was originally based on Vend 0.2 and 0.3
-# Copyright 1995 by Andrew M. Wilcox <awilcox@world.std.com>
+# Copyright 1995 by Andrew M. Wilcox <amw@wilcoxsolutions.com>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -29,7 +30,7 @@ use GDBM_File;
 use Vend::Table::Common;
 
 @ISA = qw(Vend::Table::Common);
-$VERSION = substr(q$Revision: 2.0.2.1 $, 10);
+$VERSION = substr(q$Revision: 2.12 $, 10);
 
 sub new {
 	my ($class, $obj) = @_;
@@ -45,23 +46,16 @@ sub create {
 	$File_permission_mode = 0666 unless defined $File_permission_mode;
 	$Fast_write = 1 unless defined $Fast_write;
 
-	die "columns argument $columns is not an array ref\n"
+	die ::errmsg("columns argument %s is not an array ref", $columns)
 		unless CORE::ref($columns) eq 'ARRAY';
 
-	# my $column_file = "$filename.columns";
-	# my @columns = @$columns;
-	# open(COLUMNS, ">$column_file")
-	#    or die "Couldn't create '$column_file': $!";
-	# print COLUMNS join("\t", @columns), "\n";
-	# close(COLUMNS);
-
 	my $column_index = Vend::Table::Common::create_columns($columns, $config);
 
 	my $tie = {};
 	my $flags = GDBM_NEWDB;
 	$flags |= GDBM_FAST if $Fast_write;
 	my $dbm = tie(%$tie, 'GDBM_File', $filename, $flags, $File_permission_mode)
-		or die "Could not create '$filename': $!";
+		or die ::errmsg("%s %s: %s\n", ::errmsg("create"), $filename, $!);
 
 	$tie->{'c'} = join("\t", @$columns);
 
@@ -89,11 +83,14 @@ sub open_table {
 		undef $config->{Transactions};
 		$config->{_Auto_number} = 1 if $config->{AUTO_NUMBER};
 		$flags = GDBM_WRITER;
+		$flags |= GDBM_NOLOCK if $config->{IC_LOCKING};
 		if(! defined $config->{AutoNumberCounter}) {
 			eval {
-				$config->{AutoNumberCounter} = new File::CounterFile
-											"$config->{DIR}/$config->{name}.autonumber",
-											$config->{AUTO_NUMBER} || '00001';
+				my $dot = $config->{HIDE_AUTO_FILES} ? '.' : '';
+				$config->{AutoNumberCounter} = new Vend::CounterFile
+									"$config->{DIR}/$dot$config->{name}.autonumber",
+									$config->{AUTO_NUMBER} || '00001',
+									$config->{AUTO_NUMBER_DATE};
 			};
 			if($@) {
 				::logError("Cannot create AutoNumberCounter: %s", $@);
@@ -105,16 +102,22 @@ sub open_table {
 	my $dbm;
 	my $failed = 0;
 
-	while( $failed < 10 ) {
+	my $retry = $Vend::Cfg->{Limit}{dbm_open_retries} || 10;
+
+	while( $failed < $retry ) {
 		$dbm = tie(%$tie, 'GDBM_File', $filename, $flags, 0777)
 			and undef($failed), last;
 		$failed++;
 		select(undef,undef,undef,$failed * .100);
 	}
 
-	die ::errmsg("Could not tie to '%s': %s", $filename, $!)
-		if $failed;
+	die ::errmsg("%s could not tie to '%s': %s", 'GDBM', $filename, $!)
+		unless $dbm;
+
 	my $columns = [split(/\t/, $tie->{'c'})];
+
+	$config->{VERBATIM_FIELDS} = 1 unless defined $config->{VERBATIM_FIELDS};
+
 	my $column_index = Vend::Table::Common::create_columns($columns, $config);
 
 	my $s = [
@@ -137,11 +140,14 @@ sub open_table {
 *config			= \&Vend::Table::Common::config;
 *delete_record	= \&Vend::Table::Common::delete_record;
 *each_record	= \&Vend::Table::Common::each_record;
+*errstr     	= \&Vend::Table::Common::errstr;
 *field			= \&Vend::Table::Common::field;
 *field_accessor	= \&Vend::Table::Common::field_accessor;
 *field_settor	= \&Vend::Table::Common::field_settor;
 *inc_field		= \&Vend::Table::Common::inc_field;
 *isopen			= \&Vend::Table::Common::isopen;
+*log_error  	= \&Vend::Table::Common::log_error;
+*name			= \&Vend::Table::Common::name;
 *numeric		= \&Vend::Table::Common::numeric;
 *quote			= \&Vend::Table::Common::quote;
 *record_exists	= \&Vend::Table::Common::record_exists;
