@@ -1,10 +1,11 @@
 # Vend::TextSearch - Search indexes with Perl
 #
-# $Id: TextSearch.pm,v 2.0.2.1 2002/01/24 05:07:02 jon Exp $
+# $Id: TextSearch.pm,v 2.12 2003/06/18 17:34:44 jon Exp $
 #
 # Adapted for use with Interchange from Search::TextSearch
 #
-# Copyright (C) 1996-2002 Red Hat, Inc. <interchange@redhat.com>
+# Copyright (C) 2002-2003 Interchange Development Group
+# Copyright (C) 1996-2002 Red Hat, Inc.
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -28,29 +29,29 @@ require Exporter;
 use vars qw(@ISA);
 @ISA = qw(Vend::Search);
 
-$VERSION = substr(q$Revision: 2.0.2.1 $, 10);
+$VERSION = substr(q$Revision: 2.12 $, 10);
 
 use Search::Dict;
 use strict;
 
 sub array {
 	my ($s, $opt) = @_;
-	$s->{mv_list_only} = 1;
-	Vend::Scan::perform_search($opt, undef, $s);
+	$s->{mv_list_only} = 1; # makes perform_search only return results array
+	return Vend::Scan::perform_search($opt, undef, $s);
 }
 
 sub hash {
 	my ($s, $opt) = @_;
 	$s->{mv_return_reference} = 'HASH';
-	$s->{mv_list_only} = 1;
-	Vend::Scan::perform_search($opt, undef, $s);
+	$s->{mv_list_only} = 1; # makes perform_search only return results array
+	return Vend::Scan::perform_search($opt, undef, $s);
 }
 
 sub list {
 	my ($s, $opt) = @_;
-	$s->{mv_list_only} = 1;
 	$s->{mv_return_reference} = 'LIST';
-	Vend::Scan::perform_search($opt, undef, $s);
+	$s->{mv_list_only} = 1; # makes perform_search only return results array
+	return Vend::Scan::perform_search($opt, undef, $s);
 }
 
 my %Default = (
@@ -140,20 +141,10 @@ sub search {
 	$s->{mv_return_delim} = $s->{mv_index_delim}
 		unless defined $s->{mv_return_delim};
 
-	if(ref $s->{mv_range_look}) {
-		unless( scalar(@{$s->{mv_range_look}}) == scalar(@{$s->{mv_range_min}}) and
-				scalar(@{$s->{mv_range_look}}) == scalar(@{$s->{mv_range_max}}) ) {
-			$s->{mv_search_warning}
-				= "Must have min and max values for range -- aborting range look.";
-			undef $s->{mv_range_look};
-		}
-	}
 	@specs = @{$s->{mv_searchspec}};
 
 	@pats = $s->spec_check(@specs);
 
-	return undef if $s->{matches} == -1;
-
 	if ($s->{mv_coordinate}) {
 		undef $f;
 	}
@@ -216,14 +207,15 @@ sub search {
 		my $prospect;
 
 		eval {
-			($limit_sub, $prospect) = $s->get_limit($f);
+			# 1 refers to fact you have to make ref from line
+			($limit_sub, $prospect) = $s->get_limit($f, 1);
 		};
 
 		$@  and  return $s->search_error("Limit subroutine creation: $@");
 
 		$f = $prospect if $prospect;
 
-		eval {($return_sub, $delayed_return) = $s->get_return()};
+		eval {($return_sub, $delayed_return) = $s->get_return( undef, 1 )};
 
 		$@  and  return $s->search_error("Return subroutine creation: $@");
 
@@ -267,59 +259,62 @@ sub search {
 
 		if($s->{mv_dict_end} && defined $limit_sub) {
 			while(<SEARCH>) {
-#::logDebug("$_");
-				last if &$dict_limit($_);
-				next unless ! defined $f or &$f();
-				next unless &$limit_sub($_);
+				last if $dict_limit->($_);
+				next unless $limit_sub->($_);
+				chomp;
 				(push @out, $searchfile and last)
 					if $s->{mv_return_file_name};
-				push @out, &$return_sub($_);
+				push @out, $return_sub->($_);
 			}
 		}
 		elsif($s->{mv_dict_end}) {
 			while(<SEARCH>) {
-				last if &$dict_limit($_);
+				last if $dict_limit->($_);
 				next unless &$f();
+				chomp;
 				(push @out, $searchfile and last)
 					if $s->{mv_return_file_name};
-				push @out, &$return_sub($_);
+				push @out, $return_sub->($_);
 			}
 		}
-		elsif(! defined $f and defined $limit_sub) {
+		elsif($limit_sub) {
+#::logDebug("limit_sub");
 			while(<SEARCH>) {
-				next unless &$limit_sub($_);
+				next unless $limit_sub->($_);
+				chomp;
 				(push @out, $searchfile and last)
 					if $s->{mv_return_file_name};
-				push @out, &$return_sub($_);
+				push @out, $return_sub->($_);
 			}
 		}
-		elsif(defined $limit_sub) {
-			while(<SEARCH>) {
-				next unless &$f();
-				next unless &$limit_sub($_);
-				(push @out, $searchfile and last)
-					if $s->{mv_return_file_name};
-				push @out, &$return_sub($_);
-			}
-		}
-		elsif (!defined $f) {
+		elsif (! $f) {
 			return $s->search_error('No search definition');
 		}
 		else {
+#::logDebug("no limit_sub");
 			while(<SEARCH>) {
 				next unless &$f();
+				chomp;
 				(push @out, $searchfile and last)
 					if $s->{mv_return_file_name};
-				push @out, &$return_sub($_);
+				push @out, $return_sub->($_);
 			}
 		}
 		close SEARCH;
 		$s->restore_specs();
 	}
 
+#::logDebug("before delayed return: self=" . ::Vend::Util::uneval_it({%$s}));
+#::logDebug("before delayed return: out=" . ::Vend::Util::uneval_it(\@out));
+
+	# Search the results and return
+	if($s->{mv_next_search}) {
+		@out = $s->search_reference(\@out);
+#::logDebug("did next_search: " . ::uneval(\@out));
+	}
+
 	$s->{matches} = scalar(@out);
 
-#::logDebug("before delayed return: self=" . ::Vend::Util::uneval_it({%$s}));
 	if($delayed_return and $s->{matches} > 0) {
 		$s->hash_fields($s->{mv_field_names}, qw/mv_sort_field/);
 #::logDebug("after hash fields: self=" . ::Vend::Util::uneval_it({%$s}));
@@ -328,14 +323,20 @@ sub search {
 		@out = map { $delayed_return->($_) } @out;
 	}
 #::logDebug("after delayed return: self=" . ::Vend::Util::uneval_it({%$s}));
+#::logDebug("after delayed return: out=" . ::Vend::Util::uneval_it(\@out));
 
 	if($s->{mv_unique}) {
 		my %seen;
 		@out = grep ! $seen{$_->[0]}++, @out;
-		$s->{matches} = scalar(@out);
 	}
 
-	if ($s->{matches} > $s->{mv_matchlimit}) {
+	if($s->{mv_max_matches} > 0) {
+		splice @out, $s->{mv_max_matches};
+	}
+
+	$s->{matches} = scalar(@out);
+
+	if ($s->{matches} > $s->{mv_matchlimit} and $s->{mv_matchlimit} > 0) {
 		$s->save_more(\@out)
 			or ::logError("Error saving matches: $!");
 		if ($s->{mv_first_match}) {
