@@ -1,8 +1,9 @@
 # Vend::UserDB - Interchange user database functions
 #
-# $Id: UserDB.pm,v 2.0.2.2 2002/01/24 05:07:02 jon Exp $
+# $Id: UserDB.pm,v 2.27.2.1 2004/04/19 14:32:59 mheins Exp $
 #
-# Copyright (C) 1996-2002 Red Hat, Inc. <interchange@redhat.com>
+# Copyright (C) 2002-2003 Interchange Development Group
+# Copyright (C) 1996-2002 Red Hat, Inc.
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -16,7 +17,7 @@
 
 package Vend::UserDB;
 
-$VERSION = substr(q$Revision: 2.0.2.2 $, 10);
+$VERSION = substr(q$Revision: 2.27.2.1 $, 10);
 
 use vars qw!
 	$VERSION
@@ -47,19 +48,8 @@ billing, and preference information.  It allows the user to return to a
 previous session without the requirement for a "cookie" or other persistent
 session information.
 
-It is object-oriented and called via Perl subroutine. The main software 
-is contained in a module, and is called from Interchange with a GlobalSub.
-The GlobalSub would take the form:
-
-	GlobalSub <<EOF
-	sub userdb {
-		my($function, %options) = @_;
-		use Vend::UserDB;
-		$obj = new Vend::User->DB %options;
-		$obj->$function
-			or return $obj->{ERROR};
-		return $obj->{MESSAGE};
-	}
+It is object-oriented and called via the [userdb] usertag, which calls the
+userdb subroutine.
 
 It restores and manipulates the form values normally stored in the user session
 values -- the ones set in forms and read through the C<[value variable]> tags.
@@ -176,25 +166,6 @@ box or in a set of links.
 
 =cut
 
-@S_FIELDS = ( 
-qw!
-	s_nickname
-	name
-	fname
-	lname
-	address
-	address1
-	address2
-	address3
-	city
-	state
-	zip
-	country	
-	phone_day
-	mv_shipmode
-  !
-);
-
 @B_FIELDS = ( 
 qw!
 	b_nickname
@@ -321,6 +292,7 @@ sub new {
 						LAST		=> $options{time_field} || 'mod_time',
 						EXPIRATION	=> $options{expire_field} || 'expiration',
 						OUTBOARD_KEY=> $options{outboard_key_col},
+						GROUPS		=> $options{groups_field}|| 'groups',
 						SUPER		=> $options{super_field}|| 'super',
 						ACL			=> $options{acl}		|| 'acl',
 						FILE_ACL	=> $options{file_acl}	|| 'file_acl',
@@ -334,7 +306,7 @@ sub new {
 
 	return $self if $options{no_open};
 
-	set_db($self) or die ::errmsg("user database %s does not exist.", $self->{DB_ID}) . "\n";
+	set_db($self) or die errmsg("user database %s does not exist.", $self->{DB_ID}) . "\n";
 
 	return $Vend::user_object = $self;
 }
@@ -410,15 +382,26 @@ sub create_db {
 	1;
 }
 
+sub log_either {
+	my $self = shift;
+	my $msg = shift;
+
+	if(! $self->{OPTIONS}{logfile}) {
+		return logError($msg);
+	}
+	$self->log($msg,@_);
+	return;
+}
+
 sub log {
 	my $self = shift;
-	my $time = $self->{OPTIONS}{UNIX_TIME} ?  time() :
+	my $time = $self->{OPTIONS}{unix_time} ?  time() :
 				POSIX::strftime("%Y%m%d%H%M", localtime());
 	my $msg = shift;
 	logData( ($self->{OPTIONS}{logfile} || $Vend::Cfg->{LogFile}),
 						$time,
 						$self->{USERNAME},
-						$CGI::remote_host,
+						$CGI::remote_host || $CGI::remote_addr,
 						$msg,
 						);
 	return;
@@ -428,12 +411,12 @@ sub check_acl {
 	my ($self,%options) = @_;
 
 	if(! defined $self->{PRESENT}{$self->{LOCATION}{ACL}}) {
-		$self->{ERROR} = ::errmsg('No ACL field present.');
+		$self->{ERROR} = errmsg('No ACL field present.');
 		return undef;
 	}
 
 	if(not $options{location}) {
-		$self->{ERROR} = ::errmsg('No location to check.');
+		$self->{ERROR} = errmsg('No location to check.');
 		return undef;
 	}
 
@@ -446,12 +429,12 @@ sub set_acl {
 	my ($self,%options) = @_;
 
 	if(!$self->{PRESENT}{$self->{LOCATION}{ACL}}) {
-		$self->{ERROR} = ::errmsg('No ACL field present.');
+		$self->{ERROR} = errmsg('No ACL field present.');
 		return undef;
 	}
 
 	if(!$options{location}) {
-		$self->{ERROR} = ::errmsg('No location to set.');
+		$self->{ERROR} = errmsg('No location to set.');
 		return undef;
 	}
 
@@ -472,9 +455,10 @@ sub set_acl {
 sub _check_acl {
 	my ($self, $loc, %options) = @_;
 	return undef unless $options{location};
+	$options{mode} = 'r' if ! defined $options{mode};
 	my $acl = $self->{DB}->field( $self->{USERNAME}, $loc);
 	my $f = $ready->reval($acl);
-	return 0 unless exists $f->{$options{location}};
+	return undef unless exists $f->{$options{location}};
 	return 1 if ! $options{mode};
 	if($options{mode} =~ /^\s*expire\b/i) {
 		my $cmp = $f->{$options{location}};
@@ -487,7 +471,7 @@ sub _check_acl {
 sub _set_acl {
 	my ($self, $loc, %options) = @_;
 	return undef unless $self->{OPTIONS}{location};
-	if($options{mode} =~ /^\s*expire\s+(.*)/i) {
+	if($options{mode} =~ /^\s*expires?\s+(.*)/i) {
 		my $secs = Vend::Config::time_to_seconds($1);
 		my $now = time();
 		$options{mode} = $secs + $now;
@@ -500,7 +484,7 @@ sub _set_acl {
 	else {
 		$f->{$options{location}} = $options{mode} || 'rw';
 	}
-	my $return = $self->{DB}->set_field( $self->{USERNAME}, $loc, Vend::Util::uneval_it($f) );
+	my $return = $self->{DB}->set_field( $self->{USERNAME}, $loc, uneval_it($f) );
 	return $return if $options{show};
 	return;
 }
@@ -592,10 +576,20 @@ sub clear_values {
 
 	for(@fields) {
 		if($scratch{$_}) {
-			delete $::Scratch->{$_};
+			if (exists $Vend::Cfg->{ScratchDefault}->{$_}) {
+				$::Scratch->{$_} = $Vend::Cfg->{ScratchDefault}->{$_};
+			}
+			else {
+				delete $::Scratch->{$_};
+			}
 		}
 		else {
-			delete $::Values->{$_};
+			if (exists $Vend::Cfg->{ValuesDefault}->{$_}) {
+				$::Values->{$_} = $Vend::Cfg->{ValuesDefault}->{$_};
+			}
+			else{
+				delete $::Values->{$_};
+			}
 			delete $CGI::values{$_};
 		}
 	}
@@ -609,10 +603,10 @@ sub get_values {
 	@fields = @{ $self->{DB_FIELDS} } unless @fields;
 
 	my $db = $self->{DB}
-		or die ::errmsg("No user database found.");
+		or die errmsg("No user database found.");
 
 	unless ( $db->record_exists($self->{USERNAME}) ) {
-		$self->{ERROR} = ::errmsg("username %s does not exist.", $self->{USERNAME});
+		$self->{ERROR} = errmsg("username %s does not exist.", $self->{USERNAME});
 		return undef;
 	}
 
@@ -669,7 +663,7 @@ sub get_values {
 		my $f = $self->{LOCATION}->{$area};
 		if ($self->{PRESENT}->{$f}) {
 			my $s = $self->get_hash($area);
-			die ::errmsg("Bad structure in %s: %s", $f, $@) if $@;
+			die errmsg("Bad structure in %s: %s", $f, $@) if $@;
 			$::Values->{$f} = join "\n", sort keys %$s;
 		}
 	}
@@ -689,7 +683,7 @@ sub set_values {
 	my $db = $self->{DB};
 
 	unless ( $db->record_exists($self->{USERNAME}) ) {
-		$self->{ERROR} = ::errmsg("username %s does not exist.", $self->{USERNAME});
+		$self->{ERROR} = errmsg("username %s does not exist.", $self->{USERNAME});
 		return undef;
 	}
 	my %scratch;
@@ -733,7 +727,7 @@ sub set_values {
 			push @bvals, $val;
 		}
 		else {
-			::logDebug( ::errmsg(
+			::logDebug( errmsg(
 							"cannot set unknown userdb field $_ to: %s",
 							$_,
 							$val,
@@ -748,12 +742,37 @@ sub set_values {
   };
 
 	if($@) {
-	  my $msg = ::errmsg("error saving values in userdb: %s", $@);
+	  my $msg = errmsg("error saving values in userdb: %s", $@);
 	  $self->{ERROR} = $msg;
-	  ::logError($msg);
+	  logError($msg);
 	  return undef;
 	}
 
+# Changes made to support Accounting Interface.
+
+	if(my $l = $Vend::Cfg->{Accounting}) {
+		my %hashvar;
+		my $indexvar = 0;
+		while ($indexvar <= (scalar @bfields)) {
+			$hashvar{ $bfields[$indexvar] } = $bvals[$indexvar];
+			$indexvar++;
+		};
+		my $obj;
+		my $class = $l->{Class};
+		eval {
+			$obj = $class->new;
+		};
+
+		if($@) {
+			die errmsg(
+				"Failed to save customer data with accounting system %s: %s",
+				$class,
+				$@,
+				);
+		}
+		my $returnval = $obj->save_customer_data($user, \%hashvar);
+	}
+
 	return 1;
 }
 
@@ -797,7 +816,7 @@ sub get_shipping_names {
 	my $self = shift;
 	my $ref = $self->get_hash('SHIPPING');
 	return undef unless ref $ref;
-	$::Values->{$self->{LOCATION}{SHIPPING}} = join "\n", keys %$ref;
+	$::Values->{$self->{LOCATION}{SHIPPING}} = join "\n", sort keys %$ref;
 	return $::Values->{$self->{LOCATION}{SHIPPING}} if $self->{OPTIONS}{show};
 	return '';
 }
@@ -806,7 +825,7 @@ sub get_billing_names {
 	my $self = shift;
 	my $ref = $self->get_hash('BILLING');
 	return undef unless ref $ref;
-	$::Values->{$self->{LOCATION}{BILLING}} = join "\n", keys %$ref;
+	$::Values->{$self->{LOCATION}{BILLING}} = join "\n", sort keys %$ref;
 	return $::Values->{$self->{LOCATION}{BILLING}} if $self->{OPTIONS}{show};
 	return '';
 }
@@ -815,7 +834,7 @@ sub get_preferences_names {
 	my $self = shift;
 	my $ref = $self->get_hash('PREFERENCES');
 	return undef unless ref $ref;
-	$::Values->{$self->{LOCATION}{PREFERENCES}} = join "\n", keys %$ref;
+	$::Values->{$self->{LOCATION}{PREFERENCES}} = join "\n", sort keys %$ref;
 	return $::Values->{$self->{LOCATION}{PREFERENCES}} if $self->{OPTIONS}{show};
 	return '';
 }
@@ -824,7 +843,7 @@ sub get_cart_names {
 	my $self = shift;
 	my $ref = $self->get_hash('CARTS');
 	return undef unless ref $ref;
-	$::Values->{$self->{LOCATION}{CARTS}} = join "\n", keys %$ref;
+	$::Values->{$self->{LOCATION}{CARTS}} = join "\n", sort keys %$ref;
 	return $::Values->{$self->{LOCATION}{CARTS}} if $self->{OPTIONS}{show};
 	return '';
 }
@@ -856,8 +875,8 @@ sub delete_preferences {
 sub delete_nickname {
 	my($self, $name, @fields) = @_;
 
-	die ::errmsg("no fields?") unless @fields;
-	die ::errmsg("no name?") unless $name;
+	die errmsg("no fields?") unless @fields;
+	die errmsg("no name?") unless $name;
 
 	$self->get_hash($name) unless ref $self->{$name};
 
@@ -868,11 +887,11 @@ sub delete_nickname {
 
 	my $field_name = $self->{LOCATION}->{$name};
 	unless($self->{PRESENT}->{$field_name}) {
-		$self->{ERROR} = ::errmsg('%s field not present to set %s', $field_name, $name);
+		$self->{ERROR} = errmsg('%s field not present to set %s', $field_name, $name);
 		return undef;
 	}
 
-	my $s = ::uneval_it($self->{$name});
+	my $s = uneval_it($self->{$name});
 
 	$self->{DB}->set_field( $self->{USERNAME}, $field_name, $s);
 
@@ -882,8 +901,8 @@ sub delete_nickname {
 sub set_hash {
 	my($self, $name, @fields) = @_;
 
-	die ::errmsg("no fields?") unless @fields;
-	die ::errmsg("no name?") unless $name;
+	die errmsg("no fields?") unless @fields;
+	die errmsg("no name?") unless $name;
 
 	$self->get_hash($name) unless ref $self->{$name};
 
@@ -894,7 +913,7 @@ sub set_hash {
 	$::Values->{$nick_field} = $nick;
 	$CGI::values{$nick_field} = $nick if $self->{CGI};
 
-	die ::errmsg("no nickname?") unless $nick;
+	die errmsg("no nickname?") unless $nick;
 
 	$self->{$name}{$nick} = {} unless $self->{OPTIONS}{keep}
 							   and    defined $self->{$name}{$nick};
@@ -906,11 +925,11 @@ sub set_hash {
 
 	my $field_name = $self->{LOCATION}->{$name};
 	unless($self->{PRESENT}->{$field_name}) {
-		$self->{ERROR} = ::errmsg('%s field not present to set %s', $field_name, $name);
+		$self->{ERROR} = errmsg('%s field not present to set %s', $field_name, $name);
 		return undef;
 	}
 
-	my $s = ::uneval_it($self->{$name});
+	my $s = uneval_it($self->{$name});
 
 	$self->{DB}->set_field( $self->{USERNAME}, $field_name, $s);
 
@@ -924,21 +943,21 @@ sub get_hash {
 	my ($nick, $s);
 
 	eval {
-		die ::errmsg("no name?")					unless $name;
-		die ::errmsg("%s field not present to get %s", $field_name, $name) . "\n"
+		die errmsg("no name?")					unless $name;
+		die errmsg("%s field not present to get %s", $field_name, $name) . "\n"
 										unless $self->{PRESENT}->{$field_name};
 
 		$s = $self->{DB}->field( $self->{USERNAME}, $field_name);
 
 		if($s) {
 			$self->{$name} = $ready->reval($s);
-			die ::errmsg("Bad structure in %s: %s", $field_name, $@) if $@;
+			die errmsg("Bad structure in %s: %s", $field_name, $@) if $@;
 		}
 		else {
 			$self->{$name} = {};
 		}
 
-		die ::errmsg("eval failed?") . "\n"		unless ref $self->{$name};
+		die errmsg("eval failed?") . "\n"		unless ref $self->{$name};
 	};
 
 	if($@) {
@@ -955,7 +974,7 @@ sub get_hash {
 		$nick =~ s/[\0\s]+.*//;
 		$::Values->{$nick_field} = $nick;
 		$CGI::values{$nick_field} = $nick if $self->{CGI};
-		die ::errmsg("no nickname?") unless $nick;
+		die errmsg("no nickname?") unless $nick;
 	};
 
 	if($@) {
@@ -987,22 +1006,20 @@ sub login {
 
 	# Show this generic error message on login page to avoid
 	# helping would-be intruders
-	my $stock_error = ::errmsg("Invalid user name or password.");
+	my $stock_error = errmsg("Invalid user name or password.");
 	
 	eval {
 		unless($self) {
 			$self = new Vend::UserDB %options;
 		}
-		if(
-			$Vend::Cfg->{CookieLogin}
-			and (! $Vend::Cfg->{DifferentSecure} || ! $CGI::secure)
-			)
-		{
+
+		if($Vend::Cfg->{CookieLogin}) {
 			$self->{USERNAME} = Vend::Util::read_cookie('MV_USERNAME')
 				if ! $self->{USERNAME};
 			$self->{PASSWORD} = Vend::Util::read_cookie('MV_PASSWORD')
 				if ! $self->{PASSWORD};
 		}
+
 		if ($self->{OPTIONS}{ignore_case}) {
 			$self->{PASSWORD} = lc $self->{PASSWORD};
 			$self->{USERNAME} = lc $self->{USERNAME};
@@ -1011,29 +1028,29 @@ sub login {
 		# We specifically check for login attempts with group names to see if
 		# anyone is trying to exploit a former vulnerability in the demo catalog.
 		if ($self->{USERNAME} =~ /^:/) {
-			logError("Denied attempted login with group name '%s'",
-				$self->{USERNAME});
+			$self->log_either(errmsg("Denied attempted login with group name '%s'",
+				$self->{USERNAME}));
 			die $stock_error, "\n";
 		}
 
 		# Username must be long enough
 		if (length($self->{USERNAME}) < $self->{USERMINLEN}) {
-			logError("Denied attempted login for user name '%s'; must have at least %s characters",
-				$self->{USERNAME}, $self->{USERMINLEN});
+			$self->log_either(errmsg("Denied attempted login for user name '%s'; must have at least %s characters",
+				$self->{USERNAME}, $self->{USERMINLEN}));
 			die $stock_error, "\n";
 		}
 
 		# Username must contain only valid characters
 		if ($self->{USERNAME} !~ m{^$USERNAME_GOOD_CHARS+$}) {
-			logError("Denied attempted login for user name '%s' with illegal characters",
-				$self->{USERNAME});
+			$self->log_either(errmsg("Denied attempted login for user name '%s' with illegal characters",
+				$self->{USERNAME}));
 			die $stock_error, "\n";
 		}
 
 		# Fail if password is too short
 		if (length($self->{PASSWORD}) < $self->{PASSMINLEN}) {
-			logError("Denied attempted login with user name '%s' and password less than %s characters",
-				$self->{USERNAME}, $self->{PASSMINLEN});
+			$self->log_either(errmsg("Denied attempted login with user name '%s' and password less than %s characters",
+				$self->{USERNAME}, $self->{PASSMINLEN}));
 			die $stock_error, "\n";
 		}
 
@@ -1045,47 +1062,88 @@ sub login {
 				my ($adminuser, $adminpass) = split /[\s:]+/, $pwinfo;
 				last ADMINUSER unless $adminuser eq $self->{USERNAME};
 				unless ($adminpass) {
-					logError("Refusing to use AdminUser variable with user '%s' and empty password", $adminuser);
+					$self->log_either(errmsg("Refusing to use AdminUser variable with user '%s' and empty password", $adminuser));
 					last ADMINUSER;
 				}
-				my $test = $Global::Variable->{MV_NO_CRYPT}
-						 ? $self->{PASSWORD}
-						 : crypt($self->{PASSWORD}, $adminpass);
+				my $test;
+				if($Global::Variable->{MV_NO_CRYPT}) {
+					 $test = $self->{PASSWORD}
+				}
+				elsif ($self->{OPTIONS}{md5}) {
+					 $test = generate_key($self->{PASSWORD});
+				}
+				else {
+					 $test = crypt($self->{PASSWORD}, $adminpass);
+				}
 				if ($test eq $adminpass) {
 					$user_data = {};
 					$Vend::admin = $Vend::superuser = 1;
-					logError("Successful superuser login by AdminUser '%s'", $adminuser);
+					$self->log_either( errmsg("Successful superuser login by AdminUser '%s'", $adminuser));
 				} else {
-					logError("Password given with user name '%s' didn't match AdminUser password", $adminuser);
+					$self->log_either(errmsg("Password given with user name '%s' didn't match AdminUser password", $adminuser));
 				}
 			}
 		}
 
+		my $udb = $self->{DB};
+		my $foreign = $self->{OPTIONS}{indirect_login};
+
+		if($foreign) {
+			my $uname = ($self->{PASSED_USERNAME} ||= $self->{USERNAME});
+			my $ufield = $self->{LOCATION}{USERNAME};
+			$uname = $udb->quote($uname);
+			my $q = "select $ufield from $self->{DB_ID} where $foreign = $uname";
+#::logDebug("indirect login query: $q");
+			my $ary = $udb->query($q)
+				or do {
+					my $msg = errmsg( "Database access error for query: %s", $q);
+					die "$msg\n";
+				};
+			@$ary == 1
+				or do {
+					$self->log_either(errmsg(
+						"Denied attempted login with nonexistent (indirect from %s) user name %s",
+						$foreign,
+						$uname,
+						$self->{USERNAME},
+					));
+					die $stock_error, "\n";
+				};
+			$self->{USERNAME} = $ary->[0][0];
+		}
+
 		# If not superuser, an entry must exist in access database
 		unless ($Vend::superuser) {
-			unless ($self->{DB}->record_exists($self->{USERNAME})) {
-				logError("Denied attempted login with nonexistent user name '%s'",
-					$self->{USERNAME});
+			unless ($udb->record_exists($self->{USERNAME})) {
+				$self->log_either(errmsg("Denied attempted login with nonexistent user name '%s'",
+					$self->{USERNAME}));
 				die $stock_error, "\n";
 			}
-			unless ($user_data = $self->{DB}->row_hash($self->{USERNAME})) {
-				logError("Login denied after failed fetch of user data for user '%s'",
-					$self->{USERNAME});
+			unless ($user_data = $udb->row_hash($self->{USERNAME})) {
+				$self->log_either(errmsg("Login denied after failed fetch of user data for user '%s'",
+					$self->{USERNAME}));
 				die $stock_error, "\n";
 			}
 			my $db_pass = $user_data->{ $self->{LOCATION}{PASSWORD} };
 			unless ($db_pass) {
-				logError("Refusing to use blank password from '%s' database for user '%s'", $self->{DB_ID}, $self->{USERNAME});
+				$self->log_either(errmsg("Refusing to use blank password from '%s' database for user '%s'", $self->{DB_ID}, $self->{USERNAME}));
 				die $stock_error, "\n";
 			}
 			$pw = $self->{PASSWORD};
-			$self->{PASSWORD} = crypt($pw, $db_pass) if $self->{CRYPT};
+			if($self->{CRYPT}) {
+				if($self->{OPTIONS}{md5}) {
+					$self->{PASSWORD} = generate_key($pw);
+				}
+				else {
+					$self->{PASSWORD} = crypt($pw, $db_pass);
+				}
+			}
 			unless ($self->{PASSWORD} eq $db_pass) {
-				logError("Denied attempted login by user '%s' with incorrect password",
-					$self->{USERNAME});
+				$self->log_either(errmsg("Denied attempted login by user '%s' with incorrect password",
+					$self->{USERNAME}));
 				die $stock_error, "\n";
 			}
-			logError("Successful login by user '%s'", $self->{USERNAME});
+			$self->log_either(errmsg("Successful login by user '%s'", $self->{USERNAME}));
 		}
 
 		if($self->{PRESENT}->{ $self->{LOCATION}{EXPIRATION} } ) {
@@ -1093,30 +1151,46 @@ sub login {
 			my $cmp = $now;
 			$cmp = POSIX::strftime("%Y%m%d%H%M", localtime($now))
 				unless $self->{OPTIONS}->{unix_time};
-			my $exp = $self->{DB}->field(
+			my $exp = $udb->field(
 						$self->{USERNAME},
 						$self->{LOCATION}{EXPIRATION},
 						);
-			die ::errmsg("Expiration date not set.") . "\n"
+			die errmsg("Expiration date not set.") . "\n"
 				if ! $exp and $self->{EMPTY_EXPIRE_FATAL};
 			if($exp and $exp < $cmp) {
-				die ::errmsg("Expired %s.", $exp) . "\n";
+				die errmsg("Expired %s.", $exp) . "\n";
 			}
 		}
 
+		if($self->{PRESENT}->{ $self->{LOCATION}{GROUPS} } ) {
+			$Vend::groups
+			= $Vend::Session->{groups}
+			= $udb->field(
+						$self->{USERNAME},
+						$self->{LOCATION}{GROUPS},
+						);
+		}
+
 		username_cookies($self->{USERNAME}, $pw) 
 			if $Vend::Cfg->{CookieLogin};
 
 		if ($self->{LOCATION}{LAST} ne 'none') {
+			my $now = time();
+			my $login_time;
+			unless($self->{OPTIONS}{null_time}) {
+				$login_time = $self->{OPTIONS}{iso_time}
+						? POSIX::strftime("%Y-%m-%d %H:%M:%S", localtime($now))
+						: $now;
+			}
 			eval {
-				$self->{DB}->set_field( $self->{USERNAME},
-										$self->{LOCATION}{LAST},
-										time()
-									  );
+				$udb->set_field( $self->{USERNAME},
+									$self->{LOCATION}{LAST},
+									$login_time
+									);
 			};
 			if ($@) {
-				my $msg = ::errmsg("Failed to record timestamp in UserDB: %s", $@);
-				::logError($msg);
+				my $msg = errmsg("Failed to record timestamp in UserDB: %s", $@);
+				logError($msg);
 				die $msg, "\n";
 			}
 		}
@@ -1132,7 +1206,7 @@ sub login {
 			$self->{ERROR} = $@;
 		}
 		else {
-			::logError( "Vend::UserDB error: %s\n", $@ );
+			logError( "Vend::UserDB error: %s\n", $@ );
 		}
 		return undef;
 	}
@@ -1149,7 +1223,7 @@ sub login {
 								);
 	}
 
-	$Vend::Session->{login_table} = $self->{DB_ID};
+	$Vend::login_table = $Vend::Session->{login_table} = $self->{DB_ID};
 	$Vend::username = $Vend::Session->{username} = $self->{USERNAME};
 	$Vend::Session->{logged_in} = 1;
 	
@@ -1163,6 +1237,50 @@ sub scrub {
 	}
 }
 
+sub logout {
+	my $self = shift or return undef;
+	scrub();
+
+	my $opt = $self->{OPTIONS};
+
+	if( is_yes($opt->{clear}) ) {
+		$self->clear_values();
+	}
+
+	Vend::Interpolate::tag_profile("", { restore => 1 });
+	no strict 'refs';
+
+	my @dels = qw/
+					groups
+					admin
+					superuser
+					login_table
+					username
+					logged_in
+				/;
+
+	for(@dels) {
+		delete $Vend::Session->{$_};
+		undef ${"Vend::$_"};
+	}
+
+	delete $CGI::values{mv_username};
+	delete $::Values->{mv_username};
+	$self->log('logout') if $opt->{log};
+	$self->{MESSAGE} = errmsg('Logged out.');
+	if ($opt->{clear_cookie}) {
+		my @cookies = split /[\s,\0]+/, $opt->{clear_cookie};
+		my $exp = time() + $Vend::Cfg->{SaveExpire};
+		for(@cookies) {
+			Vend::Util::set_cookie($_, '', $exp);
+		}
+	}
+	if ($opt->{clear_session}) {
+		Vend::Session::init_session();
+	}
+	return 1;
+}
+
 sub change_pass {
 
 	my ($self, $original_self);
@@ -1172,6 +1290,13 @@ sub change_pass {
 
 	my(%options) = @_;
 	
+	if ($self->{OPTIONS}{ignore_case}) {
+	   $self->{USERNAME} = lc $self->{USERNAME};
+	   $self->{OLDPASS} = lc $self->{OLDPASS};
+	   $self->{PASSWORD} = lc $self->{PASSWORD};
+	   $self->{VERIFY} = lc $self->{VERIFY};
+	}
+
 	eval {
 		my $super = $Vend::superuser || (
 			$Vend::admin and
@@ -1190,9 +1315,9 @@ sub change_pass {
 					undef $self;
 				}
 			} else {
-				::logError("Unprivileged user '%s' attempted to change password of user '%s'",
+				errmsg("Unprivileged user '%s' attempted to change password of user '%s'",
 					$Vend::username, $self->{USERNAME}) if $options{log};
-				die ::errmsg("You are not allowed to change another user's password.") . "\n";
+				die errmsg("You are not allowed to change another user's password.") . "\n";
 			}
 		}
 
@@ -1200,30 +1325,41 @@ sub change_pass {
 			$self = new Vend::UserDB %options;
 		}
 
-		die ::errmsg("Bad object.") unless defined $self;
+		die errmsg("Bad object.") unless defined $self;
 
-		die ::errmsg("'%s' not a user.", $self->{USERNAME}) . "\n"
+		die errmsg("'%s' not a user.", $self->{USERNAME}) . "\n"
 			unless $self->{DB}->record_exists($self->{USERNAME});
 
 		unless ($super and $self->{USERNAME} ne $Vend::username) {
 			my $db_pass = $self->{DB}->field($self->{USERNAME}, $self->{LOCATION}{PASSWORD});
-			$self->{OLDPASS} = crypt($self->{OLDPASS}, $db_pass)
-				if $self->{CRYPT};
-			die ::errmsg("Must have old password.") . "\n"
+			if($self->{CRYPT}) {
+				if($self->{OPTIONS}{md5}) {
+					$self->{OLDPASS} = generate_key($self->{OLDPASS});
+				}
+				else {
+					$self->{OLDPASS} = crypt($self->{OLDPASS}, $db_pass);
+				}
+			}
+			die errmsg("Must have old password.") . "\n"
 				if $self->{OLDPASS} ne $db_pass;
 		}
 
-		die ::errmsg("Must enter at least %s characters for password.",
+		die errmsg("Must enter at least %s characters for password.",
 			$self->{PASSMINLEN}) . "\n"
 			if length($self->{PASSWORD}) < $self->{PASSMINLEN}; 
-		die ::errmsg("Password and check value don't match.") . "\n"
+		die errmsg("Password and check value don't match.") . "\n"
 			unless $self->{PASSWORD} eq $self->{VERIFY};
 
 		if($self->{CRYPT}) {
-			$self->{PASSWORD} = crypt(
-									$self->{PASSWORD},
-									Vend::Util::random_string(2)
-								);
+				if($self->{OPTIONS}{md5}) {
+					$self->{PASSWORD} = generate_key($self->{PASSWORD});
+				}
+				else {
+					$self->{PASSWORD} = crypt(
+											$self->{PASSWORD},
+											Vend::Util::random_string(2)
+										);
+				}
 		}
 		
 		my $pass = $self->{DB}->set_field(
@@ -1231,8 +1367,8 @@ sub change_pass {
 						$self->{LOCATION}{PASSWORD},
 						$self->{PASSWORD}
 						);
-		die ::errmsg("Database access error.") . "\n" unless defined $pass;
-		$self->log(::errmsg('change password')) if $options{'log'};
+		die errmsg("Database access error.") . "\n" unless defined $pass;
+		$self->log(errmsg('change password')) if $options{'log'};
 	};
 
 	scrub();
@@ -1242,7 +1378,7 @@ sub change_pass {
 	if($@) {
 		if(defined $self) {
 			$self->{ERROR} = $@;
-			$self->log(::errmsg('change password failed')) if $options{'log'};
+			$self->log(errmsg('change password failed')) if $options{'log'};
 		}
 		else {
 			logError( "Vend::UserDB error: %s", $@ );
@@ -1254,12 +1390,41 @@ sub change_pass {
 }
 
 sub assign_username {
-        my $self = shift;
-        my $file = shift || $self->{OPTIONS}{'counter'};
-        my $start = $self->{OPTIONS}{username} || 'U00000';
-        $file = './etc/username.counter' if ! $file;
-        my $ctr = File::CounterFile->new($file, $start);
-        return $ctr->inc();
+	my $self = shift;
+	my $file = shift || $self->{OPTIONS}{counter};
+	my $start = $self->{OPTIONS}{username} || 'U00000';
+	$file = './etc/username.counter' if ! $file;
+
+	my $o = { start => $start, sql => $self->{OPTIONS}{sql_counter} };
+
+	my $custno;
+
+	if(my $l = $Vend::Cfg->{Accounting}) {
+
+		my $class = $l->{Class};
+
+		my $assign = defined $l->{assign_username} ? $l->{assign_username} : 1;
+
+		if($assign) {
+#::logDebug("Accounting class is $class");
+		my $obj;
+		eval {
+				$obj = $class->new;
+		};
+#::logDebug("Accounting object is $obj");
+
+		if($@) {
+			die errmsg(
+				"Failed to assign new customer number with accounting system %s",
+				$class,
+				);
+		}
+		$custno = $obj->assign_customer_number();
+		}
+#::logDebug("assigned new customer number $custno");
+	}
+
+	return $custno || Vend::Interpolate::tag_counter($file, $o);
 }
 
 sub new_account {
@@ -1275,17 +1440,20 @@ sub new_account {
 		unless($self) {
 			$self = new Vend::UserDB %options;
 		}
-		die ::errmsg("Bad object.") . "\n" unless defined $self;
 
-		die ::errmsg("Already logged in. Log out first.") . "\n"
-			if $Vend::Session->{logged_in};
-		die ::errmsg("Sorry, reserved user name.") . "\n"
+		delete $Vend::Session->{auto_created_user};
+
+		die errmsg("Bad object.") . "\n" unless defined $self;
+
+		die errmsg("Already logged in. Log out first.") . "\n"
+			if $Vend::Session->{logged_in} and ! $options{no_login};
+		die errmsg("Sorry, reserved user name.") . "\n"
 			if $self->{OPTIONS}{username_mask} 
 				and $self->{USERNAME} =~ m!$self->{OPTIONS}{username_mask}!;
-		die ::errmsg("Must enter at least %s characters for password.",
+		die errmsg("Must enter at least %s characters for password.",
 			$self->{PASSMINLEN}) . "\n"
 			if length($self->{PASSWORD}) < $self->{PASSMINLEN};
-		die ::errmsg("Password and check value don't match.") . "\n"
+		die errmsg("Password and check value don't match.") . "\n"
 			unless $self->{PASSWORD} eq $self->{VERIFY};
 
 		if ($self->{OPTIONS}{ignore_case}) {
@@ -1296,41 +1464,84 @@ sub new_account {
 		my $pw = $self->{PASSWORD};
 		if($self->{CRYPT}) {
 			eval {
-				$pw = crypt(
-										$pw,
-										Vend::Util::random_string(2)
-									);
+				if($self->{OPTIONS}{md5}) {
+					$pw = generate_key($pw);
+				}
+				else {
+					$pw = crypt( $pw, Vend::Util::random_string(2));
+				}
 			};
 		}
 	
+		my $udb = $self->{DB};
+
 		if($self->{OPTIONS}{assign_username}) {
+			$self->{PASSED_USERNAME} = $self->{USERNAME};
 			$self->{USERNAME} = $self->assign_username();
 			$self->{USERNAME} = lc $self->{USERNAME}
 				if $self->{OPTIONS}{ignore_case};
 		}
-		die ::errmsg("Can't have '%s' as username; it contains illegal characters.",
+		die errmsg("Can't have '%s' as username; it contains illegal characters.",
 			$self->{USERNAME}) . "\n"
 			if $self->{USERNAME} !~ m{^$USERNAME_GOOD_CHARS+$};
-		die ::errmsg("Must have at least %s characters in username.",
+		die errmsg("Must have at least %s characters in username.",
 			$self->{USERMINLEN}) . "\n"
 			if length($self->{USERNAME}) < $self->{USERMINLEN};
-		if ($self->{DB}->record_exists($self->{USERNAME})) {
-			die ::errmsg("Username already exists.") . "\n"
+
+		# Here we put the username in a non-primary key field, checking
+		# for existence
+		my $foreign = $self->{OPTIONS}{indirect_login};
+		if ($foreign) {
+			my $uname = ($self->{PASSED_USERNAME} ||= $self->{USERNAME});
+			$uname = $udb->quote($uname);
+			my $q = "select username from $self->{DB_ID} where $foreign = $uname";
+			my $ary = $udb->query($q)
+				or do {
+					my $msg = errmsg( "Database access error for query: %s", $q);
+					die "$msg\n";
+				};
+			@$ary == 0
+				or do {
+					my $msg = errmsg( "Username already exists (indirect).");
+					die "$msg\n";
+				};
 		}
-		my $pass = $self->{DB}->set_field(
+
+		if ($udb->record_exists($self->{USERNAME})) {
+			die errmsg("Username already exists.") . "\n";
+		}
+
+		my $pass = $udb->set_field(
 						$self->{USERNAME},
 						$self->{LOCATION}{PASSWORD},
-						$pw
+						$pw,
 						);
-		die ::errmsg("Database access error.") . "\n" unless defined $pass;
+		die errmsg("Database access error.") . "\n" unless defined $pass;
 
-		username_cookies($self->{USERNAME}, $pw) 
-			if $Vend::Cfg->{CookieLogin};
+		if($foreign) {
+			 $udb->set_field(
+						$self->{USERNAME},
+						$foreign,
+						$self->{PASSED_USERNAME},
+						)
+				or die errmsg("Database access error.");
+		}
 
-		$self->log('new account') if $options{'log'};
-		$self->set_values();
-		$self->login()
-			or die ::errmsg("Cannot log in after new account creation: %s", $self->{ERROR});
+		if($options{no_login}) {
+			$Vend::Session->{auto_created_user} = $self->{USERNAME};
+		}
+		else {
+			username_cookies($self->{USERNAME}, $pw) 
+				if $Vend::Cfg->{CookieLogin};
+
+			$self->log('new account') if $options{'log'};
+			$self->set_values() unless $self->{OPTIONS}{no_set};
+			$self->login()
+				or die errmsg(
+							"Cannot log in after new account creation: %s",
+							$self->{ERROR},
+						);
+		}
 	};
 
 	scrub();
@@ -1350,7 +1561,6 @@ sub new_account {
 
 sub username_cookies {
 		my ($user, $pw) = @_;
-		return if $Vend::Cfg->{DifferentSecure} && $CGI::secure;
 		return unless
 			 $CGI::values{mv_cookie_password}		or
 			 $CGI::values{mv_cookie_username}		or
@@ -1373,35 +1583,38 @@ sub get_cart {
 
 	my $from = $self->{NICKNAME};
 	my $to;
-	if ($options{target}) {
-		$to = ($::Carts->{$options{target}} ||= []);
+
+	my $opt = $self->{OPTIONS};
+
+	if ($opt->{target}) {
+		$to = ($::Carts->{$opt->{target}} ||= []);
 	}
 	else {
 		$to = $Vend::Items;
 	}
 
-#::logDebug ("to=$to nick=$options{target} from=$from cart=" . ::uneval_it($from));
+#::logDebug ("to=$to nick=$opt->{target} from=$from cart=" . ::uneval_it($from));
 
 	my $field_name = $self->{LOCATION}->{CARTS};
 	my $cart = [];
 
 	eval {
-		die ::errmsg("no from cart name?")				unless $from;
-		die ::errmsg("%s field not present to get %s", $field_name, $from) . "\n"
+		die errmsg("no from cart name?")				unless $from;
+		die errmsg("%s field not present to get %s", $field_name, $from) . "\n"
 										unless $self->{PRESENT}->{$field_name};
 
 		my $s = $self->{DB}->field( $self->{USERNAME}, $field_name);
 
-		die ::errmsg("no saved carts.") . "\n" unless $s;
+		die errmsg("no saved carts.") . "\n" unless $s;
 
 		my @carts = split /\0/, $from;
 		my $d = $ready->reval($s);
 #::logDebug ("saved carts=" . ::uneval_it($d));
 
-		die ::errmsg("eval failed?")				unless ref $d;
+		die errmsg("eval failed?")				unless ref $d;
 
 		for(@carts) {
-			die ::errmsg("source cart '%s' does not exist.", $from) . "\n" unless ref $d->{$_};
+			die errmsg("source cart '%s' does not exist.", $from) . "\n" unless ref $d->{$_};
 			push @$cart, @{$d->{$_}};
 		}
 
@@ -1411,9 +1624,9 @@ sub get_cart {
 		$self->{ERROR} = $@;
 		return undef;
 	}
-#::logDebug ("to=$to nick=$options{target} from=$from cart=" . ::uneval_it($cart));
+#::logDebug ("to=$to nick=$opt->{target} from=$from cart=" . ::uneval_it($cart));
 
-	if($options{merge}) {
+	if($opt->{merge}) {
 		$to = [] unless ref $to;
 		push(@$to,@$cart);
 	} else {
@@ -1426,8 +1639,11 @@ sub set_cart {
 
 	my $from;
 	my $to   = $self->{NICKNAME};
-	if ($self->{OPTIONS}{source}) {
-		$from = $::Carts->{$self->{OPTIONS}{source}} || [];
+
+	my $opt = $self->{OPTIONS};
+
+	if ($opt->{source}) {
+		$from = $::Carts->{$opt->{source}} || [];
 	}
 	else {
 		$from = $Vend::Items;
@@ -1437,17 +1653,17 @@ sub set_cart {
 	my ($cart,$s,$d);
 
 	eval {
-		die ::errmsg("no to cart name?") . "\n"					unless $to;
-		die ::errmsg('%s field not present to set %s', $field_name, $from) . "\n"
+		die errmsg("no to cart name?") . "\n"					unless $to;
+		die errmsg('%s field not present to set %s', $field_name, $from) . "\n"
 										unless $self->{PRESENT}->{$field_name};
 
 		$d = $ready->reval( $self->{DB}->field( $self->{USERNAME}, $field_name) );
 
 		$d = {} unless $d;
 
-		die ::errmsg("eval failed?")				unless ref $d;
+		die errmsg("eval failed?")				unless ref $d;
 
-		if($options{merge}) {
+		if($opt->{merge}) {
 			$d->{$to} = [] unless ref $d->{$to};
 			push(@{$d->{$to}}, @{$from});
 		}
@@ -1475,8 +1691,6 @@ sub userdb {
 
 	my %options;
 
-#::logDebug("Called userdb function=$function opt=$opt " .  Data::Dumper::Dumper($opt));
-
 	if(ref $opt) {
 		%options = %$opt;
 	}
@@ -1487,98 +1701,73 @@ sub userdb {
 	my $status = 1;
 	my $user;
 
+	my $module = $Vend::Cfg->{UserControl} ? 'Vend::UserControl' : 'Vend::UserDB';
+
 	if($function eq 'login') {
 		$Vend::Session->{logged_in} = 0;
 		delete $Vend::Session->{username};
+		delete $Vend::Session->{groups};
 		undef $Vend::username;
+		undef $Vend::groups;
 		undef $Vend::admin;
-		$user = new Vend::UserDB %options;
+		$user = $module->new(%options);
 		unless (defined $user) {
-			$Vend::Session->{failure} = ::errmsg("Unable to access user database.");
+			$Vend::Session->{failure} = errmsg("Unable to access user database.");
 			return undef;
 		}
 		if ($status = $user->login(%options) ) {
-			if(
-				! $Vend::Cfg->{AdminUserDB} or
-				$Vend::Cfg->{AdminUserDB}{$user->{PROFILE}}
-				)
-			{
+			if( $Vend::ReadOnlyCfg->{AdminUserDB}{$user->{PROFILE}} ) {
 				$Vend::admin = 1;
 			}
-			undef $Vend::Cookie
-				unless $Vend::Cfg->{StaticLogged};
+			undef $Vend::Cookie;
 			::update_user();
 		}
 	}
 	elsif($function eq 'new_account') {
-		$user = new Vend::UserDB %options;
+		$user = $module->new(%options);
 		unless (defined $user) {
-			$Vend::Session->{failure} = ::errmsg("Unable to access user database.");
+			$Vend::Session->{failure} = errmsg("Unable to access user database.");
 			return undef;
 		}
-		if($status = $user->new_account(%options)) {
+		$status = $user->new_account(%options);
+		if($status and ! $options{no_login}) {
 			$Vend::Session->{logged_in} = 1;
 			$Vend::Session->{username} = $user->{USERNAME};
-			undef $Vend::Cookie
-				unless $Vend::Cfg->{StaticLogged};
+			undef $Vend::Cookie;
 		}
 	}
 	elsif($function eq 'logout') {
-		$user = new Vend::UserDB %options;
-		scrub();
-		unless (defined $user) {
-			$Vend::Session->{failure} = ::errmsg("Unable to access user database.");
-			return undef;
-		}
-		if( is_yes($options{clear}) ) {
-			$user->clear_values();
-		}
-		Vend::Interpolate::tag_profile("", { restore => 1 });
-		delete $Vend::Session->{logged_in};
-		undef $Vend::admin;
-		undef $Vend::superuser;
-		delete $Vend::Session->{login_table};
-		delete $Vend::Session->{username};
-		delete $CGI::values{mv_username};
-		undef $Vend::username;
-		delete $::Values->{mv_username};
-		$user->log('logout') if $options{'log'};
-		$user->{MESSAGE} = ::errmsg('Logged out.');
-		if ($user->{OPTIONS}{clear_cookie}) {
-			my @cookies = split /[\s,\0]+/, $user->{OPTIONS}{clear_cookie};
-			my $exp = time() + $Vend::Cfg->{SaveExpire};
-			for(@cookies) {
-				Vend::Util::set_cookie($_, '', $exp);
-			}
-		}
-		if ($user->{OPTIONS}{clear_session}) {
-			Vend::Session::init_session();
-		}
+		$user = $module->new(%options)
+			or do {
+				$Vend::Session->{failure} = errmsg("Unable to create user object.");
+				return undef;
+			};
+		$user->logout();
 	}
 	elsif (! $Vend::Session->{logged_in}) {
-		$Vend::Session->{failure} = ::errmsg("Not logged in.");
+		$Vend::Session->{failure} = errmsg("Not logged in.");
 		return undef;
 	}
 	elsif($function eq 'save') {
-		$user = new Vend::UserDB %options;
+		$user = $module->new(%options);
 		unless (defined $user) {
-			$Vend::Session->{failure} = ::errmsg("Unable to access user database.");
+			$Vend::Session->{failure} = errmsg("Unable to access user database.");
 			return undef;
 		}
 		$status = $user->set_values();
 	}
 	elsif($function eq 'load') {
-		$user = new Vend::UserDB %options;
+		$user = $module->new(%options);
 		unless (defined $user) {
-			$Vend::Session->{failure} = ::errmsg("Unable to access user database.");
+			$Vend::Session->{failure} = errmsg("Unable to access user database.");
 			return undef;
 		}
 		$status = $user->get_values();
 	}
 	else {
-		$user = new Vend::UserDB %options;
+		$user = $module->new(%options);
 		unless (defined $user) {
-			$Vend::Session->{failure} = ::errmsg("Unable to access user database.");
+			$Vend::Session->{failure} = errmsg("Unable to access user database.");
 			return undef;
 		}
 		eval {
