@@ -1,11 +1,12 @@
 # Vend::Util - Interchange utility functions
 #
-# $Id: Util.pm,v 2.1.2.6 2002/03/01 20:10:53 racke Exp $
+# $Id: Util.pm,v 2.71 2004/03/07 03:14:41 mheins Exp $
 # 
-# Copyright (C) 1996-2002 Red Hat, Inc. <interchange@redhat.com>
+# Copyright (C) 2002-2003 Interchange Development Group
+# Copyright (C) 1996-2002 Red Hat, Inc.
 #
 # This program was originally based on Vend 0.2 and 0.3
-# Copyright 1995 by Andrew M. Wilcox <awilcox@world.std.com>
+# Copyright 1995 by Andrew M. Wilcox <amw@wilcoxsolutions.com>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -32,23 +33,30 @@ require Exporter;
 	check_security
 	copyref
 	currency
+	dbref
 	dump_structure
 	errmsg
 	escape_chars
 	evalr
+	dotted_hash
 	file_modification_time
 	file_name_is_absolute
 	find_special_page
 	format_log_msg
 	generate_key
 	get_option_hash
+	hash_string
+	hexify
+	is_hash
 	is_no
 	is_yes
+	l
 	lockfile
 	logData
 	logDebug
 	logError
 	logGlobal
+	logOnce
 	logtime
 	random_string
 	readfile
@@ -64,6 +72,7 @@ require Exporter;
 	uneval
 	uneval_it
 	uneval_fast
+	unhexify
 	unlockfile
 	vendUrl
 );
@@ -72,15 +81,13 @@ use strict;
 use Config;
 use Fcntl;
 use Errno;
+use Text::ParseWords;
+require HTML::Entities;
+use Safe;
+use Vend::File;
 use subs qw(logError logGlobal);
 use vars qw($VERSION @EXPORT @EXPORT_OK);
-$VERSION = substr(q$Revision: 2.1.2.6 $, 10);
-
-BEGIN {
-	eval {
-		require 5.004;
-	};
-}
+$VERSION = substr(q$Revision: 2.71 $, 10);
 
 my $Eval_routine;
 my $Eval_routine_file;
@@ -106,11 +113,24 @@ $ESCAPE_CHARS::ok_in_url =
 		'-_./~='
 	;
 
+## This is a character class for HTML::Entities
+$ESCAPE_CHARS::std = "^\n\t !\#\$%\'-;=?-Z\\\]-~";
+
+## Some standard error templates
+
+## This is an alias for a commonly-used function
+*dbref = \&Vend::Data::database_exists_ref;
+
 my $need_escape;
 
 sub setup_escape_chars {
     my($ok, $i, $a, $t);
 
+	## HTML::Entities caches this, let's get it cached right away so
+	## each child doesn't have to re-eval
+	my $junk = ">>>123<<<";
+	HTML::Entities::encode($junk, $ESCAPE_CHARS::std);
+
     foreach $i (0..255) {
         $a = chr($i);
         if (index($ESCAPE_CHARS::ok_in_filename,$a) == -1) {
@@ -131,7 +151,6 @@ sub setup_escape_chars {
 
 	my $string = "[^$ESCAPE_CHARS::ok_in_url]";
 	$need_escape = qr{$string};
-
 }
 
 # Replace any characters that might not be safe in a filename (especially
@@ -155,6 +174,7 @@ sub escape_chars {
 
 sub escape_chars_url {
     my($in) = @_;
+	return $in unless $in =~ $need_escape;
     my($c, $r);
 
     $r = '';
@@ -240,10 +260,10 @@ sub round_to_frac_digits {
 		$digits = 2;
 	}
 	my @frac;
-	$num =~ /^(\d*)\.(\d+)$/
+	$num =~ /^(\d*)(?:\.(\d+))?$/
 		or return $num;
 	my $int = $1;
-	@frac = split //, $2;
+	@frac = split(//, ($2 || 0));
 	local($^W) = 0;
 	my $frac = join "", @frac[0 .. $digits - 1];
 	if($frac[$digits] > 4) {
@@ -402,31 +422,75 @@ sub setlocale {
             $Vend::Cfg->{$_} = $curr->{$_}
                 if defined $curr->{$_};
         }
-        @{$Vend::Cfg->{Locale}}{@Vend::Config::Locale_keys_currency} =
-                @{$curr}{@Vend::Config::Locale_keys_currency};
     }
 
-    $::Scratch->{mv_locale}   = $locale    if $opt->{persist} and $locale;
-    $::Scratch->{mv_currency} = $currency  if $opt->{persist} and $currency;
+	if(my $ref = $Vend::Cfg->{CodeDef}{LocaleChange}) {
+		$ref = $ref->{Routine};
+		if($ref->{all}) {
+			$ref->{all}->($locale, $opt);
+		}
+		if($ref->{lc $locale}) {
+			$ref->{lc $locale}->($locale, $opt);
+		}
+	}
+
+    if($opt->{persist}) {
+		$::Scratch->{mv_locale}   = $locale		if $locale;
+		delete $::Scratch->{mv_currency_tmp};
+		delete $::Scratch->{mv_currency};
+		$::Scratch->{mv_currency} = $currency if $currency;
+	}
+	elsif($currency) {
+		Vend::Interpolate::set_tmp('mv_currency_tmp')
+			unless defined $::Scratch->{mv_currency_tmp};
+		$::Scratch->{mv_currency_tmp} = $currency;
+	}
+	else {
+		delete $::Scratch->{mv_currency_tmp};
+		delete $::Scratch->{mv_currency};
+	}
+
     return '';
 }
 
 
 sub currency {
 	my($amount, $noformat, $convert, $opt) = @_;
-#::logDebug("currency called: amount=$amount no=$noformat convert=$convert");
 	$opt = {} unless $opt;
-	
-	$amount = $amount / $Vend::Cfg->{PriceDivide}
-		if $convert and $Vend::Cfg->{PriceDivide} != 0;
+
+	my $pd = $Vend::Cfg->{PriceDivide};
+	if($opt->{locale}) {
+		$convert = 1;
+		$pd = $Vend::Cfg->{Locale_repository}{$opt->{locale}}{PriceDivide};
+	}
+
+	if($pd and $convert) {
+		$amount = $amount / $pd;
+	}
+
 	return $amount if $noformat;
-	my $loc;
 	my $sep;
 	my $dec;
 	my $fmt;
 	my $precede = '';
 	my $succede = '';
-	if ($loc = $opt->{locale} || $Vend::Cfg->{Locale}) {
+
+	my $loc = $opt->{locale}
+			|| $::Scratch->{mv_currency_tmp}
+			|| $::Scratch->{mv_currency}
+			|| $Vend::Cfg->{Locale};
+
+	if(ref($loc)) {
+		## Do nothing, is a hash reference
+	}
+	elsif($loc) {
+		$loc = $Vend::Cfg->{Locale_repository}{$loc};
+	}
+	
+	if (! $loc) {
+		$fmt = "%.2f";
+	}
+	else {
 		$sep = $loc->{mon_thousands_sep} || $loc->{thousands_sep} || ',';
 		$dec = $loc->{mon_decimal_point} || $loc->{decimal_point} || '.';
 		return picture_format($amount, $loc->{price_picture}, $sep, $dec)
@@ -444,9 +508,6 @@ sub currency {
 			}
 		}
 	}
-	else {
-		$fmt = "%.2f";
-	}
 
 	$amount = safe_sprintf($fmt, $amount);
 	$amount =~ s/\./$dec/ if defined $dec;
@@ -485,7 +546,6 @@ eval {require Digest::MD5 };
 if(! $@) {
 	$Md = new Digest::MD5;
 	$Keysub = sub {
-#::logDebug("key gen args: '@_'");
 					@_ = time() unless @_;
 					$Md->reset();
 					$Md->add(@_);
@@ -515,10 +575,19 @@ sub hexify {
 
 sub unhexify {
     my $s = shift;
-    $s =~ s/%(..)/chr(hex($1))/ge;
+    $s =~ s/%([0-9A-Fa-f][0-9A-Fa-f])/chr(hex($1))/ge;
     return $s;
 }
 
+*unescape_chars = \&unhexify;
+
+sub unescape_full {
+    my $url = shift;
+    $url =~ tr/+/ /;
+    $url =~ s/<!--.*?-->//sg;
+    return unhexify($url);
+}
+
 ## UNEVAL
 
 # Returns a string representation of an anonymous array, hash, or scaler
@@ -609,68 +678,6 @@ eval {
 *uneval_file = defined $Fast_uneval_file  ? $Fast_uneval_file  : \&uneval_it_file;
 *uneval      = defined $Pretty_uneval     ? $Pretty_uneval     : \&uneval_it;
 
-sub writefile {
-    my($file, $data, $opt) = @_;
-
-	$file = ">>$file" unless $file =~ /^[|>]/;
-	if (ref $opt and $opt->{umask}) {
-		$opt->{umask} = umask oct($opt->{umask});
-	}
-    eval {
-		unless($file =~ s/^[|]\s*//) {
-			if (ref $opt and $opt->{auto_create_dir}) {
-				my $dir = $file;
-				$dir =~ s/>+//;
-
-				## Need to make this OS-independent, requires File::Spec support
-				$dir =~ s:[\r\n]::g;   # Just in case
-				$dir =~ s:(.*)/.*:: or $dir = '';
-				if($dir and ! -d $dir) {
-					File::Path::mkpath($dir);
-				}
-			}
-			# We have checked for beginning > or | previously
-			open(MVLOGDATA, $file) or die "open\n";
-			lockfile(\*MVLOGDATA, 1, 1) or die "lock\n";
-			seek(MVLOGDATA, 0, 2) or die "seek\n";
-			if(ref $data) {
-				print(MVLOGDATA $$data) or die "write to\n";
-			}
-			else {
-				print(MVLOGDATA $data) or die "write to\n";
-			}
-			unlockfile(\*MVLOGDATA) or die "unlock\n";
-		}
-		else {
-            my (@args) = grep /\S/, Text::ParseWords::shellwords($file);
-			open(MVLOGDATA, "|-") || exec @args;
-			if(ref $data) {
-				print(MVLOGDATA $$data) or die "pipe to\n";
-			}
-			else {
-				print(MVLOGDATA $data) or die "pipe to\n";
-			}
-		}
-		close(MVLOGDATA) or die "close\n";
-    };
-
-	my $status = 1;
-    if ($@) {
-		::logError ("Could not %s file '%s': %s\nto write this data:\n%s",
-				$@,
-				$file,
-				$!,
-				$data,
-				);
-		$status = 0;
-    }
-
-    if (ref $opt and defined $opt->{umask}) {                                        
-        $opt->{umask} = umask oct($opt->{umask});                                    
-    }
-
-	return $status;
-}
 
 
 # Log data fields to a data file.
@@ -712,11 +719,6 @@ sub logData {
 }
 
 
-sub file_modification_time {
-    my ($fn, $tolerate) = @_;
-    my @s = stat($fn) or ($tolerate and return 0) or die "Can't stat '$fn': $!\n";
-    return $s[9];
-}
 
 sub quoted_comma_string {
 	my ($text) = @_;
@@ -768,13 +770,10 @@ sub check_gate {
 
 	my $gate;
 	if ($gate = readfile("$gatedir/.access_gate") ) {
-#::logDebug("found access_gate");
 		$f =~ s:.*/::;
 		$gate = Vend::Interpolate::interpolate_html($gate);
-#::logDebug("f=$f gate=$gate");
 		if($gate =~ m!^$f(?:\.html?)?[ \t]*:!m ) {
 			$gate =~ s!.*(\n|^)$f(?:\.html?)?[ \t]*:!!s;
-#::logDebug("gate=$gate");
 			$gate =~ s/\n[\S].*//s;
 			$gate =~ s/^\s+//;
 		}
@@ -790,13 +789,56 @@ sub check_gate {
 
 sub string_to_ref {
 	my ($string) = @_;
-	if(! $Vend::Cfg->{ExtraSecure} and $MVSAFE::Safe) {
+	if($MVSAFE::Safe) {
 		return eval $string;
 	}
-	elsif ($MVSAFE::Safe) {
-		die ::errmsg("not allowed to eval in Safe mode.");
+	my $safe = $Vend::Interpolate::safe_safe || new Safe;
+	return $safe->reval($string);
+}
+
+sub is_hash {
+	return ref($_[0]) eq 'HASH';
+}
+
+sub dotted_hash {
+	my($hash, $key, $value, $delete_empty) = @_;
+	$hash = get_option_hash($hash) unless is_hash($hash);
+	unless (is_hash($hash)) {
+		return undef unless defined $value;
+		$hash = {};
 	}
-	return $Vend::Interpolate::safe_safe->reval($string);
+	my @keys = split /[\.:]+/, $key;
+	my $final;
+	my $ref;
+
+	if(! defined $value) {
+		# Retrieving
+		$ref = $hash->{shift @keys};
+		for(@keys) {
+			return undef unless is_hash($ref);
+			$ref = $ref->{$_};
+		}
+		return $ref;
+	}
+
+	# Storing
+	$final = pop @keys;
+	$ref = $hash;
+
+	for(@keys) {
+		$ref->{$_} = {} unless is_hash($ref->{$_});
+		$ref = $ref->{$_};
+	}
+
+	if($delete_empty and ! length($value)) {
+		delete $ref->{$final};
+	}
+	else {
+		$ref->{$final} = $value;
+	}
+
+	$hash = uneval_it($hash);
+	return $hash;
 }
 
 sub get_option_hash {
@@ -810,7 +852,7 @@ sub get_option_hash {
 		}
 		return $string;
 	}
-	return {} unless $string =~ /\S/;
+	return {} unless $string and $string =~ /\S/;
 	$string =~ s/^\s+//;
 	$string =~ s/\s+$//;
 	if($string =~ /^{/ and $string =~ /}/) {
@@ -821,7 +863,7 @@ sub get_option_hash {
 	unless ($string =~ /,/) {
 		@opts = grep $_ ne "=", Text::ParseWords::shellwords($string);
 		for(@opts) {
-			s/^(\w+)=(["'])(.*)\2$/$1$3/;
+			s/^(\w[-\w]*\w)=(["'])(.*)\2$/$1$3/;
 		}
 	}
 	else {
@@ -831,6 +873,7 @@ sub get_option_hash {
 	my %hash;
 	for(@opts) {
 		my ($k, $v) = split /[\s=]+/, $_, 2;
+		$k =~ s/-/_/g;
 		$hash{$k} = $v;
 	}
 	if($merge) {
@@ -843,14 +886,77 @@ sub get_option_hash {
 	return \%hash;
 }
 
+## Takes an IC scalar form value (parm=val\nparm2=val) and translates it
+## to a reference
+
+sub scalar_to_hash {
+	my $val = shift;
+
+	$val =~ s/^\s+//mg;
+	$val =~ s/\s+$//mg;
+	my @args;
+
+	@args = split /\n+/, $val;
+
+	my $ref = {};
+
+	for(@args) {
+		m!([^=]+)=(.*)!
+			and $ref->{$1} = $2;
+	}
+	return $ref;
+}
+
+## Takes a form reference (i.e. from \%CGI::values) and makes into a
+## scalar value value (i.e. parm=val\nparm2=val). Also translates it
+## via HTML entities -- it is designed to make it into a hidden
+## form value
+
+sub hash_to_scalar {
+	my $ref = shift
+		or return '';
+
+	unless (ref($ref) eq 'HASH') {
+		die __PACKAGE__ . " hash_to_scalar routine got bad reference.\n";
+	}
+
+	my @parms;
+	while( my($k, $v) = each %$ref ) {
+		$v =~ s/\r?\n/\r/g;
+		push @parms, HTML::Entities::encode("$k=$v");
+	}
+	return join "\n", @parms;
+}
+
+## This simply returns a hash of words, which may be quoted shellwords
+## Replaces most of parse_hash in Vend::Config
+sub hash_string {
+	my($settings, $ref) = @_;
+
+	return $ref if ! $settings or $settings !~ /\S/;
+
+	$ref ||= {};
+
+	$settings =~ s/^\s+//;
+	$settings =~ s/\s+$//;
+	my(@setting) = Text::ParseWords::shellwords($settings);
+
+	my $i;
+	for ($i = 0; $i < @setting; $i += 2) {
+		$ref->{$setting[$i]} = $setting[$i + 1];
+	}
+	return $ref;
+}
+
 ## READIN
 
 my $Lang;
 
 sub find_locale_bit {
 	my $text = shift;
-	$Lang = $::Scratch->{mv_locale} unless defined $Lang;
-#::logDebug("find_locale: $Lang");
+	unless (defined $Lang) {
+		$Lang = $::Scratch->{mv_locale} || $Vend::Cfg->{DefaultLocale};
+	}
 	$text =~ m{\[$Lang\](.*)\[/$Lang\]}s
 		and return $1;
 	$text =~ s{\[(\w+)\].*\[/\1\].*}{}s;
@@ -860,6 +966,8 @@ sub find_locale_bit {
 sub parse_locale {
 	my ($input) = @_;
 
+	return if $::Pragma->{no_locale_parse};
+
 	# avoid copying big strings
 	my $r = ref($input) ? $input : \$input;
 	
@@ -881,6 +989,29 @@ sub parse_locale {
 	return ref($input) ? $input : $$r;
 }
 
+sub teleport_name {
+	my ($file, $teleport, $table) = @_;
+	my $db;
+	return $file
+		unless	 $teleport
+			and  $db = Vend::Data::database_exists_ref($table);
+
+	my @f = qw/code base_code expiration_date show_date page_text/;
+	my ($c, $bc, $ed, $sd, $pt) = @{$Vend::Cfg->{PageTableMap}}{@f};
+	my $q = qq{
+		SELECT $c from $table
+		WHERE  $bc = '$file'
+		AND    $ed <  $teleport
+		AND    $sd >= $teleport
+		ORDER BY $sd DESC
+	};
+	my $ary = $db->query($q);
+	if($ary and $ary->[0]) {
+		$file = $ary->[0][0];
+	}
+	return $file;
+}
+
 # Reads in a page from the page directory with the name FILE and ".html"
 # appended. If the HTMLsuffix configuration has changed (because of setting in
 # catalog.cfg or Locale definitions) it will substitute that. Returns the
@@ -888,14 +1019,26 @@ sub parse_locale {
 # Substitutes Locale bits as necessary.
 
 sub readin {
-    my($file, $only) = @_;
+    my($file, $only, $locale) = @_;
+
+	## We don't want to try if we are forcing a flypage
+	return undef if $Vend::ForceFlypage;
+
     my($fn, $contents, $gate, $pathdir, $dir, $level);
     local($/);
 
+	if($file =~ m{[\[<]}) {
+		::logGlobal("Possible code/SQL injection attempt with file name '%s'", $file);
+		$file = escape_chars($file);
+		::logGlobal("Suspect file changed to '%s'", $file);
+	}
+
 	$Global::Variable->{MV_PREV_PAGE} = $Global::Variable->{MV_PAGE}
 		if defined $Global::Variable->{MV_PAGE};
 	$Global::Variable->{MV_PAGE} = $file;
 
+	$file =~ s#^\s+##;
+	$file =~ s#\s+$##;
 	$file =~ s#\.html?$##;
 	if($file =~ m{\.\.} and $file =~ /\.\..*\.\./) {
 		::logError( "Too many .. in file path '%s' for security.", $file );
@@ -907,12 +1050,35 @@ sub readin {
 	$pathdir =~ s:^/+::;
 	my $try;
 	my $suffix = $Vend::Cfg->{HTMLsuffix};
+	my $db_tried;
+	$locale = 1 unless defined $locale;
+	my $record;
   FINDPAGE: {
-	foreach $try (
-					$Vend::Cfg->{PageDir},
-					@{$Vend::Cfg->{TemplateDir}},
-					@{$Global::TemplateDir}          )
-	{
+  	## If PageTables is set, we try to find the page in the table first
+	## but only once, without the suffix
+  	if(! $db_tried++ and $Vend::Cfg->{PageTables}) {
+		my $teleport = $Vend::Session->{teleport};
+		my $field = $Vend::Cfg->{PageTableMap}{page_text};
+		foreach my $t (@{$Vend::Cfg->{PageTables}}) {
+			my $db = Vend::Data::database_exists_ref($t);
+			if($teleport) {
+				$file = teleport_name($file, $teleport, $t);
+			}
+			$record = $db->row_hash($file)
+				or next;
+			$contents = $record->{$field};
+			last FINDPAGE if length $contents;
+			undef $contents;
+		}
+	}
+
+	my @dirs = ($Vend::Cfg->{PreviewDir},
+				$Vend::Cfg->{PageDir},
+				@{$Vend::Cfg->{TemplateDir}},
+				@{$Global::TemplateDir});
+
+	foreach $try (@dirs) {
+		next unless $try;
 		$dir = $try . "/" . $pathdir;
 		if (-f "$dir/.access") {
 			if (-s _) {
@@ -932,10 +1098,6 @@ sub readin {
 
 		if( defined $level and ! check_security($file, $level, $gate) ){
 			my $realm = $::Variable->{COMPANY} || $Vend::Cat;
-			$Vend::StatusLine = <<EOF if $Vend::InternalHTTP;
-HTTP/1.0 401 Unauthorized
-WWW-Authenticate: Basic realm="$realm"
-EOF
 			if(-f "$try/violation$suffix") {
 				$fn = "$try/violation$suffix";
 			}
@@ -969,179 +1131,151 @@ EOF
 	}
 
 	return unless defined $contents;
-
+	
 	parse_locale(\$contents);
 
-	return $contents;
-}
-
-sub readfile_db {
-	my ($name) = @_;
-	return unless $Vend::Cfg->{FileDatabase};
-	my ($tab, $col) = split /:+/, $Vend::Cfg->{FileDatabase};
-	my $db = $Vend::Interpolate::Db{$tab} || ::database_exists_ref($tab)
-		or return undef;
-#::logDebug("tab=$tab exists, db=$db");
-
-	# I guess this is the best test
-	if($col) {
-		return undef unless $db->column_exists($col);
-	}
-	elsif ( $col = $Global::Variable->{LANG} and $db->column_exists($col) ) {
-		#do nothing
-	}
-	else {
-		$col = 'default';
-		return undef unless $db->column_exists($col);
-	}
-
-#::logDebug("col=$col exists, db=$db");
-	return undef unless $db->record_exists($name);
-#::logDebug("ifile=$name exists, db=$db");
-	return $db->field($name, $col);
-}
-
-# Reads in an arbitrary file.  Returns the entire contents,
-# or undef if the file could not be read.
-# Careful, needs the full path, or will be read relative to
-# VendRoot..and will return binary. Should be tested by
-# the user.
-
-# Will also look in the *global* TemplateDir. (No need for the
-# extra overhead of local TemplateDir, probably also insecure.)
-#
-# To ensure security in multiple catalog setups, leading /
-# is not allowed if the second subroutine argument passed
-# (caller usually sends $Global::NoAbsolute) is true.
-
-# If catalog FileDatabase is enabled and there are no contents, we can retrieve
-# the file from the database.
-
-sub readfile {
-    my($ifile, $no, $loc) = @_;
-    my($contents);
-    local($/);
-
-	if($no and (file_name_is_absolute($ifile) or $ifile =~ m#\.\./.*\.\.#)) {
-		::logError("Can't read file '%s' with NoAbsolute set" , $ifile);
-		::logGlobal({ level => 'auth'}, "Can't read file '%s' with NoAbsolute set" , $ifile );
-		return undef;
-	}
-
-	my $file;
-
-#::logDebug("readfile ifile=$ifile");
-	if (file_name_is_absolute($ifile) and -f $ifile) {
-		$file = $ifile;
-	}
-	else {
-		for( ".", @{$Global::TemplateDir} ) {
-#::logDebug("trying ifile=$_/$ifile");
-			next if ! -f "$_/$ifile";
-			$file = "$_/$ifile";
-#::logDebug("readfile file=$file FOUND");
-			last;
-		}
-	}
-
-	if(! $file) {
-		$contents = readfile_db($ifile);
-		return undef unless defined $contents;
-	}
-	else {
-		return undef unless open(READIN, "< $file");
-		$Global::Variable->{MV_FILE} = $file;
-
-		binmode(READIN) if $Global::Windows;
-		undef $/;
-		$contents = <READIN>;
-		close(READIN);
-	}
-
-	if (
-		$Vend::Cfg->{Locale}
-			and
-		(defined $loc ? $loc : $Vend::Cfg->{Locale}->{readfile} )
-		)
-	{
-		my $key;
-		$contents =~ s~\[L(\s+([^\]]+))?\]([\000-\377]*?)\[/L\]~
-						$key = $2 || $3;		
-						defined $Vend::Cfg->{Locale}->{$key}
-						?  ($Vend::Cfg->{Locale}->{$key})	: $3 ~eg;
-		$contents =~ s~\[LC\]([\000-\377]*?)\[/LC\]~
-						find_locale_bit($1) ~eg;
-		undef $Lang;
-	}
-    return $contents;
+	return $contents unless wantarray;
+	return ($contents, $record);
 }
 
 sub is_yes {
-    return( defined($_[$[]) && ($_[$[] =~ /^[yYtT1]/));
+    return( defined($_[0]) && ($_[0] =~ /^[yYtT1]/));
 }
 
 sub is_no {
-	return( !defined($_[$[]) || ($_[$[] =~ /^[nNfF0]/));
+	return( !defined($_[0]) || ($_[0] =~ /^[nNfF0]/));
 }
 
 # Returns a URL which will run the ordering system again.  Each URL
 # contains the session ID as well as a unique integer to avoid caching
 # of pages by the browser.
 
+my @scratches = qw/
+				add_dot_html
+				add_source
+				link_relative
+				match_security
+				no_count
+				no_session
+				/;
+
 sub vendUrl {
-    my($path, $arguments, $r) = @_;
+    my($path, $arguments, $r, $opt) = @_;
+
+	$opt ||= {};
+
+	if($opt->{auto_format}) {
+		return $path if $path =~ m{^/};
+		$path =~ s:#([^/.])+$::
+            and $opt->{anchor} = $1;
+		$path =~ s/\.html?$//i
+			and $opt->{add_dot_html} = 1;
+	}
+
     $r = $Vend::Cfg->{VendURL}
 		unless defined $r;
 
-	my $can_cache = ! $Vend::Cfg->{NoCache}{$path};
+	my $secure;
 	my @parms;
 
 	if(exists $Vend::Cfg->{AlwaysSecure}{$path}) {
 		$r = $Vend::Cfg->{SecureURL};
 	}
 
+	my %skip = qw/form 1 href 1 reparse 1/;
+
+	for(@scratches) {
+		next if defined $opt->{$_};
+		next unless defined $::Scratch->{"mv_$_"};
+		$skip{$_} = 1;
+		$opt->{$_} = $::Scratch->{"mv_$_"};
+	}
+
+	my $extra;
+	if($opt->{form}) {
+		$path = $Vend::Cfg->{ProcessPage} unless $path;
+		if($opt->{form} eq 'auto') {
+			my $form = '';
+			while( my ($k, $v) = each %$opt) {
+				next if $skip{$k};
+				$k =~ s/^__//;
+				$form .= "$k=$v\n";
+			}
+			$opt->{form} = $form;
+		}
+		push @parms, Vend::Interpolate::escape_form($opt->{form});
+	}
+
 	my($id, $ct);
 	$id = $Vend::SessionID
-		unless $can_cache and $Vend::Cookie && $::Scratch->{mv_no_session_id};
+		unless $opt->{no_session_id}
+		or     ($Vend::Cookie and $::Scratch->{mv_no_session_id});
 	$ct = ++$Vend::Session->{pageCount}
-		unless $can_cache and $::Scratch->{mv_no_count};
+		unless $opt->{no_count};
+
+	if($opt->{match_security}) {
+		$opt->{secure} = $CGI::secure;
+	}
+
+	if($opt->{no_session}) {
+		undef $id;
+		undef $ct;
+	}
+
+	if($opt->{link_relative}) {
+		my $cur = $Global::Variable->{MV_PAGE};
+		$cur =~ s{/[^/]+$}{}
+			and $path = "$cur/$path";
+	}
+
+	if($opt->{secure} or exists $Vend::Cfg->{AlwaysSecure}{$path}) {
+		$r = $Vend::Cfg->{SecureURL};
+	}
 
 	$path = escape_chars_url($path)
 		if $path =~ $need_escape;
-    $r .= '/' . $path;
-	$r .= '.html' if $::Scratch->{mv_add_dot_html} and $r !~ /\.html?$/;
-	push @parms, "$::VN->{mv_session_id}=$id"			 	if defined $id;
-	push @parms, "$::VN->{mv_arg}=" . hexify($arguments)	if defined $arguments;
-	push @parms, "$::VN->{mv_pc}=$ct"                 	if defined $ct;
-	push @parms, "$::VN->{mv_cat}=$Vend::Cat"
-														if defined $Vend::VirtualCat;
-	if($Vend::AccumulatingLinks) {
-		my $key = $path;
-		$key =~ s/\.html?$//;
-		my $value = '';
-		if($arguments) {
-			$value = [ $key, $arguments ];
-			$key .= "/$arguments";
-		}
-		push(@Vend::Links, [$key, $value]) unless $Vend::LinkFound{$key}++;
+    	$r .= '/' . $path;
+	$r .= '.html' if $opt->{add_dot_html} and $r !~ m{(?:/|\.html?)$};
 
+	if($opt->{add_source} and $Vend::Session->{source}) {
+		my $sn = hexify($Vend::Session->{source});
+		push @parms, "$::VN->{mv_source}=$sn";
 	}
-	return $r unless @parms;
-    return $r . '?' . join($Global::UrlJoiner, @parms);
+
+	push @parms, "$::VN->{mv_session_id}=$id"		if $id;
+	push @parms, "$::VN->{mv_arg}=" . hexify($arguments) 	if defined $arguments;
+	push @parms, "$::VN->{mv_pc}=$ct"                 	if $ct;
+	push @parms, "$::VN->{mv_cat}=$Vend::Cat"            	if $Vend::VirtualCat;
+
+    	$r .= '?' . join($Global::UrlJoiner, @parms) if @parms;
+	if($opt->{anchor}) {
+		$opt->{anchor} =~ s/^#//;
+		$r .= '#' . $opt->{anchor};
+	}
+	return $r;
 } 
 
 sub secure_vendUrl {
-	return vendUrl($_[0], $_[1], $Vend::Cfg->{SecureURL});
+	return vendUrl($_[0], $_[1], $Vend::Cfg->{SecureURL}, $_[3]);
 }
 
+my %strip_vars;
+my $strip_init;
+
 sub change_url {
 	my $url = shift;
-	return $url if $url =~ m{^(?:\w+:)?/};
-#::logDebug("changed $url");
+	return $url if $url =~ m{^\w+:};
+	return $url if $url =~ m{^/};
+	if(! $strip_init) {
+		for(qw/mv_session_id mv_pc/) {
+			$strip_vars{$_} = 1;
+			$strip_vars{$::IV->{$_}} = 1;
+		}
+	}
 	my $arg;
 	my @args;
 	($url, $arg) = split /[?&]/, $url, 2;
-	@args = split $Global::UrlSplittor, $arg;
+	@args = grep ! $strip_vars{$_}, split $Global::UrlSplittor, $arg;
 	return Vend::Interpolate::tag_area( $url, '', {
 											form => join "\n", @args,
 										} );
@@ -1161,107 +1295,6 @@ my $flock_LOCK_EX = 2;          # Exclusive lock
 my $flock_LOCK_NB = 4;          # Don't block when locking
 my $flock_LOCK_UN = 8;          # Unlock
 
-sub flock_lock {
-    my ($fh, $excl, $wait) = @_;
-    my $flag = $excl ? $flock_LOCK_EX : $flock_LOCK_SH;
-
-    if ($wait) {
-        flock($fh, $flag) or die "Could not lock file: $!\n";
-        return 1;
-    }
-    else {
-        if (! flock($fh, $flag | $flock_LOCK_NB)) {
-            if ($!{EAGAIN} or $!{EWOULDBLOCK}) {
-				return 0;
-            }
-            else {
-                die "Could not lock file: $!\n";
-            }
-        }
-        return 1;
-    }
-}
-
-sub flock_unlock {
-    my ($fh) = @_;
-    flock($fh, $flock_LOCK_UN) or die "Could not unlock file: $!\n";
-}
-
-sub fcntl_lock {
-    my ($fh, $excl, $wait) = @_;
-    my $flag = $excl ? F_WRLCK : F_RDLCK;
-    my $op = $wait ? F_SETLKW : F_SETLK;
-
-	my $struct = pack('sslli', $flag, 0, 0, 0, $$);
-
-    if ($wait) {
-        fcntl($fh, $op, $struct) or die "Could not fcntl_lock file: $!\n";
-        return 1;
-    }
-    else {
-        if (fcntl($fh, $op, $struct) < 0) {
-            if ($!{EAGAIN} or $!{EWOULDBLOCK}) {
-                return 0;
-            }
-            else {
-                die "Could not lock file: $!\n";
-            }
-        }
-        return 1;
-    }
-}
-
-sub fcntl_unlock {
-    my ($fh) = @_;
-	my $struct = pack('sslli', F_UNLCK, 0, 0, 0, $$);
-	if (fcntl($fh, F_SETLK, $struct) < 0) {
-		if ($!{EAGAIN} or $!{EWOULDBLOCK}) {
-			return 0;
-		}
-		else {
-			die "Could not un-fcntl_lock file: $!\n";
-		}
-	}
-	return 1;
-}
-
-my $lock_function = \&flock_lock;
-my $unlock_function = \&flock_unlock;
-
-sub set_lock_type {
-	if ($Global::LockType eq 'none') {
-		logDebug("using NO locking");
-		$lock_function = sub {1};
-		$unlock_function = sub {1};
-	}
-	elsif ($Global::LockType =~ /fcntl/i) {
-		logDebug("using fcntl(2) locking");
-		$lock_function = \&fcntl_lock;
-		$unlock_function = \&fcntl_unlock;
-	}
-	else {
-		$lock_function = \&flock_lock;
-		$unlock_function = \&flock_unlock;
-	}
-	return; # VOID
-}
- 
-sub lockfile {
-    &$lock_function(@_);
-}
-
-sub unlockfile {
-    &$unlock_function(@_);
-}
-
-### Still necessary, sad to say.....
-if($Global::Windows) {
-	set_lock_type('none');
-}
-elsif($^O =~ /hpux/) {
-	set_lock_type('fcntl');
-}
-
 # Returns the total number of items ordered.
 # Uses the current cart if none specified.
 
@@ -1281,9 +1314,10 @@ sub tag_nitems {
 	if($opt->{qualifier}) {
 		$attr = $opt->{qualifier};
 		my $qr;
-		$qr = qr{$opt->{compare}}
-			if $opt->{compare};
-		if($opt->{compare}) {
+		eval { 
+			$qr = qr{$opt->{compare}} if $opt->{compare};
+		};
+		if($qr) {
 			$sub = sub { 
 							$_[0] =~ $qr;
 						};
@@ -1404,7 +1438,6 @@ sub check_security {
 	}
 	elsif ($reconfig eq '2') {
 		$msg = "access protected database $item";
-#::logDebug("passed gate of $gate");
 		return 1 if is_yes($gate);
 	}
 	elsif ($reconfig eq '3') {
@@ -1520,15 +1553,6 @@ EOF
 	return 1;
 }
 
-# Replace the escape notation %HH with the actual characters.
-#
-sub unescape_chars {
-    my($in) = @_;
-
-    $in =~ s/%(..)/chr(hex($1))/ge;
-    $in;
-}
-
 
 # Checks the Locale for a special page definintion mv_special_$key and
 # returns it if found, otherwise goes to the default Vend::Cfg->{Special} array
@@ -1570,6 +1594,8 @@ sub errmsg {
 	return scalar(@strings) ? sprintf $fmt, @strings : $fmt;
 }
 
+*l = \&errmsg;
+
 sub show_times {
 	my $message = shift || 'time mark';
 	my @times = times();
@@ -1580,6 +1606,7 @@ sub show_times {
 }
 
 sub logGlobal {
+	return 1 if $Vend::ExternalProgram;
     my($msg) = shift;
 	my $opt;
 	if(ref $msg) {
@@ -1628,7 +1655,7 @@ sub logGlobal {
 
 	$fn =~ s/^([^|>])/>>$1/
 		or $nolock = 1;
-#::logDebug("logging with $fn");
+
     $msg = format_log_msg($msg) if ! $nolock;
 
 	$Vend::Errors .= $msg if $Global::DisplayErrors;
@@ -1695,9 +1722,36 @@ sub logError {
     }
 }
 
+# Front-end to log routines that ignores repeated identical
+# log messages after the first occurrence
+my %logOnce_cache;
+my %log_sub_map = (
+	data	=> \&logData,
+	debug	=> \&logDebug,
+	error	=> \&logError,
+	global	=> \&logGlobal,
+);
+
+# First argument should be log type (see above map).
+# Rest of arguments are same as if calling log routine directly.
+sub logOnce {
+	return if exists $logOnce_cache{"@_"};
+	my $log_sub = $log_sub_map{ lc(shift) } || $log_sub_map{error};
+	$log_sub->(@_) and ++$logOnce_cache{"@_"};
+}
+
+
 # Here for convenience in calls
 sub set_cookie {
     my ($name, $value, $expire, $domain, $path) = @_;
+
+    # Set expire to now + some time if expire string is something like
+    # "30 days" or "7 weeks" or even "60 minutes"
+	if($expire =~ /^\s*\d+[\s\0]*[A-Za-z]\S*\s*$/) {
+		my $add = Vend::Config::time_to_seconds($expire);
+		$expire = time() + $add if $add;
+	}
+
 	if (! $::Instance->{Cookies}) {
 		$::Instance->{Cookies} = []
 	}
@@ -1718,217 +1772,6 @@ sub read_cookie {
  	return unescape_chars($1);
 }
 
-# Return a quasi-hashed directory/file combo, creating if necessary
-sub exists_filename {
-    my ($file,$levels,$chars, $dir) = @_;
-	my $i;
-	$levels = 1 unless defined $levels;
-	$chars = 1 unless defined $chars;
-	$dir = $Vend::Cfg->{ScratchDir} unless $dir;
-    for($i = 0; $i < $levels; $i++) {
-		$dir .= "/";
-		$dir .= substr($file, $i * $chars, $chars);
-		return 0 unless -d $dir;
-	}
-	return -f "$dir/$file" ? 1 : 0;
-}
-
-# Return a quasi-hashed directory/file combo, creating if necessary
-sub get_filename {
-    my ($file,$levels,$chars, $dir) = @_;
-	my $i;
-	$levels = 1 unless defined $levels;
-	$chars = 1 unless defined $chars;
-	$dir = $Vend::Cfg->{ScratchDir} unless $dir;
-    for($i = 0; $i < $levels; $i++) {
-		$dir .= "/";
-		$dir .= substr($file, $i * $chars, $chars);
-		mkdir $dir, 0777 unless -d $dir;
-	}
-    die "Couldn't make directory $dir (or parents): $!\n"
-		unless -d $dir;
-    return "$dir/$file";
-}
-
-# These were stolen from File::Spec
-# Can't use that because it INSISTS on object
-# calls without returning a blessed object
-
-my $abspat = $^O =~ /win32/i ? '^([a-z]:)?[\\\\/]' : '^/';
-
-sub file_name_is_absolute {
-    my($file) = @_;
-    $file =~ m{$abspat}oi ;
-}
-
-sub win_catfile {
-    my $file = pop @_;
-    return $file unless @_;
-    my $dir = catdir(@_);
-    $dir =~ s/(\\\.)$//;
-    $dir .= "\\" unless substr($dir,length($dir)-1,1) eq "\\";
-    return $dir.$file;
-}
-
-sub unix_catfile {
-    my $file = pop @_;
-    return $file unless @_;
-    my $dir = catdir(@_);
-    for ($dir) {
-	$_ .= "/" unless substr($_,length($_)-1,1) eq "/";
-    }
-    return $dir.$file;
-}
-
-sub unix_path {
-    my $path_sep = ":";
-    my $path = $ENV{PATH};
-    my @path = split $path_sep, $path;
-    foreach(@path) { $_ = '.' if $_ eq '' }
-    @path;
-}
-
-sub win_path {
-    local $^W = 1;
-    my $path = $ENV{PATH} || $ENV{Path} || $ENV{'path'};
-    my @path = split(';',$path);
-    foreach(@path) { $_ = '.' if $_ eq '' }
-    @path;
-}
-
-sub win_catdir {
-    my @args = @_;
-    for (@args) {
-	# append a slash to each argument unless it has one there
-	$_ .= "\\" if $_ eq '' or substr($_,-1) ne "\\";
-    }
-    my $result = canonpath(join('', @args));
-    $result;
-}
-
-sub win_canonpath {
-    my($path) = @_;
-    $path =~ s/^([a-z]:)/\u$1/;
-    $path =~ s|/|\\|g;
-    $path =~ s|\\+|\\|g ;                          # xx////xx  -> xx/xx
-    $path =~ s|(\\\.)+\\|\\|g ;                    # xx/././xx -> xx/xx
-    $path =~ s|^(\.\\)+|| unless $path eq ".\\";   # ./xx      -> xx
-    $path =~ s|\\$|| 
-             unless $path =~ m#^([a-z]:)?\\#;      # xx/       -> xx
-    $path .= '.' if $path =~ m#\\$#;
-    $path;
-}
-
-sub unix_canonpath {
-    my($path) = @_;
-    $path =~ s|/+|/|g ;                            # xx////xx  -> xx/xx
-    $path =~ s|(/\.)+/|/|g ;                       # xx/././xx -> xx/xx
-    $path =~ s|^(\./)+|| unless $path eq "./";     # ./xx      -> xx
-    $path =~ s|/$|| unless $path eq "/";           # xx/       -> xx
-    $path;
-}
-
-sub unix_catdir {
-    my @args = @_;
-    for (@args) {
-	# append a slash to each argument unless it has one there
-	$_ .= "/" if $_ eq '' or substr($_,-1) ne "/";
-    }
-    my $result = join('', @args);
-    # remove a trailing slash unless we are root
-    substr($result,-1) = ""
-	if length($result) > 1 && substr($result,-1) eq "/";
-    $result;
-}
-
-
-my $catdir_routine;
-my $canonpath_routine;
-my $catfile_routine;
-my $path_routine;
-
-if($^O =~ /win32/i) {
-	$catdir_routine = \&win_catdir;
-	$catfile_routine = \&win_catfile;
-	$path_routine = \&win_path;
-	$canonpath_routine = \&win_canonpath;
-}
-else {
-	$catdir_routine = \&unix_catdir;
-	$catfile_routine = \&unix_catfile;
-	$path_routine = \&unix_path;
-	$canonpath_routine = \&unix_canonpath;
-}
-
-sub path {
-	return &{$path_routine}(@_);
-}
-
-sub catfile {
-	return &{$catfile_routine}(@_);
-}
-
-sub catdir {
-	return &{$catdir_routine}(@_);
-}
-
-sub canonpath {
-	return &{$canonpath_routine}(@_);
-}
-
-#print "catfile a b c --> " . catfile('a', 'b', 'c') . "\n";
-#print "catdir a b c --> " . catdir('a', 'b', 'c') . "\n";
-#print "canonpath a/b//../../c --> " . canonpath('a/b/../../c') . "\n";
-#print "file_name_is_absolute a/b/c --> " . file_name_is_absolute('a/b/c') . "\n";
-#print "file_name_is_absolute a:b/c --> " . file_name_is_absolute('a:b/c') . "\n";
-#print "file_name_is_absolute /a/b/c --> " . file_name_is_absolute('/a/b/c') . "\n";
-
-#my $MIME_Lite;
-#eval {
-#	require MIME::Lite;
-#	$MIME_Lite = 1;
-#	require Net::SMTP;
-#};
-#
-#sub mime_lite_send {
-#	my ($opt, $body) = @_;
-#
-#	if(! $MIME_Lite) {
-#		return send_mail(
-#			$opt->{to},
-#			$opt->{subject},
-#			$body,
-#			$opt->{reply},
-#			undef,
-#			split /\n+/, $opt->{extra}
-#			);
-#	}
-#
-#	my %special = qw/
-#		as_string 1
-#		internal  1
-#	/;
-#	my $mime = new MIME::Lite;;
-#
-#	my @ary;
-#	my $popt = {};
-#	my $mopt = {};
-#	for(keys %$opt) {
-#		if(ref $opt->{$_}) {
-#			push @ary, [ $_, delete $opt->{$_} ];
-#		}
-#		if($special{$_}) {
-#			$popt->{$_} = delete $opt->{$_};
-#		}
-#		my $m = $_;
-#		s/_/-/g;
-#		s/(\w+)/\L\u$1/g;
-#		s/-/_/g;
-#		$mopt->{$_} = $opt->{$m};
-#	}
-#
-#}
-
 sub send_mail {
 	my($to, $subject, $body, $reply, $use_mime, @extra_headers) = @_;
 
@@ -1938,7 +1781,6 @@ sub send_mail {
 		$body = $subject;
 		undef $subject;
 		for(@$head) {
-#::logDebug("processing head=$_");
 			if( /^To:\s*(.+)/ ) {
 				$to = $1;
 			}
@@ -1986,7 +1828,7 @@ sub send_mail {
 #::logDebug("testing sendmail send none=$none");
 		last SEND if $none;
 #::logDebug("in Sendmail send $using");
-		open(MVMAIL,"|$Vend::Cfg->{SendMailProgram} $to") or last SEND;
+		open(MVMAIL,"|$Vend::Cfg->{SendMailProgram} -t") or last SEND;
 		my $mime = '';
 		$mime = Vend::Interpolate::mime('header', {}, '') if $use_mime;
 		print MVMAIL "To: $to\n", $reply, "Subject: $subject\n"
@@ -2106,5 +1948,21 @@ sub send_mail {
 	$ok;
 }
 
+### Provide stubs for former Vend::Util functions relocated to Vend::File
+*canonpath = \&Vend::File::canonpath;
+*catdir = \&Vend::File::catdir;
+*catfile = \&Vend::File::catfile;
+*exists_filename = \&Vend::File::exists_filename;
+*file_modification_time = \&Vend::File::file_modification_time;
+*file_name_is_absolute = \&Vend::File::file_name_is_absolute;
+*get_filename = \&Vend::File::get_filename;
+*lockfile = \&Vend::File::lockfile;
+*path = \&Vend::File::path;
+*readfile = \&Vend::File::readfile;
+*readfile_db = \&Vend::File::readfile_db;
+*set_lock_type = \&Vend::File::set_lock_type;
+*unlockfile = \&Vend::File::unlockfile;
+*writefile = \&Vend::File::writefile;
+
 1;
 __END__
